#ifndef _RXCOMPOUNDBODY_H_WRAP_
#define _RXCOMPOUNDBODY_H_WRAP_

#include <RxDlist.h>
#include <RxCvxPolyhed3.h>
#include <RxTrsf.h>
#include <RxPrintable.h>
#include <RxDistReport3.h>
#include <RxMaterial.h>

class RxCompoundBody;

typedef RxDlist<RxCompoundBody> RxCompoundBodyList;

/** A compound body is defined by either a single convex polyhedron,
  * or by a list of other compound bodies.
  */
class RxCompoundBody : public RxDnodeBase, public RxPrintable, public RxSolidType3
{
	// A rigid body either points to a single convex polyhedron,
	// or a list of other rigid bodies.
protected:
	char *name_;
	RxCompoundBodyList list;
	RxCvxPolyhed3 *poly_;
	RxTrsf *TPLp;		// transform from parent to local frame
	RxTrsf TRL;		// cumulative transform from root to local frame
	RxMaterial *mat;
	RxCompoundBody *parent;
	RxCompoundBody *curNode;
	void init();
	const RxCompoundBody *_getChild(int i) const;

	static class RxMuxedNameTable nameTable;

	struct DistRecInfo {
		RxDistRecord3Dlist *rlist;
		int numrecs;
		int maxrecs;
	  };
	double calcDist (const RxCompoundBody &body, const RxTrsf &T01, 
			 double distlim, DistRecInfo *dreci, 
			 RxFeatureHashTab *htab) const;
	double boundingDist (const RxCompoundBody &body) const;

	void setTPLRaw (const RxTrsf *TPLp);
	void updateTRL (int recurse);
//	void xform (const RxTrsf &TRP);
//	void copy (const RxCompoundBody &body);
//	int scanPoly (RxParseObj *obj);
	virtual RxCompoundBody *newBody();
	virtual void copyBody(const RxCompoundBody &cmpd);
	virtual int oscanBody (const char *fname, 
			      class RxParseObj *&lobj, class RxParseObj *obj);
	virtual int cprintBody (class RxCoutput *co,class RxPrintSpec *ps);
public:
	/// Class identifier constant.
#define RX_COMPOUND_BODY_ID	5
	static const int ID;

	static RxCompoundBody *findByName(const char *name)
	 { return (RxCompoundBody*)nameTable.getEntry (name);
	 }

	/** Empty the body by removing all children and references to 
	  * polyhedra. The node's parent, name, and parent-to-local transform 
	  * are unchanged.
	  */
	void clear();

//	virtual RxCompoundBody *clone ();
	void set (RxCvxPolyhed3 *poly, const RxTrsf *TPLp=NULL);
	void set (const RxCompoundBody &cmpd, const RxTrsf *TPLp=NULL);

	/** Add a child body defined by a single polyhedron. The child's
	  * root-to-local transform is updated to reflect the child's 
	  * place in the current body tree.
	  *
	  * @param	poly	Polyhedron defining the child's shape.
	  * @param	TPLp	Transform from the parent to the local frame.
	  *			If NULL, then this transform is assumed to be
	  *			the identity.
	  * @return	Created child.
	  */
	RxCompoundBody *addChild (RxCvxPolyhed3 *poly,
				  const RxTrsf *TPLp=NULL);

	/** Add a child body defined by recursively copying body1. All
	  * root-to-local transforms in this copy are automatically updated
	  * to reflect the copie's place in the current body tree.
	  *
	  * @param  	cmpd	Compound body to be copied.
	  * @param	TPLp    Transform from the parent to the local frame.
	  *			If NULL, then this transform is assumed to be
	  *			the identity.
	  * @return     Created child.
	  */
	RxCompoundBody *addChild (const RxCompoundBody &cmpd, 
				  const RxTrsf *TPLp=NULL);

	/** Remove and delete a child (and all subchildren) from the 
	  * present body tree.
	  * 
	  * @param	child	Child to be removed.
	  */
	void removeChild (RxCompoundBody *child);

	RxCompoundBody()
	 { init();
	 }
	RxCompoundBody (RxCvxPolyhed3 *poly)
	 { init();
	   set (poly);
	 }

	~RxCompoundBody();

	const char *name() const
	 { return name_;
	 }
	const char *typeName() const
	 { return "CompoundBody";
	 }
	int setName (const char *name);

	const RxTrsf *getTRL (RxTrsf *Tr) const
	 { *Tr = TRL;
	   return &TRL;
	 }
	const RxTrsf *getTRL () const
	 { return &TRL;
	 }
	const RxTrsf *getTPL (RxTrsf *Tr) const;
	const RxTrsf *getTPL () const
	 { return TPLp;
	 }
	void setTPL (const RxTrsf *TPLnew);

	const RxMaterial *getMat (RxMaterial *Mr) const;
	const RxMaterial *getMat() const
	 { return mat;
	 }
	RxMaterial *getMat()
	 { return mat;
	 }
	void setMat (const RxMaterial *Mnew);

	RxCvxPolyhed3 *poly() const
	 { return poly_;
	 }

/*	int update();	// update transform and inertia properties  */
	int addHull();
#if 0
	void calcInertia (double m)
	int setInertia (RxInertia &M, double m, RxVec3d &cm);
	int getInertia (RxInertia *M, double *m, RxVec3d *cm);

	void setMaterial (RxMaterial *M);
	RxMaterial *getMaterial();
#endif
#if 1
	double dist (RxDistRecord3Dlist *rlist, const RxCompoundBody &body,
		const RxTrsf &T01, double distlim, int maxrecs=1, 
		RxFeatureHashTab *htab=NULL) const;
#endif
	int numChildren() const
	 { return (list.length());
	 }
	int numPolys() const;
	int numNodes() const;

	RxCompoundBody *getChild(int i)
	 { return (RxCompoundBody *)_getChild (i);
	 }
	const RxCompoundBody *getChild(int i) const
	 { return _getChild (i);
	 }
	RxCompoundBody *getParent()
	 { return (parent);
	 }
	const RxCompoundBody *getParent() const
	 { return (parent);
	 }
	
	virtual int oscan (class RxParseObj *obj);
	virtual int cprint (class RxCoutput *co,class RxPrintSpec *ps) const;

	RX_DNODE_DECLS (RxCompoundBody);
};

#endif
