#ifndef _RXCVXPOLYHED3_H_WRAP_
#define _RXCVXPOLYHED3_H_WRAP_

#include <RxSolidType3.h>
#include <RxCvxPrim3.h>
#include <RxVrtx3.h>
#include <RxEdge3.h>
#include <RxFace3.h>
#include <stdlib.h>
#include <RxDlist.h>

#include <iostream.h>
#include <RxTrsf.h>
#include <RxDistReport3.h>
#include <RxFeatureHashTab.h>
#include <RxMuxedNameTable.h>
#include <RxPrintable.h>
#include <RxRefManager.h>

// The shared handle class is pointer to a multiply-referenced
// structure, that facilitates deletion of the structure when
// the number of references falls to zero.

template<class T>
class ShareHandle
{
	T *ptr;
public:
	void set(T *p) {ptr = p; p->handleCount++;}
	ShareHandle() {ptr = NULL;}
	ShareHandle(T *p) {set(p);}
	ShareHandle(const ShareHandle &orig) {set(orig.ptr);}
	~ShareHandle() {if (ptr && --ptr->handleCount == 0) delete ptr;}
	ShareHandle &operator=(const ShareHandle &orig)
	 { if (ptr && --ptr->handleCount == 0) delete ptr;
	   set(orig.ptr);
	   return *this;
	 };
	T *operator->() {return ptr;}
	T &operator*() {return *ptr;}
	const T* operator->() const {return ptr;}
	const T& operator*() const {return *ptr;}
};

// V-Clip can use either class RxTrsf or class Se3 as the basis for pose
// specification (see mv.h).  Class RxTrsf requires fewer floating point
// operations to transform points and vectors.  However the Se3
// representation may be numerically robust (further study is needed).
// During testing, cases were found where V-Clip returned an incorrect
// disjoint result for penetrating Polyhedra when using the RxTrsf
// representation; this behavior went away when the Se3 representation
// was used.  These cases were extremely contrived and unlikely to
// occur in normal use.  The RxTrsf representation is initially
// recommended for all applications.  If problems occur, consider
// switching to the Se3 representation.  The interface is the same for
// either option, except for the type of the relative pose parameter
// Xr1r2 passed to PolyTree::vclip().

// define as 1 for RxTrsf pose representation and 0 for Se3 pose
// representation.  
#define RX_VCLIP_MATRIX_POSE 1

#if RX_VCLIP_MATRIX_POSE
typedef RxTrsf VclipPose;
#else
typedef Se3 VclipPose;
#endif

// used in V-Clip to cache transforme geometry
struct XformedGeom
{
	const class RxCvxPrim3 *feat;
	RxVecd3 q;
	RxVecd3 tail;
	RxVecd3 head;
	RxVecd3 seg;
};

class RxCvxPolyhed3 : public RxSolidType3, public RxPrintable
{
protected:
	friend class PolyTree;
	friend class ShareHandle<RxCvxPolyhed3>;
	char *name_;
	RxRefManager ref;

	int handleCount;  // number of PolyTrees pointing to thie RxCvxPolyhed3

	static class RxMuxedNameTable nameTable;
#if 0
	void nameTableCreate();
	int nameTableInsert();
	int nameTableDelete();
	static RxCvxPolyhed3 *nameTableGet(const char *name);
#endif
	static int nameAllFeatures;

	RxVrtx3List vrtxs_;
	RxEdge3List edges_;
	RxFace3List faces_;

	RxCvxPrim3 *promoteVrtx (RxVrtx3 *v, 
				   const RxVecd3 &nrm, double ang) const;
	RxCvxPrim3 *promoteEdge (RxEdge3 *e, 
				   const RxVecd3 &nrm, double ang) const;

	RxEdge3 *processEdge(RxFace3 *f, RxVrtx3 *tail, RxVrtx3 *head);

	int getVrtxIdx (const RxVrtx3 *v);

	static int vertVertTest(const RxCvxPrim3 *&v1,
			const RxCvxPrim3 *&v2,
			XformedGeom &xv1, XformedGeom &xv2,
			const VclipPose &X12, const VclipPose &X21,
			RxVecd3 &cp1, RxVecd3 &cp2, double &dist);

	static int vertFaceTest(const RxCvxPrim3 *&v, 
			const RxCvxPrim3 *&f, XformedGeom &xv,
			const VclipPose &Xvf, const RxFace3List &allFaces,
			RxVecd3 &cpv, RxVecd3 &cpf, double &dist);

	static int vertEdgeTest(const RxCvxPrim3 *&v, 
			const RxCvxPrim3 *&e, 
			XformedGeom &xv, XformedGeom &xe,
			const VclipPose &Xve, const VclipPose &Xev,
			RxVecd3 &cpv, RxVecd3 &cpe, double &dist);

	static int edgeEdgeSubtest(const RxCvxPrim3 *&e, XformedGeom &xe,
			RxVecd3 &cp);

	static int edgeEdgeTest(const RxCvxPrim3 *&e1, 
			const RxCvxPrim3 *&e2, 
			XformedGeom &xe1, XformedGeom &xe2,
			const VclipPose &X12, const VclipPose &X21,
			RxVecd3 &cp1, RxVecd3 &cp2, double &dist);

	static int edgeFaceTest(const RxCvxPrim3 *&e, 
			const RxCvxPrim3 *&f, 
			XformedGeom &xe, const VclipPose &Xef, 
			RxVecd3 &cpe, RxVecd3 &cpf, double &dist);

	static double vclip(const RxCvxPolyhed3 *const poly1,
			const RxCvxPolyhed3 *const poly2, 
		    	const VclipPose &X12, const VclipPose &X21, 
		    	const RxCvxPrim3 *&feat1, const RxCvxPrim3 *&feat2,
		    	RxVecd3 &cp1, RxVecd3 &cp2, int oneStep = 0);
#if INVENTOR
	SoShapeKit *buildInvModel() const;
#endif
	int doVrtxVrtx (RxDistRecord3 &rec, const RxTrsf &T01) const;
	int doVrtxEdge (RxDistRecord3 &rec, const RxTrsf &T01) const;
	int doVrtxFace (RxDistRecord3 &rec, const RxTrsf &T01) const;
	int doEdgeVrtx (RxDistRecord3 &rec, const RxTrsf &T01) const;
	int doEdgeEdge (RxDistRecord3 &rec, const RxTrsf &T01, int parallel=0) const;
	int doEdgeFace (RxDistRecord3 &rec, const RxTrsf &T01, int final) const;
	int doFaceVrtx (RxDistRecord3 &rec, const RxTrsf &T01) const;
	int doFaceEdge (RxDistRecord3 &rec, const RxTrsf &T01, int final) const;
	int doFaceFace (RxDistRecord3 &rec, const RxTrsf &T01) const;

	void init();
	void clear();
public:
#define RX_CVX_POLYHED3_ID	4
	static const int ID;

	void *operator new (size_t n)
	 { return (RxCvxPolyhed3*)RxRefManager::alloc(n);
	 }
	void operator delete (void *p, size_t n)
	 { RxRefManager::dealloc (((RxCvxPolyhed3*)p)->ref, p, n);
	 }
	int reference()
	 { return ref.inc();
	 }

	RxCvxPolyhed3() 
	 { init();
	 }
	~RxCvxPolyhed3()
	 { if (ref.deleteOK())
	    { clear();
	      setName (NULL);
	    }
	 }
	
	static const double EPS;

	static RxCvxPolyhed3 *findByName(const char *name)
	 { return (RxCvxPolyhed3*)nameTable.getEntry (name);
	 }
	static void setNameAllFeatures(int val)
	 { nameAllFeatures = val;
	 }
	// construction
	RxVrtx3 *addVertex(const char *name, const RxVecd3 &q);
	RxVrtx3 *addVertex(const char *name, double x, double y, double z)
	 { return addVertex (name, RxVecd3 (x, y, z));
	 }
	void RxCvxPolyhed3::addFace(const char *name, RxVrtx3 **verts, 
				    int nv, int clockwise = 0);

	int buildHull();
	int check() const;

	double dist (RxDistRecord3 *rec, const RxCvxPolyhed3 &poly2, 
		     const RxTrsf &T01, RxFeatureHashTab *htab=NULL,
		     const RxSolidType3Pair *objpair=NULL) const;
	double contacts (RxDistRecord3 *rec, const RxCvxPolyhed3 &poly2, 
		     const RxTrsf &T01, double angtol, 
		     RxFeatureHashTab *htab=NULL) const;

	// examination
//	ostream& print(ostream &os) const;
	const RxVrtx3List &vrtxs() const {return vrtxs_;}
	const RxEdge3List &edges() const {return edges_;}
	const RxFace3List &faces() const {return faces_;}

	const RxVrtx3 *findVrtxByName (const char *s) const;
	const RxEdge3 *findEdgeByName (const char *s) const;
	const RxFace3 *findFaceByName (const char *s) const;
	RxEdge3 *findEdgeByVrtx (const RxVrtx3 *tail, const RxVrtx3 *head);

	const char *name() const
	 { return name_;
	 }
	const char *typeName() const
	 { return "CvxPolyhed3";
	 }
	int setName (const char *s);

#if 0
	int scan (FILE *fp=stdin);
	int xscan (FILE *fp);
	int scan (const char *buf, int *cntp=NULL);
	istream &scan(istream &is);
	int scan (class RxCinput *ci);
	int scan (class RxParseObj *obj)
	 { return (oscan (obj));
	 }
#endif
	virtual int oscan (class RxParseObj *obj);
#if 0
	int print (FILE *fp, const char *fmt) const;
	int print(const char *fmt) const
	 { return this->print(stdout, fmt);
	 }
	int print (char *buf, const char *fmt, int n=0) const;
	ostream &print(ostream &os) const;
	int print (class RxCoutput *co, class RxPrintSpec *ps) const
	 { return cprint (co, ps);
	 }
#endif
	virtual int cprint (class RxCoutput *co, class RxPrintSpec *ps) const;
};

class RxCvxPolyhed3ListNode : public RxDnodeBase
{
    public:
	RxCvxPolyhed3 p;
	RX_DNODE_DECLS (RxCvxPolyhed3ListNode);
};

typedef RxDlist<RxCvxPolyhed3ListNode> RxCvxPolyhed3List;

inline ostream &operator<<(ostream &os, const RxCvxPolyhed3 *poly)
{
	return poly->print(os);
}

inline ostream &operator<<(ostream &os, const RxCvxPolyhed3 &poly)
{
	return poly.print(os);
}

inline istream &operator>>(istream &is, RxCvxPolyhed3 &poly)
{
	return poly.scan(is);
}

#endif
