#ifndef _RXFASTALLOC_H_WRAP_
#define _RXFASTALLOC_H_WRAP_

#include <cdefs.h>
#include <thrdatadefs.h>

// Notes on using RxFastAlloc.
// 
// Suppose we have a class Foo that we want to be able to allocate quickly.
// 
// (1) Make Foo a subclass of RxFastAlloc:
// 
// class Foo : public RxFastAlloc
// {
//	...
//
//      RX_FAST_ALLOC_DECLS_L(Foo,link)
// }	
//
// RX_FAST_ALLOC_DEFS(Foo);	// in the module for Foo
// 
// where "link" is a field within Foo that can be used to
// chain together free instances of Foo.
//
// Now to allocate Foo quickly, call either
// 
// Foo::alloc()
// 
// or 
// 
// Foo::alloc(RxFastAlloc **availp)
//
// The later is faster since it avoids the need to call 
// RxFastAlloc **Foo::getAvailPtr(). To utilize this, call
// Foo::getAvailPtr() elsewhere, store the result, and use
// it to call Foo::alloc (RxFastAlloc **availp).
// 
// To free the objects, call
//
// Foo::free (Foo *F)
// 
// or
//
// Foo::free(RxFastAlloc **availp, Foo *F)
//
//

typedef struct
{
	int init;
	class RxFastAlloc *avail;	// used when not compiled _REENTRANT
	THRDATA_KEY key;
} RxFastAllocCache;

class RxFastAlloc
{
protected:
	static void initCacheKey (RxFastAllocCache *ck);
	static RxFastAlloc **initCache (RxFastAllocCache *ck);
public:
	static inline RxFastAlloc **getAvailPtrQ (RxFastAllocCache *ck);
	static RxFastAlloc **getAvailPtr (RxFastAllocCache *ck);
};

inline RxFastAlloc **RxFastAlloc::getAvailPtrQ (RxFastAllocCache *ck)
{
	void *availp = &ck->avail;
	if (ck->init == 0)
	 { initCacheKey (ck);
	 }
	THRDATA_GET_DATA (ck->key, availp);
	if (availp == NULL)
	 { return initCache (ck);
	 }
	else
	 { return (RxFastAlloc**)availp;
	 }
}

#ifdef _REENTRANT
#define RX_FAST_ALLOC_DECLS_L(T,LINK) \
private: \
	static RxFastAllocCache cacheKey; \
public: \
	static RxFastAlloc **getAvailPtr() \
	 { return RxFastAlloc::getAvailPtrQ (&cacheKey); \
	 } \
	static T *alloc(RxFastAlloc **availp) \
	 { T *p; \
	   if ((p = (T*)*availp)) \
	    { *availp = p->LINK; \
	      return p; \
	    } \
	   else \
	    { return new T; \
	    } \
	 } \
	static T *alloc() \
	 { RxFastAlloc **availp = RxFastAlloc::getAvailPtr (&cacheKey); \
	   return alloc(availp); \
	 } \
	static void free (RxFastAlloc **availp, T *vp) \
	 { vp->LINK = (T*)*availp; \
	   *availp = vp; \
	 } \
	static void free (T *vp) \
	 { RxFastAlloc **availp = RxFastAlloc::getAvailPtr (&cacheKey); \
	   free (availp, vp); \
	 }
#else
#define RX_FAST_ALLOC_DECLS_L(T,LINK) \
private: \
	static RxFastAllocCache cacheKey; \
public: \
	static RxFastAlloc **getAvailPtr() \
	 { return &cacheKey.avail; \
	 } \
	static T *alloc() \
	 { T *p; \
	   if ((p = (T*)cacheKey.avail) == NULL) \
	    { p = new T; \
	    } \
	   else \
	    { cacheKey.avail = p->LINK; \
	    } \
	   return p; \
	 } \
	static T *alloc(RxFastAlloc **/*availp*/) \
	 { return alloc(); \
	 } \
	static void free (T *vp) \
	 { vp->LINK = (T*)cacheKey.avail; \
	   cacheKey.avail = vp; \
	 } \
	static void free (RxFastAlloc **/*availp*/, T *vp) \
	 { free (vp); \
	 }
#endif

#define RX_FAST_ALLOC_DECLS(T) \
private: \
	T *alink; \
public: \
	RX_FAST_ALLOC_DECLS_L(T,alink)

#define RX_FAST_ALLOC_DEFS(T) \
RxFastAllocCache T::cacheKey = {  0, NULL };

#endif
