#ifndef _RXFRAME_H_WRAP_
#define _RXFRAME_H_WRAP_

#include <RxTrsf.h>
#include <RxTwist.h>
#include <RxDlist.h>
#include <RxPrintable.h>
#include <RxMaterial.h>
#include <RxFastAlloc.h>
#include <RxSpatialAccel.h>
#include <RxSpatialInertia.h>
#include <RxWrench.h>

class RxFrameNode;
typedef RxDlist<RxFrameNode> RxFrameNodeList;

/** An RxFrameNode is an abstract class associated
  * with a spatial transformation frame, a possible rendering
  * material, and a list of children.
  */
class RxFrameNode : public RxDnodeBase, public RxPrintable
{
protected:
	RxTrsf T_;
	RxFrameNodeList children;
	RxFrameNode *subframe_;
	RxFrameNode *parent_;
	RxMaterial *mat;
	char *name_;
	int type_;
	RxFastAlloc **linkInfoPool;

	static class RxMuxedNameTable nameTable;

	const RxFrameNode *_getChild (int i) const;
	const RxFrameNode *_getDescendant (int i, int &j) const;
	
	// clears all fields and removes all children
	virtual void clearAll();

	int oscanChild (const char *fname, class RxParseObj *obj);
	virtual int oscanField (class RxParseObj *obj);
	virtual int checkFields (class RxParseObj *obj);
	virtual int cprintFields (class RxCoutput *co,class RxPrintSpec *ps, 
				  int init=1) const;

	void maybeOutputNewline (int &init, class RxCoutput *co, 
			int lspc) const;
	void getTreeInertiaRecurs (RxSpatialInertia &Sir, 
			const RxTrsf &TRP, const RxFrameNode *omit) const;
	void init();
public:
	static const int ID;

	/** Get the node's class type ID.
	  * 
	  * @return	class type ID
	  */
	int type() const
	 { return type_;
	 }
	/** Get a string describing the node's type.
	  * 
	  * @return	class type string
	  */
	virtual const char *typeName () const
	 { return "FrameNode";
	 }
	static RxFrameNode *findByName(const char *name)
	 { return (RxFrameNode*)nameTable.getEntry (name);
	 }
	const char *getName() const
	 { return name_;
	 }
	int setName (const char *name);

	/** Create and initialize a frame node.
	  * 
	  */
	RxFrameNode ()
	 { init();
	 }
	virtual ~RxFrameNode();

	void setTrsf (const RxTrsf &T)
	 { T_ = T;
	 }
	/** Get the spatial transformation associated with the
	  * current node.
	  * 
	  * @param	T	returns the transform
	  */
	void getTrsf (RxTrsf &T) const
	 { T = T_;
	 }
	RxTrsf &getTrsf()
	 { return T_;
	 }
	const RxTrsf &getTrsf() const
	 { return T_;
	 }

	/** Add a child to a frame node.
	  * 
	  * @param	child	new child node to be added
	  * @param	i	position within the child list,
	  *                     with 0 indicating the beginning
	  *			and -1 the end
	  */
	void addChild (RxFrameNode *child, int i=-1);

	/** Remove a child.
	  * 
	  * @param	child	child node to be removed
	  */
	void removeChild (RxFrameNode *child)
	 { children.removeAndDelete (child);
	 }
	/** Remove all children.
	  * 
	  */
	void removeAllChildren ()
	 { children.removeAndDeleteAll();
	 }
	/** Return the number of children attached to the current node.
	  *
	  * @return	number of children
	  */
	int numChildren() const
	 { return (children.length());
	 }
	/** Get the i-th child of the current node.
	  * 
	  * @param	i	number of the desired child, with 0
	  *			indicating the first child and -1
	  *			indicating the last
	  * @return	child node, or NULL if not found
	  */
	RxFrameNode *getChild (int i)
	 { return (RxFrameNode*)_getChild (i);
	 }
	const RxFrameNode *getChild (int i) const
	 { return _getChild (i);
	 }

	/** Get the i-th descendant of the current frame node.
	  * 
	  * @param	i	number of the desired descendant,
	  *			in depth-first traversal order,
	  *			with 0 indicating the first child
	  *			and -1 indicating the last descendant
	  * @return	descendant node, or NULL if not found
	  */
	RxFrameNode *getDescendant (int i)
	 { int j = 0;
	   return (RxFrameNode*)_getDescendant (i, j);
	 }
	const RxFrameNode *getDescendant (int i) const
	 { int j = 0;
	   return (RxFrameNode*)_getDescendant (i, j);
	 }
	/** Get the parent of the current frame node.
	  * 
	  * @return	parent node, or NULL if there is no parent
	  */
	RxFrameNode *parent()
	 { return (parent_);
	 }
	const RxFrameNode *parent() const
	 { return (parent_);
	 }
	const RxFrameNode *subframe() const
	 { return (subframe_);
	 }

	/** Get the rendering material associated with the
	  * current frame node.
	  * 
	  * @return	rendering material, or NULL if there is none
	  */
	const RxMaterial *getMat (RxMaterial *Mr) const;
	RxMaterial *getMat() const
	 { return mat;
	 }
	/** Set the rendering material associated with the
	  * current frame node.
	  * 
	  * @param	Mnew	Rendering material
	  */
	void setMat (const RxMaterial *Mnew);

	/** Read in the current frame node, plus descendants.
	  * 
	  * @param	pob	parse object
	  * @return	1 if the read was successful and 0 otherwise
	  */
	virtual int oscan (class RxParseObj *pob);

	/** Print out the current frame node, plus descendants.
	  * 
	  * @param	co	output descriptor
	  * @param	ps	print specification
	  * @return	number of characters output
	  */
	virtual int cprint (class RxCoutput *co,class RxPrintSpec *ps) const;

	/** Set the generalized coordinates associated with this
	  * node and/or its descendants.
	  * 
	  * @param	vec	vector of generalized coordinates, indexed in
	  *			depth-first traversal order
	  * @param	j	starting index for "vec"
	  */
	virtual int setJnts (const RxVecd &vec, int j=0);

	/** Set generalized coordinate velocities associated with this
	  * node and/or its descendants.
	  * 
	  * @param	vec	vector of generalized coordinate velocities, 
	  *			indexed in depth-first traversal order
	  * @param	j	starting index for "vec"
	  */
	virtual int setJvels (const RxVecd &vec, int j=0);

	/** Get the spatial transform of the current node with
	  * respect to a specified base node.
	  * 
	  * @param	T	returns transform value
	  * @param	base	desired base node
	  */
	void getTrsf (RxTrsf &T, const RxFrameNode *base) const;

	/** Get the spatial velocity of the current node,
	  * in current node coordinates, with
	  * respect to a specified base node.
	  *
	  * @param	vr	returns the spatial velocity
	  * @param      base    desired base node
	  * @param	v0	spatial velocity of the base node
	  */
	void getVel (RxTwist &vr, const RxFrameNode *base,
		     const RxTwist &v0 = RxTwist::ZERO) const;

	/** Get the spatial acceleration of the current node,
	  * in current node coordinates, with
	  * respect to a specified base node.             
	  *
	  * @param	ar	returns the spatial acceleration
	  * @param      base    desired base node
	  * @param	jacc	vector of generalized coordinate 
	  *			accelerations along the path leading
	  *			from the base node
	  * @param      j       starting index for "jacc"
	  * @param	a0	spatial acceleration of the base node
	  * @param	v0	spatial velocity of the base node
	  */
	void getAcc (RxSpatialAccel &ar, const RxFrameNode *base,
			const RxVecd &jacc, int j=0,
			const RxSpatialAccel &a0 = RxSpatialAccel::ZERO,
			const RxTwist &v0 = RxTwist::ZERO) const
	 { RxTwist vr;

	   getAccVel (ar, vr, base, jacc, j, a0, v0);
	 }

	/** Get the spatial acceleration and velocity of the current node,
	  * in current node coordinates, with
	  * respect to a specified base node.
	  *
	  * @param	ar	returns the spatial acceleration
	  * @param	vr	returns the spatial velocity
	  * @param      base    desired base node
	  * @param	jacc	vector of generalized coordinate
	  *			accelerations along the path leading
	  *                     from the base node
	  * @param      j       starting index for "jacc"
	  * @param	a0	spatial acceleration of the base node
	  * @param	v0	spatial velocity of the base node
	  */
	void getAccVel (RxSpatialAccel &ar, RxTwist &vr, 
			const RxFrameNode *base,
			const RxVecd &jacc, int j=0,
			const RxSpatialAccel &a0 = RxSpatialAccel::ZERO,
			const RxTwist &v0 = RxTwist::ZERO) const;

	/** Get the Jacobian of the current node, in current node
	  * coordinates, with respect to an
	  * indicated base node. Each column of the Jacobian 
	  * corresponds to a generalized coordinate velocity
	  * along the path from the base node to the current
	  * node, indexed in order starting from the base node.
	  * 
	  * @param	J	returns the Jacobian
	  * @param	base	desired base node
	  */
	void getJacobian (RxMatd &J, const RxFrameNode *base);

	/** Computes the inverse dynamics for the serial chain leading
	  * from the current node to an indicated tip node, with
	  * all other branches encountered along the way assumed to
	  * be fixed.
	  * 
	  * @param	jfor	returns the computed generalized coordinate
	  *			forces along the path leading from
	  *			the current node to the tip node
	  * @param	wr	returns the wrench acting on the current node
	  * @param      tip    desired tip node
	  * @param	jacc	vector of generalized coordinate	
	  *			accelerations along the path leading
	  *                     from the current node to the tip node
	  * @param      j       starting index for "jacc" 
	  *			and "jfor"
	  * @param	a0	spatial acceleration of the current node
	  * @param	v0	spatial velocity of the current node
	  */
	void invDynSerial (RxVecd &jfor, RxWrench &wr, 
			const RxFrameNode *tip, const RxVecd &jacc, int j=0,
			const RxSpatialAccel &a0 = RxSpatialAccel::ZERO,
			const RxTwist &v0 = RxTwist::ZERO) const;

	/** Computes the inverse dynamics for articulated tree starting
	  * at the current node.
	  * 
	  * @param	jfor	returns the computed generalized coordinate
	  *			forces, in depth-first traversal order
	  *			starting from the current node
	  * @param	wr	returns the wrench acting on the current node
	  * @param	jacc	vector of generalized coordinate	
	  *			accelerations, in depth-first traversal order
	  *                     starting from the current node
	  * @param      j       starting index for "jacc" 
	  *			and "jfor"
	  * @param	a0	spatial acceleration of the current node
	  * @param	v0	spatial velocity of the current node
	  */
	void invDyn (RxVecd &jfor, RxWrench &wr, const RxVecd &jacc, int j=0,
			const RxSpatialAccel &a0 = RxSpatialAccel::ZERO,
			const RxTwist &v0 = RxTwist::ZERO) const;

	void invDynRec (RxVecd &jfor, RxWrench &wr, const RxVecd &jacc, int j,
			const RxSpatialAccel &a0, const RxTwist &v0,
			RxSpatialInertia &Si0, RxTrsf &TPB) const;


	/** Computes the forward dynamics for the serial chain leading
	  * from the current node to an indicated tip node, with
	  * all other branches encountered along the way assumed to
	  * be fixed.
	  * 
	  * @param	jacc	returns the computed generalized coordinate
	  *			accelerations along the path leading from
	  *			the current node to the tip node
	  * @param	jfor	vector of generalized coordinate	
	  *			force along the path leading
	  *                     from the current node to the tip node
	  * @param      tip     desired tip node
	  * @param      j       starting index for "jfor" 
	  *			and "jacc"
	  * @param	a0	spatial acceleration of the current node
	  * @param	v0	spatial velocity of the current node
	  */
	void fwdDynSerial (RxVecd &jacc, const RxVecd &jfor,
			const RxFrameNode *tip, int j=0,
			const RxSpatialAccel &a0 = RxSpatialAccel::ZERO,
			const RxTwist &v0 = RxTwist::ZERO) const;

	/** Computes the forward dynamics for articulated tree starting
	  * at the current node.
	  * 
	  * @param	jacc	returns the computed generalized coordinate
	  *			accelerations, in depth-first traversal order
	  *			starting from the current node
	  * @param	jfor	vector of generalized coordinate	
	  *			forces, in depth-first traversal order
	  *                     starting from the current node
	  * @param      j       starting index for "jfor" 
	  *			and "jacc"
	  * @param	a0	spatial acceleration of the current node
	  * @param	v0	spatial velocity of the current node
	  */
	void fwdDyn (RxVecd &jacc, const RxVecd &jfor, int j=0,
			const RxSpatialAccel &a0 = RxSpatialAccel::ZERO,
			const RxTwist &v0 = RxTwist::ZERO) const;

	/** Get the effective spatial inertia of the current node
	  * and all its decsendants (with positions assumed to be
	  * fixed), in the coordinate frame indicated by T
	  * with respect to the current node.
	  * 
	  * @param	Sir	returns the spatial inertia
	  * @param	T	transform from the current frame to
	  *			the frame in which the inertia should
	  *			be computed
	  * @param	omit	optional node which should be omitted,
	  *			along with its descendants, from the
	  *			calculation
	  */
	void getTreeInertia (RxSpatialInertia &Sir,
			const RxTrsf &T = RxTrsf::IDENT, 
			const RxFrameNode *omit = NULL);

	RX_DNODE_DECLS (RxFrameNode);
};

#endif
