#ifndef _RXGENKIN_H_WRAP_
#define _RXGENKIN_H_WRAP_

#include <RxVecd.h>
#include <RxMatdLu.h>
#include <RxMatdSvd.h>
#include <RxTwist.h>
#include <RxWrench.h>
#include <RxTrsf.h>

typedef struct {
	char type;
	double d;
	double a;
	double alpha;
	double theta;
	double sA;
	double cA;
	double sT;
	double cT;
} DHParams;

// return codes for invK

#define RX_GENKIN_UNIMPLEMENTED		0x80000000
#define RX_GENKIN_INDETERMINATE		0x40000000
#define RX_GENKIN_BOUNDARY_LIMIT	0x000000ff
#define RX_GENKIN_SINGULAR		0x0000ff00

#define RX_GENKIN_DEFAULT_TOL		(1e-9)

class RxGenKin
{
protected:
	int numj;
	DHParams *dhpSet;
	RxMatd *jacob;
	RxMatd *JJ;
	RxMatdLu *lu;
	RxMatdSvd *svd;
	double lambda;
	double lrange;
	double tol;
	double condlimit;
#define RX_GENKIN_JACOBIAN_VALID	0x01
#define RX_GENKIN_JACOBINV_VALID	0x02
#define RX_GENKIN_JACOBSVD_VALID	0x04
#define RX_GENKIN_NEED_SVD		0x08
	int flags;

	void init();
	void buildJacobian();
	void invertJacobian();

public:
	// assignment

	int set (const char *fileName);
	int set (const DHParams *params, int n);
	int set (const RxGenKin &kin);
	RxGenKin& operator=(const RxGenKin &kin)
	 { set (kin);
	   return (*this);
	 }
	int get (DHParams *params);

	// assignment of other properties

	virtual int setJnts (const RxVecd &jnts);
	virtual int getJnts (RxVecd *jnts);

	void setTol (double eps)
	 { tol = eps;
	 }
	double getTol ()
	 { return (tol);
	 }

	void setCondlimit (double lim)
	 { condlimit = lim;
	 }
	double getCondlimit ()
	 { return (condlimit);
	 }

	void setDamping (double lam, double rng)
	 { lambda = lam;
	   lrange = rng;
	 }
	void getDamping (double *lamp, double *rngp)
	 { *lamp = lambda;
	   *rngp = lrange;
	 }

	// constructors 

	RxGenKin ()
	 { init();
	 }
	RxGenKin (const char *fileName)
	 { init();
	   set (fileName);
	 }
	RxGenKin (const DHParams *params, int n)
	 { init();
	   set (params, n);
	 }
	RxGenKin (const RxGenKin &kin)
	 { init();
	   set (kin);
	 }

	~RxGenKin ();

	// query

	int numJnts()
	 { return numj;
	 }

	// forward kinematics

	virtual int fwdK (RxTrsf *T, int *conf);
	virtual int fwdK (RxTrsf *T, int *conf, const RxVecd &jnts)
	 { setJnts (jnts);
	   return (fwdK (T, conf));
	 }

	virtual RxTrsf& A (RxTrsf *mr, int j);
	virtual RxTrsf& A (RxTrsf *mr, int j, double jval);
	RxTrsf A (int j, double jval)
	 { RxTrsf mr;
	   A (&mr, j, jval);
	   return (mr);
	 }
	RxTrsf A (int j)
	 { RxTrsf mr;
	   A (&mr, j);
	   return (mr);
	 }

	virtual RxMatd& Jset (RxMatd *J);
	virtual RxMatd& Jset (RxMatd *J, const RxVecd jnts)
	 { setJnts(jnts);
	   return (Jset (J));
	 }
	virtual RxTwist& Jmul (RxTwist *tw, const RxVecd &dj);
	virtual RxTwist& Jmul (RxTwist *tw, const RxVecd &dj, 
			       const RxVecd jnts)
	 { setJnts(jnts);
	   return (Jmul (tw, dj));
	 }
	virtual RxVecd& JTmul (RxVecd *tor, const RxWrench &wr);
	virtual RxVecd& JTmul (RxVecd *tor, const RxWrench &wr, 
			       const RxVecd jnts)
	 { setJnts(jnts);
	   return (JTmul (tor, wr));
	 }

	// inverse kinematics

	virtual int invK (RxVecd *jnts, const RxTrsf &T, int conf, 
			  const RxVecd *jref=NULL, int update=0);

	virtual int Jsolve (RxVecd *dj, const RxTwist &tw);
	virtual int Jsolve (RxVecd *dj, const RxTwist &tw,
			    const RxVecd jnts)
	 { setJnts (jnts);
	   return (Jsolve (dj, tw));
	 }
	virtual int JTsolve (RxWrench *wr, const RxVecd &tor);
	virtual int JTsolve (RxWrench *wr, const RxVecd &tor,
			     const RxVecd jnts)
	 { setJnts (jnts);
	   return (JTsolve (wr, tor));
	 }
};

#endif
