#ifndef _RXSPATIALINERTIA_H_WRAP_
#define _RXSPATIALINERTIA_H_WRAP_

#include <RxMatdBase.h>
#include <RxMatd3x3.h>
#include <RxSpatialMat.h>
#include <RxSpatialVec.h>
#include <RxTrsf.h>
#include <mathx.h>
#include <stdarg.h>
#include <cdefs.h>

	// Here is what the matrix looks like:
	//	      0      1      2      3      4      5
	//	  ---------------------------------------------
	//	0 [   0     mpc.z -mpc.y   m      0      0    ]
	//	1 [ -mpc.z   0     mpc.x   0      m      0    ]
	//	2 [  mpc.y -mpc.x   0      0      0      m    ]
	//	3 [  I.00   I.01   I.02    0    -mpc.z  mpc.y ]
	//	4 [  I.10   I.11   I.12   mpc.z   0    -mpc.x ]
	//	5 [  I.20   I.21   I.22  -mpc.y  mpc.x   0    ]     


class RxSpatialInertia : public RxSpatialMat
{
	RxVecd3 com_;

    private:
	void doCopy (const RxSpatialInertia &m1)
	 { RxSpatialMat::doCopy (m1);
	   com_.set (m1.com_);
	 }
	void doInit ();
	void doZeros ();
	void updateMass (double m)
	 { double mcx, mcy, mcz;

	   B_.xx() = m;
	   B_.yy() = m;
	   B_.zz() = m;
	   mcx = m*com_.x();
	   mcy = m*com_.y();
	   mcz = m*com_.z();
	   A_.xy() =  mcz;
	   A_.xz() = -mcy;
	   A_.yx() = -mcz;
	   A_.yz() =  mcx;
	   A_.zx() =  mcy;
	   A_.zy() = -mcx;

	   D_.xy() = -mcz;
	   D_.xz() =  mcy;
	   D_.yx() =  mcz;
	   D_.yz() = -mcx;
	   D_.zx() = -mcy;
	   D_.zy() =  mcx;
	 }
    public:

	RxMatd3x3 &I()
	 { return C_;
	 }
	const RxMatd3x3 &I() const
	 { return C_;
	 }
	double getMass() const
	 { return B_.xx();
	 }
	void setMass (double m);
	void setCom (const RxVecd3 &com);
	double getCom (RxVecd3 &com)
	 { com.set (com_);
	 }
	const RxVecd3 &getCom() const
	 { return com_;
	 }

	// assignment

	RxSpatialInertia& set (const RxSpatialInertia& in1)
	 { doCopy (in1);
	   return *this;
	 }
	RxSpatialInertia& operator=(const RxSpatialInertia& m2)
	 { return (set (m2));
	 }
	RxSpatialInertia& set (const RxMatd3x3& I, double m);
	RxSpatialInertia& set (const RxMatd3x3& I, double m, 
			       const RxVecd3 &com);

	// constructors

	RxSpatialInertia ()
	 { doInit();
	 }
	RxSpatialInertia (const RxSpatialInertia &m2)
         { doZeros();
	   doCopy (m2);
	 }
	RxSpatialInertia (const RxMatd3x3& I, double m)
	 { doZeros();
	   set (I, m);
	 }
	RxSpatialInertia (const RxMatd3x3& I, double m,
			  const RxVecd3 &com)
	 { doZeros();
	   set (I, m, com);
	 }

	// T <- T,T operations
	
	RxSpatialInertia& add (const RxSpatialInertia &m1, const RxSpatialInertia &m2);
	RxSpatialInertia& add (const RxSpatialInertia &m2)
	 { return (add (*this, m2));
	 }
	RxSpatialInertia& operator+=(const RxSpatialInertia &m2)
	 { return (add (*this, m2));
	 }
	friend RxSpatialInertia operator+(const RxSpatialInertia &m1, const RxSpatialInertia &m2);

	RxSpatialInertia& sub (const RxSpatialInertia &m1, const RxSpatialInertia &m2);
	RxSpatialInertia& sub (const RxSpatialInertia &m2)
	 { return (sub (*this, m2));
	 }
	RxSpatialInertia& operator-=(const RxSpatialInertia &m2)
	 { return (sub (*this, m2));
	 }
	friend RxSpatialInertia operator-(const RxSpatialInertia &m1, const RxSpatialInertia &m2);

	// T <- T,O operations

	RxSpatialInertia& scale (double a, const RxSpatialInertia &m1);
	RxSpatialInertia& scale (double a);
	RxSpatialInertia& operator *= (double a)
	 { return (scale (a));
	 }
	RxSpatialInertia& operator /= (double a)
	 { return (scale (1/a));
	 }
	friend RxSpatialInertia operator*(const RxSpatialInertia &m1, double a);
	friend RxSpatialInertia operator*(double a, const RxSpatialInertia &m1);
	friend RxSpatialInertia operator/(const RxSpatialInertia &m1, double a);

	RxSpatialInertia& xform (const RxSpatialInertia &v1, const RxRot& R);
	RxSpatialInertia& xform (const RxRot& R)
	 { return (xform (*this, R));
	 }
	RxSpatialInertia& Ixform (const RxSpatialInertia &v1, const RxRot& R);
	RxSpatialInertia& Ixform (const RxRot& R)
	 { return (Ixform (*this, R));
	 }
	RxSpatialInertia& xform (const RxSpatialInertia &v1, const RxTrsf& T);
	RxSpatialInertia& xform (const RxTrsf& T)
	 { return (xform (*this, T));
	 }
	RxSpatialInertia& Ixform (const RxSpatialInertia &v1, const RxTrsf& T);
	RxSpatialInertia& Ixform (const RxTrsf& T)
	 { return (Ixform (*this, T));
	 }

	// O <- T,O operations

	RxSpatialVec &Vmul (RxSpatialVec *vr, const RxSpatialVec &v1) const;
	RxSpatialVec &mulV (RxSpatialVec *vr, const RxSpatialVec &v1) const;
	friend RxSpatialVec operator*(const RxSpatialInertia &m1, const RxSpatialVec &v1);
	friend RxSpatialVec operator*(const RxSpatialVec &v1, const RxSpatialInertia &m1);

	// T <- T operations

	RxSpatialInertia &neg (const RxSpatialInertia &m1);
	RxSpatialInertia &neg()
	 { return (neg(*this));
	 }
	friend RxSpatialInertia operator-(const RxSpatialInertia &m1);

	// T <- O operations

	RxSpatialInertia& setZero()
	 { RxSpatialMat::setZero();
	   com_.setZero();
	 }
	RxSpatialInertia& setUnit();

	int oscan (class RxParseObj *obj);
	int cprint (class RxCoutput *co, class RxPrintSpec *fs) const;

	// constants

	static const RxSpatialInertia UNIT;
	static const RxSpatialInertia ZERO;
};

inline RxSpatialVec operator*(const RxSpatialInertia &m1, const RxSpatialVec &v1)
{
	RxSpatialVec vr;
	return (m1.mulV(&vr,v1));
}

inline RxSpatialVec operator*(const RxSpatialVec &v1, const RxSpatialInertia &m1)
{
	RxSpatialVec vr;
	return (m1.Vmul(&vr,v1));
}

inline RxSpatialInertia operator-(const RxSpatialInertia &m1)
{
	RxSpatialInertia mr;
	return (mr.neg(m1));
}

#endif
