/*---------McGill-Computer-Vision-and-Robotics-Laboratory--C-Module---------
 *		
 * NAME:	command_match
 *		
 * USAGE:	index = command_match (prompt, keywords);
 *		int index;
 *		char *prompt;
 *		char **keywords;
 *
 * DESCRIPTION: A sort of bare bones keyword matcher. You provide a string
 *		'prompt' and an array of 'keywords' and the routine returns
 *		the array index of the keyword that was matched. The array
 *		MUST be terminated by the string "EOA", or the matcher will
 *		crash furiously. This makes it easier to modify the size of
 *		the keyword list without having to worry about it's size.
 * 
 * FILES:	/u1/robot/util/keyword.c	: source
 *		/u1/robot/util/useful.a		: archive
 *
 * DIAGNOSTICS: Returns -1 on end of file and -2 if an error occurs.
 * 
 * NOTES:	The routine sets the terminal into CBREAK and NOECHO;
 *		a sort of rude effort has been made to unset these conditions
 *		upon reciept of a signal, but the results may be a bit
 *		unpredicatble if the signal does not kill the process.
 *
 *		A high level description of the implementation is given
 *		below.
 *
 * AUTHOR/DATE: John Lloyd
 *
 * UPDATES:
 *
 *-------------McGill-Computer-Vision-and-Robotics-Laboratory--------------*/

/*---------------------------------------------------------------------------
 *
 * IMPLEMENTATION:
 *	
 *	Initialize.
 *
 *      for each input charcter, switch on the character function:
 *	   normal character:
 *	      if a lookahead has been possible
 *	         match the character to the first lookahead character;
 *		    echo charcacter and increase index
 *	         else
 *		    ring the bell
 *	      else
 *	         for all active keywords
 *		    deactivate keywords which don't match the character
 *	         if all keywords have been deactivated,
 *		    then the character is bogus;
 *		    reactivate those keywords and ring the bell;
 *	         else
 *		    remember (stack) the lookahead point;
 *		    lookahead as far as possible
 *	   DELETE:
 *	      if the lookahead stack is not empty,
 *	         pop the present lookahead stack
 *	         undelete all entries at that lookahead point, and
 *	         erase back to that index
 *	   HELP:
 *	      print a list of the keywords
 *	   EXPAND:
 *	      if lookahead,
 *	         print out the lookahead and increase index accordingly
 *	      else
 *	         ring the bell
 *	   ENTER:
 *	      if number of active keywords is one, or there is an
 *		active keyword at full length,
 *	         return with that keyword
 *	      else
 *	         ring the bell
 *	   EOF:
 *	      return (-1)
 *
 *--------------------------------------------------------------------------*/

#include <sgtty.h>
#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
 
#define NORMAL		1
#define EXPAND		2
#define HELP		3
#define ENTER		4
#define DELETE		5
#define MAXKEYS 	256
#define MAXSTRLEN 	256

#define TRUE		1
#define FALSE		0

#define BELL		'\7'

typedef struct _keyword_record {
  char *string;
  int length;
  int remove; } KEYWORD_RECORD;

extern int errno;
static int user_fd = -1;
static struct sgttyb dev_chars, old_chars;

static int (*old_inthandler[NSIG])();

static inthandler (sig)
int sig;
{
 ioctl (user_fd, TIOCSETP, &old_chars);
 signal (sig, old_inthandler[sig]);
 kill (0, sig);
}

command_match_open ()
{
 if (user_fd == -1)
  { if ((user_fd = open ("/dev/tty", 2)) < 0)
       return (-1);
    else 
       return (0);
  }
}

command_match (prompt, keywords)
char *prompt;
char **keywords;
{
 register int i, j;

 char input, match;
 char char_status;
 char lahead_buffer [MAXSTRLEN];
 int lahead = 0;
 int index = 0;
 int lahead_stack [MAXSTRLEN];
 int lahead_sp = -1;
 KEYWORD_RECORD keys [MAXKEYS];
 char *print_list [MAXKEYS];

#define PAGE_WIDTH 78
 
 int max_key_length = 0;
 int num_columns;
 char printf_ctrl [10];

 int keylist_size;
 int num_entries;
 int result;
 int noblock;
 
 /* find the size of the keylist */
 
 for (i=0; (strcmp (keywords[i], "EOA") != 0) && i<MAXKEYS; i++)
    ;
 if (i == MAXKEYS)
  { fprintf (stderr, "Error, number of keywords exceeds %d\n", MAXKEYS);
    return (-2);
  }
 keylist_size = i++;
 num_entries = keylist_size;

 for (i=0; i<keylist_size; i++)
  { keys[i].string = keywords[i];
    keys[i].length = strlen (keywords [i]);
    if (keys[i].length > max_key_length)
       max_key_length = keys[i].length;
    keys[i].remove = -1;
    print_list[i] = keywords[i];
  }

 num_columns = PAGE_WIDTH / (max_key_length + 2);
 sprintf (printf_ctrl, "%%-%ds", (max_key_length + 2));

 for (i=0; i<(keylist_size-1); i++)
  { for (j=i+1; j<keylist_size; j++)
     { if (strcmp (print_list[i], print_list[j]) > 0)
	{ char *temp;
	  
	  temp = print_list[i];
	  print_list[i] = print_list[j];
	  print_list[j] = temp;
	}
     }
  }

 /* open the tty file only if we have to */

 if (user_fd == -1)
  { if ((user_fd = open ("/dev/tty", 2)) < 0)
       return (-2);
  }

 ioctl (user_fd, TIOCGETP, &dev_chars);
 old_chars = dev_chars;
 dev_chars.sg_flags |= CBREAK;
 dev_chars.sg_flags &= ~ECHO;
 ioctl (user_fd, TIOCSETP, &dev_chars);

 /* take care of signals  (cough) */

 for (i=0; i<NSIG; i++)
  { 
    old_inthandler[i] = signal (i, inthandler);
  }

 noblock = TRUE;
 ioctl (fileno(stdin), FIONBIO, &noblock);

 while (TRUE)
  { input = getchar();
    if (input == EOF && errno == EWOULDBLOCK)
     { clearerr (stdin);
       break;
     }
    else if (! (isspace (input)))
     { ungetc (input, stdin);
       break;
     }
  }

 noblock = FALSE;
 ioctl (fileno(stdin), FIONBIO, &noblock);
 
 printf ("%s", prompt);
 while (1)
  { 
    fflush (stdout);
    input = getchar();
    switch (input)
     { case '?':
	  char_status = HELP;
	  break;
       case ' ':
	  char_status = EXPAND;
	  break;
       case '\n':
	  char_status = ENTER;
	  break;
       case '\177':
       case '\b':
	  char_status = DELETE;
	  break;
       case EOF:
	  char_status = EOF;
	  break;
       default:
	  char_status = NORMAL;
	  break;
     }


    switch (char_status)
     { case NORMAL:
	  if (lahead > index)
	   { if (lahead_buffer [index] == input)
	      { putchar (input);
		index++;
	      }
	     else
	      { putchar (BELL);
	      }
	   }
	  else
	   { for (i=0; i<keylist_size; i++)
	      { if (keys[i].remove == -1)
		 { if (index < keys[i].length)
	            { if (keys[i].string[index] != input)
		       { keys[i].remove = index;
		         --num_entries;
		       }
		    }
		   else
		    { keys[i].remove = index;
		      --num_entries;
		    }
		 }
	      }
	     if (num_entries == 0)
	      { for (i=0; i<keylist_size; i++)
		 { 
		   if (keys[i].remove == index)
		    { keys[i].remove = -1;
		      num_entries++;
		    }
		 }
		putchar (BELL);
	      }
	     else
	      { putchar (input);
	        lahead_stack [++lahead_sp] = index;
		lahead_buffer [lahead] = input;
		j = lahead;
		lahead++;
		index++;
		do
		 { j++;
		   match = 0;
		   for (i=0; i<keylist_size; i++)
		    { if (keys[i].remove == -1)
		       { if (j >= keys[i].length)
			  { match = 0;
			    break;
			  }
			 else
			  { if (match == 0)
		    	     { match = keys[i].string[j];
			     }
			    else
		             { if (keys[i].string[j] != match)
				{ match = 0;
				  break;
				}
			     }
			  }
		       }
		    }
		   if (match != 0)
		    { lahead_buffer [lahead++] = match;
		    }
		 }
		while (match != 0);
	      }
	   }
	  break;
       case DELETE:
	  if (lahead_sp > -1)
	   { lahead = lahead_stack [lahead_sp--];
	     for (i=0; i<keylist_size; i++)
	      { if (keys[i].remove == lahead)
		 { keys[i].remove = -1;
		   num_entries++;
		 }
	      }
	     for (i=0; i<(index-lahead); i++)
	      { printf ("\b \b");
	      }
	     index = lahead;
	   }
	  break;
       case EXPAND:
	  if (lahead > index)
	   { for (i=index; i<lahead; i++)
	        putchar (lahead_buffer[i]);
	     index = lahead;
	   }
	  else
	   { putchar (BELL);
	   }
	  break;
       case ENTER:
	  result = -1;
	  if (num_entries == 1)
	   { for (i=0; i<keylist_size; i++)
	      { if (keys[i].remove == -1)
		 { result = i;
		   break;
		 }
	      }
	   }
	  else
	   { for (i=0; i<keylist_size; i++)
	      { if ((keys[i].remove == -1) && (keys[i].length == index))
		 { result = i;
		   break;
		 }
	      }
	   }
	  if (result != -1)
	   { for (i=index; i<keys[result].length; i++)
	        putchar (keys[result].string[i]);
             ioctl (user_fd, TIOCSETP, &old_chars);
	     for (i=0; i<NSIG; i++)
	      { signal (i, old_inthandler[i]);
	      }
	     return (result);
	   }
	  else
	   { putchar (BELL);
	   }
	  break;
       case HELP:
	  putchar ('\n');
	  for (i=0; i<keylist_size;)
	   { for (j=0; (j<num_columns)&& (i<keylist_size); j++)
	      { printf (printf_ctrl, print_list[i++]);
	      }
	     putchar ('\n');
	   }
	  printf ("%s", prompt);
	  for (i=0; i<index; i++)
	     putchar (lahead_buffer[i]);
	  break;
       case EOF:
	  ioctl (user_fd, TIOCSETP, &old_chars);
	  for (i=0; i<NSIG; i++)
	   { signal (i, old_inthandler[i]);
	   }
	  return (-1);
     }
  }
}
