#include <stdiox.h>
#include <string.h>
#include <stringx.h>
#include <stdlib.h>
#include <ctype.h>
#include <mutexdefs.h>

DECLARE_STATIC_MUTEX(mux);

int convFieldWidth (const char *conv)
{
	static const char *flags = "-+ #0";
	const char *cp;
	int width = 0;

	cp = conv+1;
	if (*cp == '%')
	 { return (1);
	 }
	while (strchr (flags, *cp))
	 { cp++;
	 }
	while (isdigit(*cp))
	 { width = width*10 + (*cp++ -'0');
	 }	
	return (width);
}
	
int fmtSpecAlloc (FmtSpec *spec, const char *fmt)
{
	char *convBegin;
	char *convEnd;
	int stat;

	spec->prefix = NULL;
	spec->conv = NULL;
	spec->suffix = NULL;
	spec->width = 0;

	if (fmt == NULL)
	 { return 0;
	 }
	stat = fmtindex (&convBegin, &convEnd, fmt);
	if (!convBegin)
	 { spec->prefix = strdup(fmt);
	   return (0);
	 }
	else
	 { if (fmt > convBegin)
	    { spec->prefix = strndup(fmt, fmt-convBegin);
	    }
	   spec->conv = strndup (convBegin, convEnd-convBegin+1);
	   if (convEnd[1] != '\0')
	    { spec->suffix = strdup (convEnd+1);
	    }
	   spec->width = convFieldWidth(spec->conv);
	   return (stat);
	 }
}

void fmtSpecFree (FmtSpec *spec)
{
	if (spec->prefix)
	 { free (spec->prefix);
	 }
	if (spec->conv)
	 { free (spec->conv);
	 }
	if (spec->suffix)
	 { free (spec->suffix);
	 }
}

typedef struct streamData {
	struct streamData *next;
	int lineno;
	FILE *fp;
} StreamData;

static StreamData *dataList;

static StreamData *getData (FILE *fp)
{
	StreamData *data;

	MUTEX_LOCK(&mux);
	for (data=dataList; data; data=data->next)
	 { if (data->fp == fp)
	    { MUTEX_UNLOCK(&mux);
	      return (data);
	    }
	 }
	data = (StreamData*)malloc(sizeof(StreamData));
	if (data)
	 { data->next = dataList;
	   data->fp = fp;
	   dataList = data;
	   data->lineno = 1;
	 }
	MUTEX_UNLOCK(&mux);
	return (data);
}

static StreamData *findData (FILE *fp)
{
	StreamData *data;

	MUTEX_LOCK(&mux);
	for (data=dataList; data; data=data->next)
	 { if (data->fp == fp)
	    { MUTEX_UNLOCK(&mux);
	      return (data);
	    }
	 }
	MUTEX_UNLOCK(&mux);
	return (NULL);
}

static void *freeData (FILE *fp)
{
	StreamData *data, *last;

	last = NULL;
	MUTEX_LOCK(&mux);
	for (data=dataList; data; data=data->next)
	 { if (data->fp == fp)
	    { if (last)
	       { last->next = data->next;
	       }
	      else
	       { dataList = data->next;
	       }
	      MUTEX_UNLOCK(&mux);
	      free (data);
	      return;
	    }
	   last = data;
	 }
	MUTEX_UNLOCK(&mux);
}

static void eatWhiteSpace (StreamData *data)
{
	int c;
	do
	 { c = getc(data->fp);
	   if (c == '#')
	    { do
	       { c = getc(data->fp);
	       }
	      while (c != '\n' && c != EOF);
	    }
	   if (c == '\n') data->lineno++;
	 }
	while (isspace (c));
	ungetc (c, data->fp);
}

int xscanf (FILE *fp, const char *fmt, void *ptr)
{
	StreamData *data;

	if ((data = getData (fp)) == NULL)
	 { return -1;
	 }
	eatWhiteSpace (data);
	return (fscanf (data->fp, fmt, ptr));
}

int xgetc (FILE *fp)
{
	StreamData *data;
	int c;

	if ((data = getData (fp)) == NULL)
	 { return -1;
	 }
	c = getc(fp);
	if (c == '#')
	 { do
	    { c = getc(fp);
	    }
	   while (c != '\n');
	 }
	if (c == '\n')
	 { data->lineno++;
	 }
	return c;
}

int xungetc (int c, FILE *fp)
{
	StreamData *data;
	int stat;

	if ((data = getData (fp)) == NULL)
	 { return -1;
	 }
	stat = ungetc (c, fp);
	if (stat != EOF && c == '\n')
	 { data->lineno--;
	 }
	return stat;
}

void xscanfBegin (FILE *fp)
{
	StreamData *data;

	if ((data = getData (fp)))
	 { data->lineno = 1;
	 }
}

void xscanfEnd (FILE *fp)
{
	freeData (fp);
}

int xscanfLine (FILE *fp)
{
	StreamData *data;

	if ((data = findData (fp)))
	 { return (data->lineno);
	 }
	else
	 { return (-1);
	 }
}
