#include <RxCio.h>
#include <stdiox.h>
#include <snprintf.h>
#include <ctype.h>

void RxCoutput::setStream (FILE *fp)
{
	u.fp = fp;
	type_ = OUT_FILE;
	nchars_ = 0;
}

void RxCoutput::setStream (char *buf, int n)
{
	u.buf = buf;
	nchars_ = 0;
	if (n <= 0)
	 { type_ = STRING;
	 }
	else
	 { type_ = STRING_N;
	   if (n == 1)
	    { buf[0] = '\0';	// don't write anything else!
	    }
	   nmax = n;
	 }
}

void RxCoutput::setStream (ostream *osp)
{
	u.osp = osp;
	type_ = OSTREAM;
	nchars_ = 0;
}

void *RxCoutput::getStream () const
{
	switch (type_)
	 { case OUT_FILE:
	    { return u.fp;
	    }
	   case STRING:
	   case STRING_N:
	    { return u.buf;
	    }
	   case OSTREAM:
	    { return u.osp;
	    }
	   default:
	    { return NULL;
	    }
	 }
}

#define IOS_BUF_MAX	80	/* this is just a guess */

#define OUTPUT_FXN(N,T) \
int RxCoutput::N (const char *fmt, T val) \
{ \
	int n; \
	if (nchars_ == -1) \
	 { return -1; \
	 } \
	if (type_ == OUT_FILE) \
	 { n = fprintf (u.fp, fmt, val); \
	 } \
	else if (type_ == STRING) \
	 { n = sprintf (u.buf+nchars_, fmt, val); \
	 } \
	else if (type_ == STRING_N) \
	 { if (nchars_ >= nmax-1) \
	    { return -1; \
	    } \
	   n = snprintf (u.buf+nchars_, nmax-nchars_, fmt, val); \
	   if (n >= nmax-nchars_) \
	    { n = -1; \
	    } \
	   if (n == -1) \
	    { nchars_ = -1; \
	      return -1; \
	    } \
	 } \
	else if (type_ == OSTREAM) \
	 { int w0, w1; \
	   w0 = u.osp->rdbuf()->out_waiting(); \
	   if (w0 > IOS_BUF_MAX) \
	    { u.osp->flush(); \
	      w0 = u.osp->rdbuf()->out_waiting(); \
	    } \
	   *u.osp << val; \
	   w1 = u.osp->rdbuf()->out_waiting(); \
	   n = (u.osp->good() ? w1-w0 : -1); \
	 } \
	if (n < 0) \
	 { nchars_ = -1; \
	 } \
	else \
	 { nchars_ += n; \
	 } \
	return n; \
}

OUTPUT_FXN (output_double, double)
OUTPUT_FXN (output_float, float)
OUTPUT_FXN (output_int, int)
OUTPUT_FXN (output_str, const char *)

int RxCoutput::output_str (const char *s)
{
	if (s == NULL)
	 { return 0;
	 }
	else if (type_ != OSTREAM)
	 { return output_str ("%s", s);
	 }
	else
	 { *u.osp << s;
	   if (u.osp->good())
	    { int n;
	      n = strlen(s);
	      nchars_ += n;
	      return n;
	    }
	   else
	    { nchars_ = -1;
	      return -1;
	    }
	 }
}

int RxCoutput::output_char (char c, int n)
{
	int k;

	if (nchars_ == -1)
	 { return -1;
	 }
	for (k=0; k<n; k++)
	 { if (type_ == OUT_FILE)
	    { if (putc (c, u.fp) ==EOF)
	       { nchars_ = -1;
		 return -1;
	       }
	    }
	   else if (type_ == STRING)
	    { u.buf[nchars_+k] = c;
	    }
	   else if (type_ == STRING_N)
	    { /* Number of characters presently written is nchars_+k.
		 We want to write another character. That means that
		 we must have nchars_+k < nmax-1, or we get an overflow */
	      if (nchars_+k >= nmax-1)
	       { nchars_ = -1;
		 return -1;
	       }
	      u.buf[nchars_+k] = c;
	    }
	   else if (type_ == OSTREAM)
	    { u.osp->put(c);
	      if (!u.osp->good())
	       { nchars_ = -1;
		 return -1;
	       }
	    }
	 }
	if (type_ == STRING || type_ == STRING_N)
	 { u.buf[nchars_+n] = '\0';
	 }
	nchars_ += n;
	return n;
}

#define TAB_SIZE	8

int RxCoutput::output_newline (int leadspacing)
{
	int ntabs;
	int nspaces;
	int i;

	ntabs = leadspacing/TAB_SIZE;
	nspaces = leadspacing%TAB_SIZE;
	output_char ('\n');
	for (i=0; i<ntabs; i++)
	 { output_char ('\t');
	 }
	for (i=0; i<nspaces; i++)
	 { output_char (' ');
	 }
	return 1+nspaces+ntabs;	
}

int RxCoutput::output_doubleArray (const char *fmt0, const char *fmt1,
			const char *fmt2, double *vec, int len)
{
	int i, n = 0;

	if (len <= 0)	// sanity check
	 { return 0;
	 }
	n += output_double (fmt0, vec[0]);
	for (i=1; i<len; i++)
	 { n += output_str (fmt1 ? fmt1 : " ");
	   n += output_double (fmt2 ? fmt2 : fmt0, vec[i]);
	 }
	return n;
}

void RxCinput::init()
{
	htab = NULL;
}

void RxCinput::setStream (FILE *fp, int useXscan)
{
	u.fp = fp;
	type_ = (useXscan ? IN_FILE_X : IN_FILE);
	status_ = 1;
	cntp_ = NULL;
}

void RxCinput::setStream (const char *buf, int *cntp)
{
	u.buf = buf;
	type_ = STRING;
	status_ = 1;
	nchars_ = 0;
	cntp_ = cntp;
}

void RxCinput::setStream (istream *isp)
{
	u.isp = isp;
	type_ = ISTREAM;
	nchars_ = 0;
}

void *RxCinput::getStream () const
{
	switch (type_)
	 { case IN_FILE:
	    { return u.fp;
	    }
	   case STRING:
	    { return (void*)u.buf;
	    }
	   case ISTREAM:
	    { return u.isp;
	    }
	   default:
	    { return NULL;
	    }
	 }
}

#define INPUT_FXN(N,T,FMT) \
int RxCinput::N (T *val) \
{ \
	int k; \
	if (status_ != 1) \
	 { return status_; \
	 } \
	if (type_ == IN_FILE) \
	 { k = fscanf (u.fp, FMT, val); \
	 } \
	else if (type_ == IN_FILE_X) \
	 { k = xscanf (u.fp, FMT, val); \
	 } \
	else if (type_ == STRING) \
	 { char fmt[16]; \
	   int n; \
	   strcpy (fmt, FMT); \
	   strcat (fmt, "%n"); \
	   k = sscanf (u.buf+nchars_, fmt, val, &n); \
	   if (k == 1) \
	    { nchars_ += n; \
	    } \
	 } \
	else if (type_ == ISTREAM) \
	 { *u.isp >> *val; \
	   if (u.isp->eof()) \
	    { k = EOF; \
	    } \
	   else if (u.isp->good()) \
	    { k = 1; \
	    } \
	   else \
	    { k = 0; \
	    } \
	 } \
	if (k != 1) \
	 { status_ = k; \
	 } \
	return status_; \
}

INPUT_FXN (input_double, double, "%lf")
INPUT_FXN (input_float, float, "%f")
INPUT_FXN (input_int, int, "%d")

int RxCinput::getChar ()
{
	int c;

	if (type_ == IN_FILE)
	 { c = getc (u.fp);
	 }
	else if (type_ == IN_FILE_X)
	 { c = xgetc (u.fp);
	 }
	else if (type_ == STRING)
	 { c = *(u.buf+nchars_);
	   if (c == '\0')
	    { c = EOF;
	    }
	   else
	    { nchars_++;
	    }
	 }
	else if (type_ == ISTREAM)
	 { c = u.isp->get();
	 }
	return c;
}
	
int RxCinput::ungetChar (int c)
{
	if (c == EOF)
	 { return EOF;
	 }
	if (type_ == IN_FILE)
	 { return ungetc (c, u.fp);
	 }
	else if (type_ == IN_FILE_X)
	 { return xungetc (c, u.fp);
	 }
	else if (type_ == STRING)
	 { if (nchars_ > 0)
	    { nchars_--;
//	      *(u.buf+nchars_) = c;
	      if (c != *(u.buf+nchars_))
	       { fprintf (stderr,
"RxCinput::ungetChar: chaning c=%c breaks in the case of strings\n");
		 exit (1);
	       }
	      return c;
	    }
	   else
	    { return EOF;
	    }
	 }
	else if (type_ == ISTREAM)
	 { u.isp->putback(c);
	   return (u.isp->good() ? c : EOF);
	 }
	else
	 { fprintf (stderr,
"RxCinput::ungetChar: internal error: bogus type %d\n", type_);
	   exit (1);
	   return -1;
	 }
}
	
int RxCinput::eatWhiteSpace ()
{
	int c;
	do
	 { c = getChar();
	 }
	while (isspace(c));
	ungetChar (c);
	return c;
}

int RxCinput::input_str (char *s, int maxn)
{
	int c, k;

	if (status_ != 1)
	 { return status_;
	 }
	// start by skipping white space
	do
	 { c = getChar();
	 }
	while (isspace(c));
	for (k=0; k<maxn-1; k++)
	 { if (k > 0)
	    { c = getChar();
	    }
	   if (c == EOF)
	    { break;
	    }
	   else
	    { s[k] = c;
	    }
	 }
	s[k] = '\0';
	if (c == EOF)
	 { status_ = EOF;
	 }
	return status_;
}

int RxCinput::setHashTab (int size)
{
	if (htab)
	 { delete htab;
	 }
	htab = new RxStrHashTab (size);
	return htab->size();
}
