#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <RxError.h>
#include <RxCio.h>
#include <signal.h>
#include <mutexdefs.h>

#define ASSERT(e,s) if (e) { abort(); }

#define DEFAULT_ACTION	THROW
#define MAXSTRING	1024

DECLARE_STATIC_MUTEX(mux)
int RxError::idx = 0;

int RxError::assignClassIdx(int *idxp)
{
	static int classIdxCnt = 1;
	MUTEX_LOCK(&mux);
	if (*idxp == 0)
	 { *idxp = classIdxCnt++;
	 }
	MUTEX_UNLOCK(&mux);
	return *idxp;
}

int RxError::getClassIdx(int *idxp)
{
	if (idxp != &idx && *idxp == 0)
	 { assignClassIdx (idxp);
	 }
	return *idxp;
}

int RxError::getClassIdx() const
{
	int *idxp;

	idxp = getClassIdxPtr();
	if (idxp != &idx && *idxp == 0)
	 { assignClassIdx (idxp);
	 }
	return *idxp;
}

void RxError::initAction()
{
	action act;
	int *idxp;

	idxp = getClassIdxPtr();
	act = getActionIdx(idxp);
	if (act != DEFAULT)
	 { action_ = act;
	 }
	else if (getClassIdx() == 0)
	 { action_ = DEFAULT_ACTION;
	 }
}

RxError::RxError (const RxError &rec)
{
	code = rec.code;
	msg = (rec.msg ? strdup(rec.msg) : NULL);
	action_ = rec.action_;
}

RxError::~RxError ()
{
	if (msg)
	 { free (msg);
	 }
}

RxError::action RxError::setActionIdx(action act, int *idxp)
{
	return (RxErrCtx::getCur()->setClassAction(getClassIdx(idxp), act));
}

RxError::action RxError::getActionIdx(int *idxp)
{
	
	return (RxErrCtx::getCur()->getClassAction(getClassIdx(idxp)));
}

RxError::RxError()
{
	msg = NULL;
	code = 0;
	initAction();	
}

RxError::RxError (int n, const char *fmt, ... )
{
	va_list ap;

	msg = NULL;
	va_start (ap, fmt);
	vset (n, fmt, ap);
	va_end (ap);
	initAction();
}

RxError::RxError (int n)
{
	msg = NULL;
	set (n, NULL);
	initAction();
}

void RxError::set (int n, const char *fmt, ... )
{
	va_list ap;

	va_start (ap, fmt);
	vset (n, fmt, ap);
	va_end (ap);
}

void RxError::vset (int n, const char *fmt, va_list ap)
{
	char newfmt[MAXSTRING];
	char buf[MAXSTRING];
	const char *cp, *kp;
	char *np;

	code = n;
	if (fmt)
	 { 
	   for (cp=fmt, np=newfmt; *cp; )
	    { if (cp[0] == '%' && cp[1] == 'K')
	       { if ((kp = codeToStr(code)) != NULL)
		  { strcpy (np, kp);
		    np += strlen(kp);
		  }
		 cp += 2;
	       }
	      else
	       { *np++ = *cp++;
	       }
	    }
	   *np = '\0';
	   vsprintf (buf, newfmt, ap);
	   setMsg (buf);
	 }
}

void RxError::setMsg (const char *s)
{
	if (msg)
	 { free (msg);
	 }
	msg = strdup(s);
}

const char *RxError::codeToStr(int c) const
{
	const CodeList *list = getCodeList();
	const char *str = NULL;

	if (list != NULL)
	 { while (list->str)
	    { if (list->code == c)
	       { str = list->str;
		 break;
	       }
	      list++;
	    }
	 }
	return (str);
}

void RxError::raise()
{
	RxErrCtx *ctx;

	ctx = RxErrCtx::getCur();
	switch (action_)
	 {
	   case THROW:
	    { 
#ifndef NO_EXCEPTIONS
	      dothrow();
#else
	      ctx->elist.addTail(docopy());
	      ctx->print (stderr);
	      abort();
#endif
	      break;
	    }
	   case EXIT:
	    { ctx->elist.addTail(docopy());
	      ctx->print (stderr);
	      exit (1);
	      break;
	    }
	   case ABORT:
	    { ctx->elist.addTail(docopy());
	      ctx->print (stderr);
	      abort ();
	      break;
	    }
	   case POST:
	    { ctx->elist.addTail(docopy());
	      break;
	    }
	   case IGNORE:
	    { break;
	    }
	   default:
	    { fprintf (stderr, "RxError:raise: bogus action code %d\n", action_);
	      abort();
	    }
	 }
}

void RxError::post()
{
	RxErrCtx *ctx;

	ctx = RxErrCtx::getCur();
	ctx->elist.addTail(docopy());
}

RxErrCtx *RxErrCtx::ctx0 = NULL;
THRDATA_KEY RxErrCtx::key;
int RxErrCtx::keyinit = 0;

RxErrorAction RxErrCtx::setClassAction (int idx, RxErrorAction act)
{
	RxErrorAction oldact;
	int i;

	if (idx >= numClasses)
	 { int oldnum = numClasses;
	   while (idx >= numClasses)
	    { numClasses *= 2;
	    }
	   classActions = 
	      (RxErrorAction*)realloc(classActions,
				      sizeof(RxErrorAction)*numClasses);
	   for (i=oldnum; i<numClasses; i++)
	    { classActions[i] = RxError::DEFAULT;
	    }
	 }
	oldact = classActions[idx];
	classActions[idx] = act;
	return oldact;
}

RxErrorAction RxErrCtx::getClassAction (int idx)
{
	if (idx >= numClasses)
	 { return RxError::DEFAULT;
	 }
	else
	 { return classActions[idx];
	 }
}

RxErrCtx *RxErrCtx::getCur()
{
	void *ctx = ctx0;
	if (keyinit == 0)
	 { MUTEX_LOCK(&mux);
	   if (keyinit == 0)
	    { THRDATA_KEY_CREATE (&key, 0);
	      keyinit = 1;
	    }
	   MUTEX_UNLOCK(&mux);
	 }
	THRDATA_GET_DATA (key, ctx);
	if (ctx == NULL)
	 { ctx = new RxErrCtx();
	   THRDATA_SET_DATA (key, ctx);
	   ctx0 = (RxErrCtx*)ctx;
	 }
	return ((RxErrCtx*)ctx);
}

int RxErrCtx::clear()
{
	return elist.removeAndDeleteAll();
}

int RxErrCtx::numErrs()
{
	return elist.length();
}

int RxErrCtx::print (RxCoutput *co)
{
	RxError *err;

	for (err=elist.head(); err; err=err->next())
	 { co->output_str (err->msg);
	 }
	return co->nchars();
}

#if 0
void RxErrCtx::terminate()
{
	RxErrCtx::getCur()->print(stderr);
	abort();
}
#endif

RxErrCtx::RxErrCtx ()
{      
	numClasses = DEFAULT_NUM_CLASSES;
	int i;

	classActions = 
	   (RxErrorAction*)malloc(sizeof(RxErrorAction)*numClasses);
	for (i=0; i<numClasses; i++)
	 { classActions[i] = RxError::DEFAULT;
	 }
	err = NULL;
}

RxErrCtx::~RxErrCtx ()
{
	elist.removeAndDeleteAll();
	if (classActions)
	 { free (classActions);
	 }
}
