#include <RxError.h>
#include <assert.h>
#include <signal.h>
#include <setjmp.h>

class RxTestError : public RxError
{
	RX_ERROR_DECLS (RxTestError)
    public:
	enum {
	   IndexRange,
	   MemAlloc
	};
	RxTestError (int n, const char *fmt, ...);
	RxTestError (int n) : RxError (n)
	 { RX_ERROR_CONSTRUCTOR_CODE();
	 }
	RxTestError ()
	 { RX_ERROR_CONSTRUCTOR_CODE();
	 }
};

RX_ERROR_DEFS_BEGIN(RxTestError)
	IndexRange, "index range error",
	MemAlloc,   "out of memory",
RX_ERROR_DEFS_END()

RxTestError::RxTestError (int n, const char *fmt, ...)
{
	va_list ap;

	va_start (ap, fmt);
	vset (n, fmt, ap);
	va_end (ap);
	RX_ERROR_CONSTRUCTOR_CODE();
}

class RxNewError : public RxTestError {
	RX_ERROR_DECLS (RxNewError)
    public:
	enum {
	  Foo,
	  Bar
	};
	RxNewError (int n, const char *fmt, ...);
	RxNewError (int n) : RxTestError (n)
	 { RX_ERROR_CONSTRUCTOR_CODE ()
	 };
	RxNewError ()
	 { RX_ERROR_CONSTRUCTOR_CODE ()
	 }
};

//RX_ERROR_CODE(RxNewError)
RX_ERROR_DEFS_BEGIN(RxNewError)
	Foo, "foo error",
	Bar, "bar error",
RX_ERROR_DEFS_END()

RxNewError::RxNewError (int n, const char *fmt, ...)
{
	va_list ap;

	va_start (ap, fmt);
	vset (n, fmt, ap);
	va_end (ap);
	RX_ERROR_CONSTRUCTOR_CODE ()
}

bar()
{
	RxError (2, "%K this is error #%d %K\n", 2).raise();
}

foo()
{
	bar ();
}

gnew()
{
	RxNewError (RxNewError::Foo, "this is error %K %d\n", 3).raise();
}

gen()
{
	RxTestError (RxTestError::MemAlloc, "gen(): %K here\n").raise();
}

typedef struct {
	int code;
	char *name;	// used only if RxError contains a name field
	char *msg;
	char *trace[10];
} ErrorCheck;

checkError(RxError &err, ErrorCheck &check)
{
	const char *trace;
	char **tp;

	if (strcmp(err.getName(),check.name) != 0)
	 { fprintf (stderr, "err.getName() = %s vs. %s\n",
		    err.getName(), check.name);	    
	   return (0);
	 }
	if (err.getCode() != check.code)
	 { fprintf (stderr, "err.getCode() = %d vs. %d\n",
		    err.getCode(), check.code);
	   return (0);
	 }
	if (strcmp (err.getMsg(), check.msg) != 0)
	 { fprintf (stderr, "err.getMsg() = %s vs. %s\n",
		    err.getMsg(), check.msg);
	   return (0);
	 }
#if 0
	tp = check.trace;
	for (trace=err.getFirstTrace(); trace; trace=err.getNextTrace())
	 { if (*tp == NULL || strcmp (*tp, trace) != 0)
	    { fprintf (stderr, "*tp=%s vs. %s\n", 
		       *tp ? *tp : "NULL", trace);
	      return (0);
	    }
	   tp++;
	 }
	if (*tp != NULL)
	 { fprintf (stderr, "*tp != NULL\n");
	   return (0);
	 }
#endif
	return 1;
}

jmp_buf env;
int caught = 0;

void onAbort(int)
{
	caught = 1;
	longjmp (env, 1);
}

tryGen(ErrorCheck &check)
{
#ifndef NO_EXCEPTIONS
	try
#endif
	 { gen();
	 }
#ifndef NO_EXCEPTIONS
	catch (RxTestError err)
	 { caught = 1;
	   assert (checkError (err, check));
	   RxErrCtx::clearCur();
	 }
#endif
}

tryBar(ErrorCheck &check)
{
#ifndef NO_EXCEPTIONS
	try
#endif
	 { bar();
	 }
#ifndef NO_EXCEPTIONS
	catch (RxError err)
	 { caught = 1;
	   assert (checkError (err, check));
	   RxErrCtx::clearCur();
	 }
#endif
}

tryFoo(ErrorCheck &check)
{
#ifndef NO_EXCEPTIONS
	try
#endif
	 { foo();
	 }
#ifndef NO_EXCEPTIONS
	catch (RxError err)
	 { caught = 1;
       	   assert (checkError (err, check));
	   RxErrCtx::clearCur();
	 }
#endif
}

tryGnew(ErrorCheck &check)
{
#ifndef NO_EXCEPTIONS
	try
#endif
	 { gnew();
	 }
#ifndef NO_EXCEPTIONS
	catch (RxNewError err)
	 { caught = 1;
	   assert (checkError (err, check));
	   RxErrCtx::clearCur();
	 }
#endif
}

main()
{
	int dummy;
	RxError *err;

	RxError::setAction(RxError::IGNORE);
	assert (RxError::getAction() == RxError::IGNORE);
	assert (RxTestError::getAction() == RxError::DEFAULT);
	assert (RxNewError::getAction() == RxError::DEFAULT);
	assert (RxError().netAction() == RxError::IGNORE);
	assert (RxTestError().netAction() == RxError::IGNORE);
	assert (RxNewError().netAction() == RxError::IGNORE);
	
	RxTestError::setAction(RxError::POST);
	assert (RxError::getAction() == RxError::IGNORE);
	assert (RxTestError::getAction() == RxError::POST);
	assert (RxNewError::getAction() == RxError::DEFAULT);
	assert (RxError().netAction() == RxError::IGNORE);
	assert (RxTestError().netAction() == RxError::POST);
	assert (RxNewError().netAction() == RxError::POST);
	
	RxNewError::setAction(RxError::THROW);
	assert (RxError::getAction() == RxError::IGNORE);
	assert (RxTestError::getAction() == RxError::POST);
	assert (RxNewError::getAction() == RxError::THROW);
	assert (RxError().netAction() == RxError::IGNORE);
	assert (RxTestError().netAction() == RxError::POST);
	assert (RxNewError().netAction() == RxError::THROW);
	
	static ErrorCheck echeck2 = { 2, "RxError", 
	   " this is error #2 \n",
	   { "main", NULL }};

	static ErrorCheck echeck1 = { 2, "RxError", 
	   " this is error #2 \n",
	   { "foo", "main", NULL }};

#ifndef NO_EXCEPTIONS
	RxError::setAction(RxError::THROW);
	caught = 0;
 	tryFoo (echeck1);
	assert (caught == 1);
	assert (RxErrCtx::lastErrCur() == (RxError*)NULL);

	caught = 0;
	tryBar(echeck2);
	assert (caught == 1);
	assert (RxErrCtx::lastErrCur() == (RxError*)NULL);
	assert (RxErrCtx::numErrsCur() == 0);
#endif

	RxError::setAction(RxError::IGNORE);
	caught = 0;
	tryFoo (echeck1);
	assert (caught == 0);
	assert (RxErrCtx::numErrsCur() == 0);
	caught = 0;
	tryBar (echeck2);
	assert (caught == 0);
	assert (RxErrCtx::numErrsCur() == 0);

	RxError::setAction(RxError::POST);
	tryFoo (echeck1);
	assert (caught == 0);
	assert (RxErrCtx::numErrsCur() == 1);
	assert (checkError (*RxErrCtx::firstErrCur(), echeck1));
	assert (RxErrCtx::firstErrCur() == RxErrCtx::lastErrCur());
	assert (RxErrCtx::nextErrCur() == NULL);
	RxErrCtx::clearCur();
	assert (RxErrCtx::numErrsCur() == 0);
	assert (RxErrCtx::firstErrCur() == NULL);
	assert (RxErrCtx::nextErrCur() == NULL);
	assert (RxErrCtx::prevErrCur() == NULL);
	assert (RxErrCtx::lastErrCur() == NULL);

	caught = 0;
	tryBar(echeck2);
	assert (caught == 0);
	assert (RxErrCtx::numErrsCur() == 1);
	assert (checkError (*RxErrCtx::lastErrCur(), echeck2));

	static ErrorCheck genCheck1 = { RxTestError::MemAlloc, "RxTestError",
	   "gen(): out of memory here\n",
	   { "main", NULL } };

	RxErrCtx::clearCur();
#ifndef NO_EXCEPTIONS
	RxTestError::setAction(RxError::THROW);
	caught = 0;
	tryGen (genCheck1);
	assert (caught == 1);
	assert (RxErrCtx::numErrsCur() == 0);
#endif

	RxErrCtx::clearCur();
	RxTestError::setAction(RxError::POST);
	caught = 0;
 	tryGen (genCheck1);
	assert (caught == 0);
	assert (RxErrCtx::numErrsCur() == 1);
	RxTestError *generr = (RxTestError*)RxErrCtx::lastErrCur();
	assert (strcmp(RxErrCtx::lastErrCur()->getName(), 
		      genCheck1.name) == 0);
	assert (checkError (*RxErrCtx::lastErrCur(), genCheck1));

	static ErrorCheck newCheck1 = { RxNewError::Foo, "RxNewError",
	   "this is error foo error 3\n",
	   { "main", NULL } };

	RxErrCtx::clearCur();
#ifndef NO_EXCEPTIONS
	RxNewError::setAction(RxError::THROW);
	caught = 0;
	tryGnew(newCheck1);
	dummy = 1;
	assert (caught == 1);
	assert (RxErrCtx::numErrsCur() == 0);
#endif
	RxErrCtx::clearCur();
	RxNewError::setAction(RxError::POST);
	caught = 0;
	tryGnew(newCheck1);
	dummy = 1;
	assert (caught == 0);
	assert (RxErrCtx::numErrsCur() == 1);
	assert (strcmp(RxErrCtx::firstErrCur()->getName(), 
		       newCheck1.name) == 0);
	assert (checkError (*(RxNewError*)RxErrCtx::firstErrCur(),newCheck1));

	signal (SIGABRT, onAbort);
	if (setjmp(env))
	 { /* then the abort happened */
	   assert (caught == 1);
	   assert (RxErrCtx::numErrsCur() == 1);
	 }
	else
	 { RxErrCtx::clearCur();
	   RxError::setAction(RxError::ABORT);
	   caught = 0;
	   foo();
	   fprintf (stderr, "Shouldn't get here; should have aborted\n");
	   exit (1);
	 }

	printf ("\nROUTINES OK\n\n");
	return (0);
}
