#include <RxPrintSpec.h>
#include <stdio.h>
#include <stdiox.h>
#include <stringx.h>

RxFmtNode::RxFmtNode (const char *s, int n, int type)
{
	if (n > 0)
	 { fmt_ = strndup (s, n);
	 }
	else
	 { fmt_ = strdup (s);
	 }
	width_ = 0;
	type_ = type;
}

RxFmtNode::RxFmtNode (const char *s1, const char *s2, int type)
{
	fmt_ = (char*)malloc(strlen(s1)+strlen(s2)+1);
	strcpy (fmt_, s1);
	strcat (fmt_, s2);
	width_ = 0;
	type_ = type;
}

const RxFmtNode *RxFmtNode::nextConv() const
{
	const RxFmtNode *node;

	for (node=next(); node; node=node->next())
	 { if (node->type_ == RxFmtNode::CONV)
	    { return node;
	    }
	 }
	return NULL;
}

const RxFmtNode *RxFmtNode::nextStr() const
{
	const RxFmtNode *node;

	for (node=next(); node; node=node->next())
	 { if (node->type_ == RxFmtNode::STR)
	    { return node;
	    }
	 }
	return NULL;
}

RxFmtNode *RxPrintSpec::addConvNode (const char *s, int n)
{
	RxFmtNode *fnode;

	fnode = new RxFmtNode (s, n, RxFmtNode::CONV);
	fnode->width_ = convFieldWidth(fnode->fmt());
	fmtList.addTail (fnode);
	nconv++;
	return fnode;
}

RxFmtNode *RxPrintSpec::addStrNode (const char *s, int n)
{	
	RxFmtNode *fnode;

	fnode = new RxFmtNode (s, n, RxFmtNode::STR);
	fnode->width_ = strlen(fnode->fmt());
	fmtList.addTail (fnode);
	nstr++;
	return fnode;
}

RxFmtNode *RxPrintSpec::addSpecialNode (const char *s, int n)
{
	RxFmtNode *fnode;

	fnode = new RxFmtNode (s, n, RxFmtNode::CONV);
	fnode->width_ = convFieldWidth(fnode->fmt());
	specialList.addTail (fnode);
	nspecial++;
	return fnode;
}
       
void RxPrintSpec::init(int leadspc)
{
	prefixNode_ = NULL;
	suffixNode_ = NULL;
	leadSpacingBase = leadspc;
	leadSpacing = leadspc;
	nconv = 0;
	nstr = 0;
	nspecial = 0;
}

void RxPrintSpec::clear()
{
	fmtList.removeAndDeleteAll();
	specialList.removeAndDeleteAll();
	prefixNode_ = NULL;
	suffixNode_ = NULL;
	nconv = 0;
	nstr = 0;
	nspecial = 0;
}

int RxPrintSpec::set(const char *s)
{
	const char *cp;
	char *bp, *ep;
	int status;

	clear();

	/* parse the input string s to find the internal
	   format sequence */

	if (s)
	 { cp = s;
	   do
	    { status = fmtindex (&bp, &ep, cp);
	      if (status != 0)
	       { if ((bp-cp) > 0)
		  { addStrNode (cp, bp-cp);
		    if (cp == s)
		     { prefixNode_ = fmtList.head();
		     }
		    cp = bp;
		  }
		 if (status == 1)
		  { addConvNode (cp, ep+1-cp);
		  }
		 else
		  { // this is an unrecognized format; add to the special list
		    addSpecialNode (cp, ep+1-cp);
		  }
		 cp = ep+1;
	       }
	    }
	   while (status != 0);
	   if (*cp != '\0' && nconv > 0)
	    { suffixNode_ = addStrNode (cp, 0);
	    }
	 }
	if (nconv == 0 && nspecial == 0)
	 { prefixNode_ = addStrNode (s, 0);
	 }
#define TAB_SIZE	8
	/* compute the column which we are left in after printing prefix,
	   assuming eight character tabs. */

	if (prefix())
	 { /* print buf should be *less* than strlen(prefix), but we
	      allocate a bigger buffer anyway */

	   char *cp;
	   int col;

	   char *buf = (char*)malloc(2*strlen(prefix())+1);
	   sprintf (buf, prefix());
	   col = 0;
	   for (cp=buf; *cp !='\0'; cp++)
	    { if (*cp == '\t')
	       { col += TAB_SIZE - (col%TAB_SIZE);
	       }
	      else if (*cp == '\n' || *cp == '\r')
	       { col = 0;
	       }
	      else
	       { col++;
	       }
	    }
	   leadSpacing = leadSpacingBase + col;
	   free (buf);
	 }
	   
	return (1);
}

int RxPrintSpec::set (const RxFmtNode *fnode1, const RxFmtNode *fnode2)
{
	const RxFmtNode *fnode;

	clear();
	
	for (fnode=fnode1; fnode!=fnode2->next(); fnode=fnode->next())
	 { if (fnode->type() == RxFmtNode::STR)
	    { addStrNode (fnode->fmt(), 0);
	    }
	   else
	    { addConvNode (fnode->fmt(), 0);
	    }
	 }
	if (nstr > 0 && fmtList.head()->type() == RxFmtNode::STR)
	 { prefixNode_ = fmtList.head();
	 }
	if (nconv > 0 && fmtList.tail()->type() == RxFmtNode::STR)
	 { suffixNode_ = fmtList.tail();
	 }
	return 1;
}

int RxPrintSpec::set (const RxPrintSpec &ref)
{
	const RxFmtNode *fnode;

	clear();
	for (fnode=ref.fmtList.head(); fnode; fnode=fnode->next())
	 { if (fnode->type() == RxFmtNode::STR)
	    { addStrNode (fnode->fmt(), 0);
	    }
	   else
	    { addConvNode (fnode->fmt(), 0);
	    }
	   if (fnode == ref.prefixNode_)
	    { prefixNode_ = fmtList.tail();
	    } 
	   if (fnode == ref.suffixNode_)
	    { suffixNode_ = fmtList.tail();
	    }
	 }
	for (fnode=ref.specialList.head(); fnode; fnode=fnode->next())
	 { addSpecialNode (fnode->fmt(), 0);
	 }
	nconv = ref.nconv;
	nstr = ref.nstr;
	nspecial = ref.nspecial;
	leadSpacingBase = ref.leadSpacingBase;
	leadSpacing = ref.leadSpacing;
	return 1;
}

void RxPrintSpec::clearPrefixAndSuffix()
{
	if (prefixNode_)
	 { fmtList.removeAndDelete (prefixNode_);
	   nstr--;
	   prefixNode_ = NULL;
	 }
	if (suffixNode_)
	 { fmtList.removeAndDelete (suffixNode_);
	   nstr--;
	   suffixNode_ = NULL;
	 }
}

RxPrintSpec::~RxPrintSpec ()
{
	fmtList.removeAndDeleteAll();
	specialList.removeAndDeleteAll();
}

const RxFmtNode *RxPrintSpec::fmtNode(int n) const
{
	const RxFmtNode *node;
	int i;

	if (n < 0)
	 { n += nconv+nstr;
	 }
	i=0;
	for (node=fmtList.head(); node; node=node->next())
	 { if (i++ == n)
	    { break;
	    }
	 }
	return node;
}

const RxFmtNode *RxPrintSpec::convNode(int n) const
{
	const RxFmtNode *node;
	int i;

	if (n < 0)
	 { n += nconv;
	 }
	i=0;
	for (node=fmtList.head(); node; node=node->next())
	 { if (node->type() == RxFmtNode::CONV)
	    { if (i++ == n)
	       { break;
	       }
	    }
	 }
	return node;
}

const RxFmtNode *RxPrintSpec::strNode(int n) const
{
	const RxFmtNode *node;
	int i;

	if (n < 0)
	 { n += nstr;
	 }
	i=0;
	for (node=fmtList.head(); node; node=node->next())
	 { if (node->type() == RxFmtNode::STR)
	    { if (i++ == n)
	       { break;
	       }
	    }
	 }
	return node;
}

const RxFmtNode *RxPrintSpec::specialNode(int n) const
{
	const RxFmtNode *node;
	int i;

	if (n < 0)
	 { n += nspecial;
	 }
	i=0;
	for (node=specialList.head(); node; node=node->next())
	 { if (i++ == n)
	    { break;
	    }
	 }
	return node;
}

const RxFmtNode *RxPrintSpec::findSpecial (const char *s) const
{
	const RxFmtNode *node;

	for (node=specialList.head(); node; node=node->next())
	 { if (strcmp (s, node->fmt()) == 0)
	    { return node;
	    }
	 }
	return NULL;
}
