#include <RxCvxPolygon2.h>
#include <RxPrintSpec.h>
#include <RxParseObj.h>
#include <cdefs.h>
#include <math.h>
#include <mathx.h>
#include <mutexdefs.h>
#include <assert.h>
#include <iostream.h>
#include <RxCio.h>

#define EPSILON	(1e-14)	/* double precision limit */

#define SGN3(x)	((x) > 0 ? 1 : (x) < 0 ? -1 : 0)

RxEdge2 *RxEdge2::cache;
RxVrtx2 *RxVrtx2::cache;

DECLARE_STATIC_MUTEX(mux)

#define DEBUG	1

#ifdef DEBUG
#define DEBUG_STATEMENT(xxx) if (RxCvxPolygon2::debug) xxx;
#else
#define DEBUG_STATEMENT(xxx)
#endif

RxEdge2& RxEdge2::setV(const RxVecd2 &qh, const RxVecd2 &qt)
{
	if (vt == NULL)
	 { vt = RxVrtx2::alloc();
	 }
	if (vh == NULL)
	 { vh = RxVrtx2::alloc();
	 }
	vh->setV(qh);
	vt->setV(qt);
	u_.subQ (qh, qt);
	return *this;
}

RxEdge2& RxEdge2::setV(double xh, double yh, double xt, double yt)
{
	if (vt == NULL)
	 { vt = RxVrtx2::alloc();
	 }
	if (vh == NULL)
	 { vh = RxVrtx2::alloc();
	 }
	vh->setV(xh, yh);
	vt->setV(xt, yt);
	u_.set (xh-xt, yh-yt);
	return *this;
}

RxEdge2& RxEdge2::setV(const RxVecd2 &qh)
{
	if (vh == NULL)
	 { vh = RxVrtx2::alloc();
	 }
	vh->setV(qh);
	vt = vh;
	u_.setZero();
	return *this;
}

RxEdge2& RxEdge2::setV(double xh, double yh)
{
	if (vh == NULL)
	 { vh = RxVrtx2::alloc();
	 }
	vh->setV(xh, yh);
	vt = vh;
	u_.setZero();
	return *this;
}

RxEdge2::~RxEdge2()
{
	if (vt)
	 { RxVrtx2::free (vt);
	   if (vh && vh != vt)
	    { RxVrtx2::free (vh);
	    }
	 }
}

RxEdge2 *RxEdge2::alloc()
{ 
	RxEdge2 *ep;
	MUTEX_LOCK(&mux);
	if ((ep = cache))
	 { cache = ep->link;
	   MUTEX_UNLOCK(&mux);
	   return (ep);
	 }
	MUTEX_UNLOCK(&mux);
	return (new RxEdge2);
}

RxEdge2 *RxEdge2::allocList(int n)
{ 
	RxEdge2 *ep, *pp;

	if (n < 1)	// sanity check
	 { return (NULL);
	 }
	MUTEX_LOCK(&mux);
	pp = ep = cache;
	while (pp && --n)
	 { pp = pp->link;
	 }
	if (pp)
	 { cache = pp->link;
	   pp->link = NULL;
	 }
	else
	 { cache = NULL;
	 }
	MUTEX_UNLOCK(&mux);
	while (n--)
	 { pp = ep;
	   ep = new RxEdge2;
	   ep->link = pp;
	 }
	return (ep);
}

void RxEdge2::free (RxEdge2 *ep)
{
	MUTEX_LOCK(&mux);
	ep->link = cache;
	cache = ep;
	MUTEX_UNLOCK(&mux);
}

void RxEdge2::freeList (RxEdge2 *ep)
{
	RxEdge2 *lp;
	for (lp=ep; lp->link; lp=lp->link)
	   ;
	MUTEX_LOCK(&mux);
	lp->link = cache;
	cache = ep;
	MUTEX_UNLOCK(&mux);
}


RxVrtx2 *RxVrtx2::alloc()
{ 
	RxVrtx2 *vp;
	MUTEX_LOCK(&mux);
	if ((vp = cache))
	 { cache = vp->link;
	   MUTEX_UNLOCK(&mux);
	   return (vp);
	 }
	MUTEX_UNLOCK(&mux);
	return (new RxVrtx2);
}

RxVrtx2 *RxVrtx2::allocList(int n)
{ 
	RxVrtx2 *vp, *pp;

	if (n < 1)	// sanity check
	 { return (NULL);
	 }
	MUTEX_LOCK(&mux);
	pp = vp = cache;
	while (pp && --n)
	 { pp = pp->link;
	 }
	if (pp)
	 { cache = pp->link;
	   pp->link = NULL;
	 }
	else
	 { cache = NULL;
	 }
	MUTEX_UNLOCK(&mux);
	while (n--)
	 { pp = vp;
	   vp = new RxVrtx2;
	   vp->link = pp;
	 }
	return (vp);
}

void RxVrtx2::free (RxVrtx2 *vp)
{
	MUTEX_LOCK(&mux);
	vp->link = cache;
	cache = vp;
	MUTEX_UNLOCK(&mux);
}

void RxVrtx2::freeList (RxVrtx2 *vp)
{
	RxVrtx2 *lp;
	for (lp=vp; lp->link; lp=lp->link)
	   ;
	MUTEX_LOCK(&mux);
	lp->link = cache;
	cache = vp;
	MUTEX_UNLOCK(&mux);
}

double RxEdge2::intersect (RxVecd2 *p, const RxLine2 &l2, double tol) const
{
	/*
	   Suppose we have two line segments, given by

		v = q1 + lam1 u1   and    v = q2 + lam2 u2

	   where q1, u1, q2, and u2 are 2 vectors. In general, finding 
	   the intersection of these two points amounts to solving

		[ u1x  -u2x ] [ lam1 ]
		[           ] [      ] = q2 - q1 = del;
		[ u1y  -u2y ] [ lam2 ] 

	   for which we get

		       -u2y delx + u2x dely    u2 X del
		lam1 = --------------------  = -------
			 u2x u1y - u1x u2y     u2 X u1

		       -u1y delx + u1x dely    u1 X del
		lam2 = --------------------  = -------
			 u2x u1y - u1x u2y     u2 X u1
	*/

	RxVecd2 del;
	double x, l;

	x = l2.u().cross(u_);
	if (ABS(x) <= tol)
	 { // then the lines are parallel
	   if (p)
	    { del.sub(l2.q(),vh->q_);
	      l = del.x()*u_.x()+del.y()*u_.y();
	      p->scaleQ (l, u_);
	      p->addQ(*p, vh->q_);
	    }
	   return (HUGE_VAL);
	 }
	else
	 { del.sub(l2.q(),vh->q_);
	   l = l2.u().cross(del)/x;
	   if (p)
	    { p->scaleQ (l, u_);
	      p->addQ(*p, vh->q_);
	    }
	   return (l);
	 }
}

int RxCvxPolygon2::debug;

void RxCvxPolygon2::setDebug (int on)
{
	debug = on;
}

int RxCvxPolygon2::consistencyCheck (int)
{
	const RxEdge2 *ep, *tp;
	const RxVrtx2 *vp, *pp;
	int nvtx = 0;
	int k;

	if ((ep = vlist.head()))
	 { pp = vlist.tail()->vh;
	   do
	    { vp = ep->vh;
	      if (vp->e != ep)
	       { printf (
"RxCvxPolygon2::sanityCheck: vrtx doesn't point to edge: ep=%x vp->e=%x %x %x\n", 
			 ep, ep->vh, ep->next(), ep->prev());
		 return (0);
	       }
	      if ((vp->q()-pp->q()-ep->u()).norm2() > getCharDist()*EPSILON)
	       { printf (
"RxCvxPolygon2::sanityCheck: u is inconsitent, vertex %d, tol=%g:\n", 
			  nvtx, getCharDist()*EPSILON);
		 vp->q().print ("  vp->q: %12.7f\n");
		 pp->q().print ("  pp->q: %12.7f\n");
		 ep->u().print ("  ep->u: %12.7f\n");
		 return (0);
	       }
	      if (numv > 1 && (vp->q()-pp->q()).norm2() < getCharDist()*EPSILON)
	       { printf (
"RxCvxPolygon2::sanityCheck: repeated vertex at %d, tol=%g\n", 
			nvtx, getCharDist()*EPSILON);
		 vp->q().print ("vp->q %12.7f\n");
		 pp->q().print ("pp->q %12.7f\n");
		 return (0);
	       }
	      for (k=0, tp=vlist.head(); tp!=ep; k++, tp=tp->next())
		 ;
	      if (k != nvtx)
	       { printf (
"RxCvxPolygon2::sanityCheck: vertex %d duplicated at %d:\n", nvtx, k);
		 return (0);
	       }
	      nvtx++;
	      pp = vp;
	      ep = ep->next();
	    }
	   while (ep != vlist.head());
	 }
	if (nvtx != numv)
	 { printf ( 
"RxCvxPolygon2::sanityCheck: numv=%d, actual number is %d\n", numv, nvtx);
	   return (0);
	 }
	return (1);
}

void RxCvxPolygon2::init()
{
	vlist.init();
	chardist = -1.0;
	eps = -1.0;
	xmin = HUGE_VAL;
	numv = 0;
//	curEdge = 0;
}

double RxCvxPolygon2::getEps()
{
	double d;

	if (eps == -1)
	 { if (numv == 1)
	    { d = vlist.head()->vh->q().norm1();
	    }
	   else
	    { calcBoundingBox();
	      d = (xmax-xmin)+(ymax-ymin);
	    }
	   eps = d*d*EPSILON;
	 }
	return (eps);
}

double RxCvxPolygon2::getCharDist()
{
	double cdist;

	if (chardist != -1)
	 { cdist = chardist;
	 }
	else
	 { cdist = sqrt(getEps()/EPSILON);
	 }
	return (cdist);
}

void RxCvxPolygon2::setCharDist(double cdist)
{
	chardist = cdist;
	eps = cdist*cdist*EPSILON;
}

void RxCvxPolygon2::calcBoundingBox()
{
	RxEdge2 *ep;
	RxVrtx2 *vp;

	if (xmin == HUGE_VAL)
	 { if ((ep = vlist.head()))
	    { vp = ep->vh;
	      xmin = vp->q().x();
	      xmax = vp->q().x();
	      ymin = vp->q().y();
	      ymax = vp->q().y();

	      do
	       { if (vp->q().x() < xmin)
		  { xmin = vp->q().x();
		  }
		 else if (vp->q().x() > xmax)
		  { xmax = vp->q().x();
		  }
		 if (vp->q().y() < ymin)
		  { ymin = vp->q().y();
		  }
		 else if (vp->q().y() > ymax)
		  { ymax = vp->q().y();
		  }
		 ep = ep->next();
		 vp = ep->vh;
	       }
	      while (ep != vlist.head());
	    }
	   else
	    { xmin = xmax = 0;
	      ymin = ymax = 0;
	    }
	 }
}

int RxCvxPolygon2::setVertices (double *vtxs, int n)
{
	RxEdge2 *ep;
	RxVrtx2 *vp;
	double x0, y0;
	double xp, yp;

	clear();
	if (n < 1)
	 { return 0;
	 }
	x0 = vtxs[0];
	y0 = vtxs[1];
	ep=RxEdge2::allocList(n);
	vp=RxVrtx2::allocList(n);
	for ( ; ep; ep=ep->link, vp=vp->link)
	 { ep->vh = vp;
	   vp->e = ep;
	   vp->setV (vtxs[0], vtxs[1]);
	   if (!vlist.empty())
	    { ep->setU (vtxs[0]-xp, vtxs[1]-yp);
//	      ep->vt = vlist.tail()->vh;
	    }
	   xp = vtxs[0];
	   yp = vtxs[1];
	   vlist.addTail (ep);
	   vtxs += 2;
	 }
	vlist.head()->setU(x0-xp, y0-yp);
//	vlist.head()->vt = vlist.tail()->vh;
	numv = n;
	return 1;
}

int RxCvxPolygon2::getVertices (double *vtxs) const
{
	const RxEdge2 *ep;
	int n = 0;

	if ((ep = vlist.head()))
	 { do
	    { vtxs[2*n+0] = ep->vh->q().x();
	      vtxs[2*n+1] = ep->vh->q().y();
	      ep = ep->next();
	      n++;
	    }
	   while (ep != vlist.head());
	 }
	return (n);
}

int RxCvxPolygon2::set (const RxCvxPolygon2 &poly1)
{
	const RxEdge2 *ep1;
	RxEdge2 *ep;
	RxVrtx2 *vp;

	clear();
	ep = RxEdge2::allocList(poly1.numv);
	vp = RxVrtx2::allocList(poly1.numv);
	for (ep1=poly1.vlist.head(); ep; ep=ep->link, vp=vp->link)
	 { ep->vh = vp;
	   vp->e = ep;
	   vp->setV (ep1->nextVrtx()->q());
	   ep->setU (ep1->u());
//	   if (vlist.tail())
//	    { ep->vt = vlist.tail()->vh;
//	    }
	   vlist.addTail (ep);
	   ep1=ep1->next();
	 }
//	if (poly1.numv > 0)
//	 { vlist.head()->vt = vlist.tail()->vh;
//	 }
	if (chardist == -1)
	 { eps = poly1.eps;
	 }
	numv = poly1.numv;
	xmin = poly1.xmin;
	xmax = poly1.xmax;
	ymin = poly1.ymin;
	ymax = poly1.ymax;
	return 1;
}

int RxCvxPolygon2::setRect (double x, double y, double w, double h)
{
	RxVrtx2 *vp;
	RxEdge2 *ep;

	clear();
	ep = RxEdge2::allocList(4);
	vp = RxVrtx2::allocList(4);
	ep->vh = vp;
	vp->e = ep;
	vp->setV (x, y);
	ep->setU (0, -h);
	vlist.addTail (ep);

	ep = ep->link;
//	ep->vt = vp;
	vp = vp->link;
	ep->vh = vp;
	vp->e = ep;
	vp->setV (x+w, y);
	ep->setU (w, 0);
	vlist.addTail (ep);

	ep = ep->link;
//	ep->vt = vp;
	vp = vp->link;
	ep->vh = vp;
	vp->e = ep;
	vp->setV (x+w, y+h);
	ep->setU (0, h);
	vlist.addTail (ep);

	ep = ep->link;
//	ep->vt = vp;
	vp = vp->link;
	ep->vh = vp;
	vp->e = ep;
	vp->setV (x, y+h);
	ep->setU (-w, 0);
	vlist.addTail (ep);

//	vlist.head()->vt = vp;

	if (w > 0)
	 { xmin = x;
	   xmax = x+w;
	 }
	else
	 { xmin = x+w;
	   xmax = x;
	 }
	if (h > 0)
	 { ymin = y;
	   ymax = y+h;
	 }
	else
	 { ymin = y+h;
	   ymax = y;
	 }
	numv = 4;
	return 1;
}

int RxCvxPolygon2::setBegin (int n)
{
	RxEdge2 *ep;
	RxVrtx2 *vp;

	clear();
	if (n > 0)
	 { ep = RxEdge2::allocList(n);
	   vp = RxVrtx2::allocList(n);
	   curEdge = ep;
	   for ( ; ep; ep=ep->link, vp=vp->link)
	    { ep->vh = vp;
//	      if (vlist.tail())
//	       { ep->vt = vlist.tail()->vh;
//	       }
	      vp->e = ep;
	    }
//	   curEdge->vt = vlist.tail()->vh;
	 }
	else
	 { curEdge = NULL;
	 }
	return 1;				
}

RxVrtx2 *RxCvxPolygon2::addVrtx (double x, double y)
{
	RxEdge2 *ep;
	RxVrtx2 *vp;

	if ((ep = curEdge))
	 { vp = ep->vh;
	   curEdge = curEdge->link;
	 }
	else
	 { ep = RxEdge2::alloc();
	   vp = RxVrtx2::alloc();
	   ep->vh = vp;
	   vp->e = ep;
//	   if (vlist.tail())
//	    { ep->vt = vlist.tail()->vh;
//	    }
	 }
	vp->setV (x, y);
	vlist.addTail (ep);
	if (ep != vlist.head())
	 { RxVrtx2 *pp;
	   pp = ep->prev()->vh;
	   ep->setU (vp->q().x()-pp->q().x(),
		     vp->q().y()-pp->q().y());
	 }
	numv++;
	return (vp);
}

int RxCvxPolygon2::setEnd()
{
	RxVrtx2 *vp;
	RxVrtx2 *pp;

	if (numv > 0)
	 { 
	   vp = vlist.head()->vh;
	   pp = vlist.tail()->vh;	

	   vlist.head()->setU (vp->q().x()-pp->q().x(),
			       vp->q().y()-pp->q().y());
//	   vlist.head()->vt = vlist.tail()->vh;
	 }
	return 1;
}

void RxCvxPolygon2::linkNodes (RxEdge2 *ep0, RxEdge2 *ep1)
{
	// link the edge and vertex nodes from ep0 to ep1
	// in preparation for deallocation

	RxEdge2 *ep, *epNext = NULL;

	for (ep=ep0; ep!=ep1; ep=epNext)
	 { epNext = ep->next();
	   ep->link = epNext;
	   ep->vh->link = epNext->vh;
	 }
	ep->link = NULL;
	ep->vh->link = NULL;
}

void RxCvxPolygon2::clear()
{
	if (numv != 0)
	 { linkNodes (vlist.head(), vlist.tail());
	   RxVrtx2::freeList (vlist.head()->vh);
	   RxEdge2::freeList (vlist.head());
	   vlist.removeAll();
	   numv = 0;
	 }
	if (chardist == -1)
	 { eps = -1;
	 }
	xmin = HUGE_VAL;
}

void RxCvxPolygon2::shiftVtxs (int n)
{
	vlist.shift (n);	
}

/* The data structure IntersectCtx holds various computational values 
   used in the intersection computation.

   The intersection of polyP and polyQ is computed by starting with
   a specific edge on each one, and then advancing each of these edges
   according to certain rules. After each advance, the new leading edges
   are checked for intersection. 

   Notation: 

   The leading edge of polyP will be called ``P'', and will be denoted
   by the vertices p1 and p2 (in counterclockwise order, respectively).
   The preceding edge is defined by the vertices p0 and p1. 
   Likewise, leading edge of polyQ will be called ``Q'', with vertices
   q1 and q2, and the preceding edge defined by q0 and q1.
*/

void RxCvxPolygon2::centroid (RxVecd2 *v) const
{
	/* Compute the centroid of a polygon 'pgn'. */

	const RxEdge2 *ep;
	register double x, y;

	x = y = 0;
	if ((ep = vlist.head()))
	 { do
	    { x += ep->vh->q().x();
	      y += ep->vh->q().y();
	      ep = ep->next();
	    }
	   while (ep != vlist.head());
	 }
	v->x() = x/numv;
	v->y() = y/numv;
}

double RxCvxPolygon2::area () const
{
	/* Compute the area of the polygon */

	const RxEdge2 *ep;
	const RxVrtx2 *vp;
	double v0x, v0y;    /* coordinates of first vertex v0 */
	double d1x, d1y;    /* vector from v0 to some vertex vi */
	double d2x, d2y;    /* vector from v0 to vertex v(i+1) */
	double area;	    /* accumulated area */

	area = 0;

	ep = vlist.head();
	vp = ep->vh;
	v0x = vp->q().x();
	v0y = vp->q().y();

	ep = ep->next();
	vp = ep->vh;
	d2x = vp->q().x() - v0x;
	d2y = vp->q().y() - v0y;

	for (ep=ep->next(); ep!=vlist.head(); ep=ep->next())
	 { d1x = d2x;
	   d1y = d2y;
	   vp = ep->vh;
	   d2x = vp->q().x() - v0x;
	   d2y = vp->q().y() - v0y;
	   area += (d1x*d2y - d1y*d2x);
	 }
	return (area/2);
}

int RxCvxPolygon2::pointIsInside (const RxVecd2 &v)
{
	// returns 1 if v is inside, 0 if it is on the boundary,
	// and -1 if it is outside.

	RxEdge2 *ep;
	RxVrtx2 *vp;
	double prod;

	getEps();
	if ((ep = vlist.head()))
	 { do
	    { vp = ep->vh;
	      prod = ep->u().x()*(v.y()-vp->q().y()) - ep->u().y()*(v.x()-vp->q().x());
	      if (prod <= eps)
	       { return (prod < -eps ? -1 : 0);
	       }
	      ep = ep->next();
	    }
	   while (ep != vlist.head());
	   return 1;
	 }
	else
	 { return -1;
	 }
}

void RxCvxPolygon2::xform (double px, double py, double ang)
{
	RxEdge2 *ep;
	RxVrtx2 *vp;
	double x, y;
	double s, c;
	
	s = sin(ang);
	c = cos(ang);

	if ((ep = vlist.head()))
	 { do
	    { vp = ep->vh;
	      x = ep->u().x()*c - ep->u().y()*s;
	      y = ep->u().x()*s + ep->u().y()*c;
	      ep->setU (x, y);

	      x = vp->q().x()*c - vp->q().y()*s + px;
	      y = vp->q().x()*s + vp->q().y()*c + py;
	      ep->nextVrtx()->setV (x, y);

	      if (xmin != HUGE_VAL)
	       { if (ep == vlist.head())
		  { xmin = xmax = x;
		    ymin = ymax = y;
		  }
		 else
		  { if (x < xmin)
		     { xmin = x;
		     }
		    else if (x > xmax)
		     { xmax = x;
		     }
		    if (y < ymin)
		     { ymin = y;
		     }
		    else if (y > ymax)
		     { ymax = y;
		     }
		  }
	       }
	      ep = ep->next();
	    }
	   while (ep != vlist.head());
	 }
	if (xmin != HUGE_VAL)
	 { x = xmin*c - ymin*s + px;
	   y = xmin*s + ymin*c + py;
	   xmin = x;
	   ymin = y;
	   x = xmax*c - ymax*s + px;
	   y = xmax*s + ymax*c + py;
	   xmax = x;
	   ymax = y;
	 }
}

void RxCvxPolygon2::xform (const RxCvxPolygon2 &poly1, 
			     double px, double py, double ang)
{
	const RxEdge2 *ep1;
	RxEdge2 *ep;
	const RxVrtx2 *vp1;
	RxVrtx2 *vp;
	double x, y;
	double s, c;
	
	s = sin(ang);
	c = cos(ang);

	clear();
	ep = RxEdge2::allocList(poly1.numv);
	vp = RxVrtx2::allocList(poly1.numv);
	for (ep1=poly1.vlist.head(); ep; ep=ep->link, vp=vp->link)
	 { x = ep1->u().x()*c - ep1->u().y()*s;
	   y = ep1->u().x()*s + ep1->u().y()*c;
	   ep->setU (x, y);

	   ep->vh = vp;
	   vp->e = ep;
//	   if (vlist.tail())
//	    { ep->vt = vlist.tail()->vh;
//	    }
	   vp1 = ep1->vh;
	   x = vp1->q().x()*c - vp1->q().y()*s + px;
	   y = vp1->q().x()*s + vp1->q().y()*c + py;
	   ep->nextVrtx()->setV (x, y);

	   if (poly1.xmin != HUGE_VAL)
	    { if (ep1 == poly1.vlist.head())
	       { xmin = xmax = x;
		 ymin = ymax = y;
	       }
	      else
	       { if (x < xmin)
		  { xmin = x;
		  }
		 else if (x > xmax)
		  { xmax = x;
		  }
		 if (y < ymin)
		  { ymin = y;
		  }
		 else if (y > ymax)
		  { ymax = y;
		  }
	       }
	    }
	   vlist.addTail (ep);
	   ep1 = ep1->next();
	 }
//	if (poly1.numv > 0)
//	 { vlist.head()->vt = vlist.tail()->vh;
//	 }
	if (chardist == -1)
	 { eps = poly1.eps;
	 }
	if (poly1.xmin == HUGE_VAL)
	 { xmin = HUGE_VAL;
	 }
	numv = poly1.numv;
}

#define RX_NO_OVERLAP	0x0
#define RX_OVERLAP	0x1
#define RX_EDGE_WITHIN	0x3
#define RX_POLY_WITHIN	0x5

int RxCvxPolygon2::isectEdge (int *np, RxVecd2 *q1, RxVecd2 *q2, 
			      const RxEdge2 &e1)
{
	double lam[2];
	RxLine2 line;
	double tol;
	tol = getEps();
	int n;

	if ((e1.vt && e1.vt == e1.vh) ||
	    (e1.vt==NULL && e1.next() == NULL))
	 { // degenerate edge; single vertex
	   if (pointIsInside (e1.vh->q()) >= 0)
	    { *q1 = e1.vh->q();
	      *np = 1;
	      return RX_EDGE_WITHIN;
	    }
	   else
	    { *np = 0;
	      return RX_NO_OVERLAP;
	    }
	 }
	else
	 { line.set (e1.prevVrtx()->q(), e1.u());
	   n = intersectLine (lam, line);
	   if (n == 1)
	    { if (lam[0] < -tol || lam[0] > 1+tol)
	       { n = 0;
	       }
	      lam[0] = MAX(0.0,MIN(1.0,lam[0]));
	    }
	   else if (n == 2)
	    { 
	      if (lam[0] > 1+tol || lam[1] < -tol)
	       { n = 0;
	       }
	      else
	       { lam[0] = MAX(0.0,MIN(1.0,lam[0]));
		 lam[1] = MAX(0.0,MIN(1.0,lam[1]));
		 if (ABS(lam[0]-lam[1]) <= 1.0*tol)
		  { n = 1;
		  }
	       }
	    }
	   *np = n;
	   if (n == 0)
	    { return RX_NO_OVERLAP;
	    }
	   else if (n == 1)
	    { q1->extrap (line.q(), lam[0], line.u());
	      return RX_OVERLAP;
	    }
	   else if (n == 2)
	    { q1->extrap (line.q(), lam[0], line.u());
	      q2->extrap (line.q(), lam[1], line.u());
	      if (lam[0] == 0 && lam[1] == 1)
	       { return RX_EDGE_WITHIN;
	       }
	      else if (lam[0] > 0 && lam[1] < 1 && numv == 2)
	       { return RX_POLY_WITHIN;
	       }
	      else
	       { return RX_OVERLAP;
	       }
	    }
	   else
	    { fprintf (stderr,
"RxCvxPolygon2::isectEdge: internal error: n=%d\n", n);
	      exit (1);
	      return -1;
	    }
	 }
}

RxCvxPolygon2* RxCvxPolygon2::intersectLineSeg(RxCvxPolygon2 &poly1, 
					   RxCvxPolygon2 &poly2)
{
	// Special case intersection when poly1 is a 2-edge line segment

	RxVrtx2 *v1, *v2;
	RxEdge2 *e1, *e2;
	RxVecd2 q1, q2;
	int n;
	int code;

	code = poly2.isectEdge (&n, &q1, &q2, *poly1.vlist.head());
	if (n == 1)
	 { e1 = RxEdge2::alloc();
	   v1 = RxVrtx2::alloc();
	   e1->vh = v1;
	   v1->e = e1;
	   v1->setV (q1);
	   e1->setU (0, 0);
	   vlist.addTail (e1);
	   numv = 1;
	 }
	else if (n == 2)
	 { e1 = RxEdge2::allocList(2);
	   v1 = RxVrtx2::allocList(2);
	   e1->vh = v1;
	   v1->e = e1;
	   v1->setV (q1);
	   e2 = e1->link;
	   v2 = v1->link;
	   e2->vh = v2;
	   v2->e = e2;
	   v2->setV (q2);
	   vlist.addTail (e1);
	   e1->u_.sub(v1->q(), v2->q());
	   vlist.addTail (e2);
	   e2->u_.sub(v2->q(), v1->q());
	   numv = 2;
	 }
	if (code == RX_EDGE_WITHIN)
	 { return &poly1;
	 }
	else if (code == RX_POLY_WITHIN)
	 { return &poly2;
	 }
	else
	 { return this;
	 }
}

int doColinear (RxVecd2 *pnts, IsectCtx &ctx)
{
	/* 
	   All the cases for this can be a bit tedious. Here,
	   we do the following. Base coordinates are assigned to
	   the edge associated with vp1, such that the head is
	   at 0 and the tail is at -1. Within this coordinate
	   system, point associated with the edge of vp2 are
	   (a,b), where (a,b) = (t2,h2) if vp1 and vp2 are
	   pointing in the same direction, and (a,b) = (h2,t2)
	   otherwise. la and lb describe the positions at a and
	   b, respectively.

	       vp2:  a--------------->b
		     |   	      |
		     |   	      |
		     V   	      V
	       vp1:            t1----------------->h1
	     
		     la        -1     lb          0
	*/
	RxVecd2 da, db;
	const RxVecd2 *a, *b;
	double la, lb, dot;
	int icnt = 0;

	if (ctx.ep1->u().dot(ctx.ep2->u()) > 0)
	 { a = &ctx.pp2->q();
	   b = &ctx.vp2->q();
	 }
	else
	 { a = &ctx.vp2->q();
	   b = &ctx.pp2->q();
	 }
	da.sub (*a, ctx.vp1->q());
	db.sub (*b, ctx.vp1->q());

	dot = da.dot (ctx.ep1->u());
	if (ABS(dot) <= ctx.tol)
	 { dot = 0;
	 }
	la = dot/ctx.ep1->u().dot(ctx.ep1->u());
	dot = db.dot (ctx.ep1->u());
	if (ABS(dot) <= ctx.tol)
	 { dot = 0;
	 }
	lb = dot/ctx.ep1->u().dot(ctx.ep1->u());

	if (lb >= -1 && la <= 0)
	 { 
	   if (la <= -1)
	    { pnts[icnt++] = ctx.pp1->q();
	    }
	   else if (la < 0)
	    { pnts[icnt++] = *a;
	    }
	   if (lb >= 0)
	    { pnts[icnt++] = ctx.vp1->q();
	    }
	   else if (lb > -1)
	    { pnts[icnt++] = *b;
	    }
	 }
	return (icnt);
}

int isectLineSegs (RxVecd2 *pnts, IsectCtx &ctx, int colinearOK)
{
	RxVecd2 dp;
	double ah2at2;
	double ah1at1;

#if DEBUG
	if (RxCvxPolygon2::debug)
	 { printf ("ah2=%22.16e at2=%22.16e\n", ctx.ah2, ctx.at2);
	   printf ("ah1=%22.16e at1=%22.16e\n", ctx.ah1, ctx.at1);
	 }
#endif
	ctx.colinear = 0;

	if ((ah2at2 = ctx.ah2*ctx.at2) > 0 || (ah1at1 = ctx.ah1*ctx.at1) > 0)
	 { return (0);
	 }
	else if (ah2at2 < 0 && ah1at1 < 0)
	 { pnts[0].extrap (ctx.vp1->q(), ctx.ah1/(ctx.at1-ctx.ah1), ctx.ep1->u());
	   ctx.inflag = ctx.ah1 > 0 ? 'p' : 'q';
	   return (1);
	 }
	else if ((ctx.ah1 == 0 && ctx.at1 == 0) || 
		 (ctx.ah2 == 0 && ctx.at2 == 0))
	 { ctx.colinear = 1;
	   if (colinearOK)
	    { return (doColinear (pnts, ctx));
	    }
	   else
	    { return (0);
	    }
	 }
	else if (ah2at2 < 0)
	 { pnts[0] = (ctx.ah1 == 0 ? ctx.vp1->q() : ctx.pp1->q());
	 }
	else if (ah1at1 < 0)
	 { pnts[0] = (ctx.ah2 == 0 ? ctx.vp2->q() : ctx.pp2->q());
	 }
	else if (ctx.at1 == 0)
	 { if (ctx.at2 == 0)
	    { dp.sub (ctx.pp1->q(), ctx.pp2->q());
	      pnts[0] = (dp.dot(ctx.ep2->u()) >= 0 ? ctx.pp2->q():ctx.pp1->q());
	    }
	   else // ctx.ah2 == 0
	    { dp.sub (ctx.pp1->q(), ctx.vp2->q());
	      pnts[0] = (dp.dot(ctx.ep2->u()) >= 0 ? ctx.vp2->q() : ctx.pp1->q());
	    }
	 }
	else if (ctx.ah1 == 0)
	 { if (ctx.at2 == 0)
	    { dp.sub (ctx.vp1->q(), ctx.pp2->q());
	      pnts[0] = (dp.dot(ctx.ep2->u()) >= 0 ? ctx.pp2->q() : ctx.vp1->q());
	    }
	   else // ctx.ah2 == 0
	    { dp.sub (ctx.vp1->q(), ctx.vp2->q());
	      pnts[0] = (dp.dot(ctx.ep2->u()) >= 0 ? ctx.vp2->q() : ctx.vp1->q());
	    }
	 }
	else
	 { fprintf (stderr, "isectLineSegs: help! we forgot a case!\n");
	   exit (1);
	 }

	if (ctx.ah1 > 0)
	 { if (ah2at2 < 0)
	    { ctx.inflag = 'p';
	    }
	   else if (ctx.at2 == 0)
	    { dp.sub(ctx.vp1->q(),ctx.pp2->q());
	      if (ctx.ep2->prev()->u().cross (dp) > ctx.tol)
	       { ctx.inflag = 'p';
	       }
	    }
	 }
	else if (ctx.ah2 > 0)
	 { if (ah1at1 < 0)
	    { ctx.inflag = 'q';
	    }
	   else if (ctx.at1 == 0)
	    { dp.sub(ctx.vp2->q(),ctx.pp1->q());
	      if (ctx.ep1->prev()->u().cross (dp) > ctx.tol)
	       { ctx.inflag = 'q';
	       }
	    }
	 }
	return (1);
}

RxCvxPolygon2 *RxCvxPolygon2::copyIsect (RxCvxPolygon2 &poly, IsectCtx &ctx)
{
	RxEdge2 *ep, *pp;
	RxVrtx2 *vp;

	ep = ctx.esupply;
	vp = ctx.vsupply;

	if ((pp = poly.vlist.head()))
	 { do
	    { ep->vh = vp;
	      vp->e = ep;
	      vp->setV(pp->nextVrtx()->q());
	      ep->setU(pp->u());
	      vlist.addTail (ep);
// vp->q().print ("copy vp->q() %12.7f\n");
// ep->u().print ("copy ep->u() %12.7f\n");
	      ep = ep->link;
	      vp = vp->link;
	      pp = pp->next();
	    }
	   while (pp != poly.vlist.head());
	 }
	if (chardist == -1)
	 { eps = poly.eps;
	 }
	xmin = poly.xmin;
	xmax = poly.xmax;
	ymin = poly.ymin;
	ymax = poly.ymax;
	numv = poly.numv;

	if (ep)
	 { RxEdge2::freeList (ep);
	 }
	if (vp)
	 { RxVrtx2::freeList (vp);
	 }
	return (&poly);
}

int RxCvxPolygon2::addIsectPoint (const RxVecd2 &p, IsectCtx &ctx)
{
	RxEdge2 *e1, *e2;
	RxVecd2 u2, uh;
	RxVrtx2 *v2;

	if ((e1 = vlist.tail()))
	 { u2.sub (p, e1->vh->q());
	   if (ABS(u2.x()) <= ctx.tol*100 && ABS(u2.y()) <= ctx.tol*100)
	    { return 0;
	    }
	 }
	e2 = ctx.esupply;
	ctx.esupply = e2->link;
	v2 = ctx.vsupply;
	ctx.vsupply = v2->link;
	e2->vh = v2;
	v2->e = e2;
	e2->nextVrtx()->setV (p);
	if (e1)
	 { e2->setU (u2);
	 }
	vlist.addTail (e2);
	numv++;
	return 1;
}

RxCvxPolygon2 *RxCvxPolygon2::finishIsect (IsectCtx &ctx)
{
	/* Complete the Polygon structure describing the intersection
	   polygon. */

	RxEdge2 *ep, *pp;
	RxVrtx2 *vp;

	if (numv == 0)
	 { return (this);
	 }

	ep = vlist.head();
	vp = ep->vh;
	pp = vlist.tail();
	ep->u_.sub (vp->q(), pp->vh->q());

	if (numv > 1)
	 { if (ABS(ep->u().x()) <= ctx.tol*100 && ABS(ep->u().y()) <= ctx.tol*100)
	    { vlist.remove (pp);
	      RxVrtx2::free (pp->vh);
	      RxEdge2::free (pp);
	      pp = vlist.tail();
	      ep->u_.sub (vp->q(), pp->vh->q());
	      numv--;
	    }
	 }
	if (ctx.esupply)
	 { RxEdge2::freeList (ctx.esupply);
	 }
	if (ctx.vsupply)
	 { RxVrtx2::freeList (ctx.vsupply);
	 }
	return (this);
}

inline double withinTol (double x, double tol)
{
	if (x <= tol && x >= -tol)
	 { return (0);
	 }
	else
	 { return (x);
	 }
}

void advanceQ (IsectCtx &ctx)
{
	ctx.pp2 = ctx.vp2;
	ctx.ep2 = ctx.ep2->next();
	ctx.vp2 = ctx.ep2->nextVrtx();

	ctx.dh2.sub (ctx.vp2->q(), ctx.vp1->q());
	ctx.dt2.sub (ctx.pp2->q(), ctx.vp1->q());

	ctx.at2 = ctx.ah2;
	ctx.ah2 = withinTol (ctx.ep1->u().cross(ctx.dh2), ctx.tol);
	ctx.ah1 = withinTol (-ctx.ep2->u().cross(ctx.dh2), ctx.tol);
	ctx.at1 = withinTol (ctx.ah2-ctx.at2+ctx.ah1, ctx.tol);

	ctx.qcnt++;
}

void advanceP (IsectCtx &ctx)
{
	ctx.pp1 = ctx.vp1;
	ctx.ep1 = ctx.ep1->next();
	ctx.vp1 = ctx.ep1->nextVrtx();

	ctx.dh2.sub (ctx.vp2->q(), ctx.vp1->q());
	ctx.dt2.sub (ctx.pp2->q(), ctx.vp1->q());

	ctx.at1 = ctx.ah1;
	ctx.ah1 = withinTol (-ctx.ep2->u().cross(ctx.dh2), ctx.tol);
	ctx.ah2 = withinTol (ctx.ep1->u().cross(ctx.dh2), ctx.tol);
	ctx.at2 = withinTol (ctx.ah2-ctx.at1+ctx.ah1, ctx.tol);

	ctx.pcnt++;
}

#define ADVANCE_Q(s) \
	DEBUG_STATEMENT (printf ("advance Q %s\n", s)) \
	if (ctx.inflag == 'q') \
	 { addIsectPoint (ctx.vp2->q(), ctx); \
	 } \
	advanceQ (ctx); \
	lastAdvance = 'q';

#define ADVANCE_P(s) \
	DEBUG_STATEMENT (printf ("advance P %s\n", s)) \
	if (ctx.inflag == 'p') \
	 { addIsectPoint (ctx.vp1->q(), ctx); \
	 } \
	advanceP (ctx); \
	lastAdvance = 'p';

RxCvxPolygon2 *RxCvxPolygon2::intersect 
	(RxCvxPolygon2 &poly1, RxCvxPolygon2 &poly2,
	 int bboxCheck)
{
	RxVecd2 ipnt;
	int lastAdvance = 0;
	IsectCtx ctx;
	RxCvxPolygon2 *polyi = this;

	clear();	

	DEBUG_STATEMENT (printf ("\n"));

	// Do the bounding box check, if called for

	if (bboxCheck)
	 { if (poly1.xmin == HUGE_VAL)
	    { poly1.calcBoundingBox();
	    }
	   if (poly2.xmin == HUGE_VAL)
	    { poly2.calcBoundingBox();
	    }
	   if (poly1.xmax <= poly2.xmin || poly1.xmin >= poly2.xmax ||
	       poly1.ymax <= poly2.ymin || poly1.ymin >= poly2.ymax)
	    { return (polyi);
	    }
	 }

	if (chardist == -1)
	 { double eps1, eps2;

	   eps1 = poly1.getEps();
	   eps2 = poly2.getEps();
	   eps = MAX(eps1,eps2);
	 }

	// Trap degeneracies here. In theory, the general algorithm should
	// be able to handle degeneracies, but in practice it can only
	// do so with a bunch of extra special case code that reduces
	// the overall efficiency.

	if (poly1.numv == 2)
	 { return (intersectLineSeg (poly1, poly2));
	 }
	else if (poly2.numv == 2)
	 { return (intersectLineSeg (poly2, poly1));
	 }

	ctx.tol = eps;
	ctx.inflag = 0;
	ctx.pcnt = 0;
	ctx.qcnt = 0;
	ctx.esupply = RxEdge2::allocList (poly1.numv + poly2.numv);
	ctx.vsupply = RxVrtx2::allocList (poly1.numv + poly2.numv);

	ctx.ep1 = poly1.vlist.head();
	ctx.ep2 = poly2.vlist.head();
	ctx.vp1 = ctx.ep1->vh;
	ctx.vp2 = ctx.ep2->vh;
	ctx.pp1 = ctx.ep1->prev()->vh;
	ctx.pp2 = ctx.ep2->prev()->vh;

	ctx.dh2.sub (ctx.vp2->q(), ctx.vp1->q());
	ctx.dt2.sub (ctx.pp2->q(), ctx.vp1->q());

	ctx.ah2 = withinTol (ctx.ep1->u().cross(ctx.dh2), ctx.tol);
	ctx.at2 = withinTol (ctx.ep1->u().cross(ctx.dt2), ctx.tol);
	ctx.ah1 = withinTol (-ctx.ep2->u().cross(ctx.dh2), ctx.tol);
	ctx.at1 = withinTol (ctx.ah2-ctx.at2+ctx.ah1, ctx.tol);

	do
	 { double cross;
	   RxVecd2 ddt;
#if DEBUG
	   int oldflag = ctx.inflag;
#endif
	   cross = ctx.ah2-ctx.at2;

	   /* Check for intersection of both segments */

	   if (isectLineSegs (&ipnt, ctx, /*doColinear=*/0))
	    { 
#if DEBUG
	      if (RxCvxPolygon2::debug)
	       { ipnt.print ("Intersection at %g\n");
		 if (oldflag != ctx.inflag)
		  { printf ("inflag set to %c\n", ctx.inflag);
		  }
	       }
#endif
	      if (numv == 0)
	       { ctx.pcnt = ctx.qcnt = 0;
		 ctx.q0 = ipnt;
	       }
	      else if (numv > 1)
	       { if (ABS(ipnt.x()-ctx.q0.x()) <= ctx.tol*100 && 
		     ABS(ipnt.y()-ctx.q0.y()) <= ctx.tol*100)
		  { return (finishIsect (ctx));
		  }
	       }
	      addIsectPoint (ipnt, ctx);
	    }

DEBUG_STATEMENT (printf ("cross=%g\n", cross));
/*  if ((ctx.ah1 == 0 && ctx.at1 == 0) || (ctx.ah2 == 0 && ctx.at2 == 0)) */
	   if (ctx.colinear)
	    { 
	      if (ctx.ep1->u().dot (ctx.ep2->u()) < 0)	// well defined 
	       { 
		 if (lastAdvance == 'p')
		  { ADVANCE_Q ("A");
		  }
		 else
		  { ADVANCE_P ("B");
		  }
	       }
	      else
	       { 
		 if (ctx.ep1->u().dot (ctx.dh2) >= 0)
		  { ADVANCE_P ("AA");
		  }
		 else
		  { ADVANCE_Q ("BB");
		  }
	       }
	    }

	   // These next two rules say: If the head of X lies on the line 
	   // of Y, is behind Y, and X is pointing away from Y, advance Y

	   else if (ctx.ah1 == 0 && cross*ctx.at2 > 0)
	    { ADVANCE_Q ("JJ");
	    }
	   else if (ctx.ah2 == 0 && cross*ctx.at1 < 0)
	    { ADVANCE_P ("KK");
	    }
	   else if (cross >= 0)
	    { if (ctx.ah2 > 0)
	       { ADVANCE_P ("C");
	       }
	      else
	       { ADVANCE_Q ("D");
	       }
	    }
	   else
	    { if (ctx.ah1 > 0)
	       { ADVANCE_Q ("E");
	       }
	      else
	       { ADVANCE_P ("F");
	       }
	    }
	 }
	while ((ctx.pcnt < poly1.numv || ctx.qcnt < poly2.numv) && 
		ctx.pcnt < 2*poly1.numv && ctx.qcnt < 2*poly2.numv);

	if (numv > 0)
	 { return (finishIsect (ctx));
	 }

	/* If we got this far, we know thctx.at either (1) there is
	   no intersection, (2) poly1 is contained in poly2, or
	   (3) poly2 is contained in poly1. */
	 { 
	    /* Contact did not occur. This implies either 
	      there is no intersection, or one polygon is completely
	      inside the other. */

	   RxVecd2 cnt1;	/* centroid of poly1 */
	   RxVecd2 cnt2;	/* centroid of poly2 */
	   int insideTheOther = 0;
	   double a1, a2;	/* polygon areas */

	   poly1.centroid (&cnt1);
	   if (poly2.pointIsInside (cnt1) == 1)
	    { insideTheOther = 1;
	    }
	   else
	    { poly2.centroid (&cnt2);
	      if (poly1.pointIsInside (cnt2) == 1)
	       { insideTheOther = 1;
	       }
	    }
	   if (insideTheOther == 0)
	    { return (this);
	    }

	   a1 = poly1.area();
	   a2 = poly2.area();
	   if (a1 > a2)
	    { return (copyIsect (poly2, ctx));
	    }
	   else
	    { return (copyIsect (poly1, ctx));
	    }
	 }
}

RxCvxPolygon2 *RxCvxPolygon2::intersect (
	RxCvxPolygon2 &poly1, int bboxCheck)
{
	RxCvxPolygon2 polytmp;
	RxCvxPolygon2 *resptr;

	resptr = polytmp.intersect (*this, poly1, bboxCheck);
	clear();
	vlist = polytmp.vlist;
	numv = polytmp.numv;
	polytmp.numv = 0;
	polytmp.vlist.removeAll();
	return (resptr == &polytmp ? this : resptr);
}

RxCvxPolygon2 *RxCvxPolygon2::intersect (
	RxCvxPolygon2 &poly1, const RxLine2 &hp, int bboxCheck)
{
	set (poly1);
	return (intersect (hp, bboxCheck));
}

RxCvxPolygon2 *RxCvxPolygon2::intersect (const RxLine2 &hp, int /*bboxCheck*/)
{
	RxEdge2 *e1, *e2;
	RxVrtx2 *v1, *v2;
	RxVrtx2 *vp;
	RxEdge2 *firstNotIn = NULL;
	RxEdge2 *lastNotIn = NULL;
	int insideVtxExists = 0;
	int side1, side2;
	RxVecd2 qnew;

	if (numv == 0)
	 { return (this);
	 }
	getEps();

	e1 = vlist.head();
	v1 = e1->vh;
	side1 = hp.sideQ (v1->q(), eps);
	do
	 { e2 = e1->next();
	   v2 = e2->vh;

	   if (side1 == 1)
	    { insideVtxExists = 1;
	    }

	   side2 = hp.sideQ (v2->q(), eps);
	   if (side1 == 1 && side2 != 1)
	    { if (side2 == -1)
	       { e2->intersect (&qnew, hp, eps);
		 firstNotIn = RxEdge2::alloc();
		 vp = RxVrtx2::alloc();
		 firstNotIn->vh = vp;
//		 firstNotIn->vt = e1->vh;
		 vp->e = firstNotIn;
		 firstNotIn->nextVrtx()->setV(qnew.x(), qnew.y());
		 firstNotIn->setU (qnew.x()-v1->q().x(), qnew.y()-v1->q().y());
		 vlist.addAfter (e1, firstNotIn);
		 numv++;
	       }
	      else
	       { firstNotIn = e2;
	       }
	    }
	   else if (side1 != 1 && side2 == 1)
	    { if (side1 == -1)
	       { e2->intersect (&qnew, hp, eps);
		 lastNotIn = RxEdge2::alloc();
		 vp = RxVrtx2::alloc();
		 lastNotIn->vh = vp;
//		 lastNotIn->vt = e1->vh;
		 vp->e = lastNotIn;
		 e2->setU (v2->q().x()-qnew.x(), v2->q().y()-qnew.y());
		 lastNotIn->nextVrtx()->setV (qnew);
		 vlist.addAfter (e1, lastNotIn);
		 numv++;
	       }
	      else
	       { lastNotIn = e1;
	       }
	    }
	   e1 = e2;
	   v1 = v2;
	   side1 = side2;
	 }
	while (e1!=vlist.head() && (firstNotIn==NULL||lastNotIn==NULL));

	if (firstNotIn==NULL && lastNotIn==NULL)
	 { if (insideVtxExists == 0)
	    { clear();
	    }
	 }
	else if (firstNotIn && lastNotIn)
	 { if (firstNotIn != lastNotIn)
	    { if (firstNotIn->next() != lastNotIn)
	       { e1 = firstNotIn->next();
		 e2 = lastNotIn->prev();
		 linkNodes (e1, e2);
		 numv -= vlist.remove (e1, e2);
		 RxVrtx2::freeList (e1->vh);
		 RxEdge2::freeList (e1);
      	       }
	      lastNotIn->u_.sub (lastNotIn->vh->q(), firstNotIn->vh->q());
	    }
	 }
	else
	 { fprintf (stderr,
"Internal error: Polytope2D::addHalfPlane: firstNotIn=%x, lastNotIn=%x\n", 
		    firstNotIn, lastNotIn);
	   exit (1);
	 }
	return (this);	
}

int RxCvxPolygon2::intersect (RxEdge2 *er, const RxEdge2 &e1)
{
	RxVecd2 q1, q2;
	int n;

	isectEdge (&n, &q1, &q2, e1);
	if (n == 1)
	 { er->setV (q1);
	 }
	else // n == 2
	 { er->setV (q1, q2);
	 }
	return n;
}

static void updateLam (double *lam, const RxVecd2 p, const RxLine2 &hp)
{
	double l;

	l = hp.project((RxVecd2*)0, p);
	if (l < lam[0])
	 { lam[0] = l;
	 }
	if (l > lam[1])
	 { lam[1] = l;
	 }
}

int RxCvxPolygon2::intersectLine (double *lam, const RxLine2 &hp)
{
	RxEdge2 *e1, *e2;
	RxVrtx2 *v1, *v2;
	RxVecd2 dp;
	RxVecd2 pi;
	double at, ah;
	double tol;

	lam[0] =  HUGE_VAL;
	lam[1] = -HUGE_VAL;

	tol = getEps();

DEBUG_STATEMENT(printf ("tol=%g\n", tol));

	if ((e1 = vlist.head()) == NULL)
	 { return 0;
	 }
	else
	 { v1 = e1->vh;
	   at = hp.u().cross (dp.sub (v1->q(), hp.q()));
	   do
	    { e2 = e1->next();
	      v2 = e2->vh;
	      ah = hp.u().cross (dp.sub (v2->q(), hp.q()));

DEBUG_STATEMENT(printf ("at=%g ah=%g %d %d\n", at, ah, 
		ABS(ah) <= tol, ABS(at) <= tol));
#if 1
	      if (ABS(ah) <= tol || ABS(at) <= tol)
	       { if (ABS(ah) <= tol)
		  { updateLam (lam, v2->q(), hp);
DEBUG_STATEMENT(printf ("ah lam=%g %g\n", lam[0], lam[1]));
		  }
		 if (ABS(at) <= tol)
		  { updateLam (lam, v1->q(), hp);
DEBUG_STATEMENT(printf ("at lam=%g %g\n", lam[0], lam[1]));
		  }
	       }
	      else
	       { if (ah*at < 0)
		  { pi.extrap (v2->q(), -ah/(ah-at), e2->u());
		    updateLam (lam, pi, hp); 
DEBUG_STATEMENT(printf ("xx lam=%g %g\n", lam[0], lam[1]));
		  }
		 else if (lam[0] != HUGE_VAL && lam[0] != lam[1])
		  { // short-cut: we have found a finite interval, and we
		    // didn't add to it this time, so was are done.
DEBUG_STATEMENT(printf ("break\n"));
		    break;
		  }
	       }
#else
	      if (ah*at < 0)
	       { pi.extrap (v2->q(), -ah/(ah-at), e2->u());
		 updateLam (lam, pi, hp); 
	       }
	      if (ah == 0)
	       { updateLam (lam, v2->q(), hp);
	       }
	      if (at == 0)
	       { updateLam (lam, v1->q(), hp);
	       }
	      if (ah*at > 0 && lam[0] != HUGE_VAL && lam[0] != lam[1])
	       { // short-cut: we have found a finite interval, and we 
		 // didn't add to it this time, so was are done.
		 break;
	       }
#endif
	      e1 = e2;
	      v1 = v2;
	      at = ah;
	    }
	   while (e1!=vlist.head());
	 }
	if (lam[0] == HUGE_VAL)
	 { swapDouble (&lam[0], &lam[1]);
	   return (0);
	 }
	else if (lam[0] != lam[1])
	 { return (2);
	 }
	else
	 { return (1);
	 }
}

int RxCvxPolygon2::maxPoint (RxVecd2 *v, const RxVecd2 &dir) const
{
	const RxEdge2 *ep, *np;
	double vdot;
	double ndot;

	if (numv == 0)
	 { return (0);
	 }
	else if (numv == 1)
	 { *v = vlist.head()->vh->q();
	   return 1;
	 }
	else
	 { ep = vlist.head();
	   vdot = ep->u().dot(dir);
	   do
	    { np = ep->next();
	      ndot = np->u().dot(dir);
	      if (vdot >= 0 && ndot <= 0)
	       { *v = ep->vh->q();
		 return 1;
	       }
	      ep = np;
	      vdot = ndot;
	    }
	   while (ep != vlist.head());
	   fprintf (stderr, "RxCvxPolygon2::maxPoint(): internal error\n");
	   exit (1);
	   return -1;
	 }
}

#if 0
#define MAXCOL 80

int RxCvxPolygon2::print (FILE *fp, const char *fmt) const
{
	RxPrintSpec spec;
	RxFmtNode *fnode;
	char *pfmt;
	const RxEdge2 *ep;
	int ntotal = 0;
	int col = 0;
	int n;
	int rwidth;

	spec.set(fmt);
	if (spec.prefix)
	 { if ((n = fprintf (fp, spec.prefix)) < 0)
	    { return n;
	    }
	   ntotal += n;
	 }
	if ((n = fprintf (fp, "%2d:\n", numv)) < 0)
	 { return n;
	 }
	ntotal += n;
	if ((ep = vlist.head()))
	 { if ((fnode = spec.middle.head()))
	    { pfmt = fnode->fmt;
	      rwidth = fnode->width ? 1+2*fnode->width : 0;
	    }
	   else
	    { pfmt = "%g";
	      rwidth = 0;
	    }
	   do
	    { if ((n = fprintf (fp, pfmt, ep->vh->q().x())) < 0)
	       { return n;
	       }
	      putc (' ', fp);
	      col += n+1;
	      ntotal += n+1;
	      if ((n = fprintf (fp, pfmt, ep->vh->q().y())) < 0)
	       { return n;
	       }
	      col += n;
	      ntotal += n;
	      ep = ep->next();
	      if (ep == vlist.head() || col+rwidth+2 > MAXCOL)
	       { putc ('\n', fp);
		 ntotal++;
		 col = 0;
	       }
	      else
	       { putc (' ', fp);
		 putc (' ', fp);
		 ntotal += 2;
		 col += 2;
	       }
	    }
	   while (ep != vlist.head());
	 }
	if (spec.suffix)
	 { if ((n = fprintf (fp, spec.suffix)) < 0)
	    { return n;
	    }
	   ntotal += n;
	 }
	return (ntotal);
}

int RxCvxPolygon2::scan (FILE *fp)
{
	RxEdge2 *ep;
	RxVrtx2 *vp;
	double x0, y0;
	double xp, yp;
	double x, y;
	int n, num;

	if ((n = fscanf (fp, "%d:", &num)) <= 0)
	 { return (n);
	 }
	clear();
	if (num > 0)
	 { ep = RxEdge2::allocList(num);
	   vp = RxVrtx2::allocList(num);
	   for ( ; ep; ep=ep->link, vp=vp->link)
	    { if ((n = fscanf (fp, "%lf%lf", &x, &y)) <= 0)
	       { return (n);
	       }
	      ep->vh = vp;
	      vp->e = ep;
	      vp->setV (x, y);
	      if (!vlist.empty())
	       { ep->setU (x-xp, y-yp);
//		 ep->vt = vlist.tail()->vh;
	       }
	      else
	       { x0 = x;
		 y0 = y;
	       }
	      xp = x;
	      yp = y;
	      vlist.addTail (ep);
	    }
	   vlist.head()->setU(x0-xp, y0-yp);
//	   vlist.head()->vt = vlist.tail()->vh;
	 }
	numv = num;
	return (1);
}

int RxCvxPolygon2::xscan (FILE *fp)
{
	RxEdge2 *ep;
	RxVrtx2 *vp;
	double x0, y0;
	double xp, yp;
	double x, y;
	int n, num;

	clear();
	if ((n = xscanf (fp, "%d:", &num)) <= 0)
	 { return (n);
	 }
	if (num > 0)
	 { ep = RxEdge2::allocList(num);
	   vp = RxVrtx2::allocList(num);
	   for ( ; ep; ep=ep->link, vp=vp->link)
	    { if ((n = xscanf (fp, "%lf", &x)) <= 0)
	       { return (n);
	       }
	      if ((n = xscanf (fp, "%lf", &y)) <= 0)
	       { return (n);
	       }
	      ep->vh = vp;
	      vp->e = ep;
	      vp->setV(x, y);
	      if (!vlist.empty())
	       { ep->setU (x-xp, y-yp);
//		 ep->vt = vlist.tail()->vh;
	       }
	      else
	       { x0 = x;
		 y0 = y;
	       }
	      xp = x;
	      yp = y;
	      vlist.addTail (ep);
	    }
	   vlist.head()->setU(x0-xp, y0-yp);
//	   vlist.head()->vt = vlist.tail()->vh;
	 }
	numv = num;
	return (1);
}

#else

int RxCvxPolygon2::scan (FILE *fp)
{
	RxCinput ci(fp, /*useXscan=*/0);
	return scan (&ci);
}

int RxCvxPolygon2::xscan (FILE *fp)
{
	RxCinput ci(fp, /*useXscan=*/1);
	return scan (&ci);
}

int RxCvxPolygon2::scan (const char *buf, int *cntp)
{
	RxCinput ci(buf, cntp);
	return scan (&ci);
}

istream &RxCvxPolygon2::scan(istream &is)
{
	RxCinput ci(&is);
	scan (&ci);
	return is;
}

int RxCvxPolygon2::scan (RxCinput *ci)
{
	RxParseObj *obj;

	if ((obj = RxParseObj::scan (ci)) == NULL)
	 { return -1;
	 }
	return oscan (obj);
}

int RxCvxPolygon2::oscan (RxParseObj *obj)
{
	RxEdge2 *ep;
	RxVrtx2 *vp;
	double x0, y0;
	double xp, yp;
	double x, y;
	RxParseObj *op;
	int len;

	if ((obj->getType() & ~(RxParseObj::NUMBER)) != RxParseObj::LIST)
	 { return -1;
	 }
	len = obj->getList().length();
	if (len < 0 || len%2 != 0)
	 { return -1;
	 }
	clear();
	if (len > 0)
	 { ep = RxEdge2::allocList(len/2);
	   vp = RxVrtx2::allocList(len/2);
	   op=obj->firstChild();
	   for ( ; ep; ep=ep->link, vp=vp->link)
	    { x = op->getNumberValue();
	      op = obj->nextChild();
	      y = op->getNumberValue();
	      op = obj->nextChild();
	      ep->vh = vp;
	      vp->e = ep;
	      vp->setV (x, y);
	      if (!vlist.empty())
	       { ep->setU (x-xp, y-yp);
	       }
	      else
	       { x0 = x;
		 y0 = y;
	       }
	      xp = x;
	      yp = y;
	      vlist.addTail (ep);
	    }
	   vlist.head()->setU(x0-xp, y0-yp);
	 }
	numv = len/2;
	return 1;
}

int RxCvxPolygon2::print (FILE *fp, const char *s) const
{
	RxCoutput co(fp);
	RxPrintSpec fs(s);
	return cprint (&co, &fs);
}

int RxCvxPolygon2::print (char *buf, const char *s, int n) const
{
	RxCoutput co(buf, n);
	RxPrintSpec fs(s);
	return cprint (&co, &fs);
}

ostream &RxCvxPolygon2::print(ostream &os) const
{
	RxCoutput co(&os);
	RxPrintSpec fs("%g");	// fake print spec -- format will be ignored
	cprint (&co, &fs);	
	return os;
}

#define MAX_COLS	79

int RxCvxPolygon2::cprint (RxCoutput *co, RxPrintSpec *ps) const
{
	const RxEdge2 *ep;
	RxFmtNode *cf0;
	int ncOne = -1;
	int nc = 0;
	int i;

	co->output_str (ps->prefix());
	cf0 = (RxFmtNode*)ps->convNode(0);
	if (cf0 == NULL)
	 { cf0 = new RxFmtNode ("%g", 0, RxFmtNode::CONV);
	 }
	co->output_char ('[');
	nc = 0;
	ep = vlist.head();
	for (i=0; i<numv; i++)
	 { nc += co->output_double (cf0->fmt(), ep->vh->q().x());
	   nc += co->output_str (" ");
	   nc += co->output_double (cf0->fmt(), ep->vh->q().y());
	   if (ncOne == -1)
	    { ncOne = nc + 2;
	    }
	   if (i<numv-1)
	    { if (nc + ncOne > MAX_COLS)
	       { co->output_str ("\n ");
		 nc = 0;
	       }
	      else
	       { co->output_str ("  ");
	       }
	    }
	   ep = ep->next();
	 }
	co->output_char (']');
	co->output_str (ps->suffix());
	if (ps->numConvs() == 0)
	 { delete cf0;
	 }
	return co->nchars();
}

#endif

int RxCvxPolygon2::equal (RxCvxPolygon2 &p2, double prec)
{
	RxEdge2 *ep1;
	const RxEdge2 *ep2;
	RxVecd2 qdel;
	int equal;
	double tol;

	if (numv != p2.numv)
	 { return (0);
	 }
	if (chardist == -1)
	 { double cdist1 = getCharDist();
	   double cdist2 = p2.getCharDist();
	   tol = square(MAX(cdist1, cdist2))*prec;
	 }
	else
	 { tol = chardist*chardist*prec;
	 }
	if ((ep1 = vlist.head()))
	 { 
	   ep2 = p2.vlist.head();
	   do
	    { qdel.sub (ep2->vh->q(), ep1->vh->q());
	      if (qdel.dot(qdel) <= tol)
	       { // found a beginning to second polytope
		 equal = 1;
		 ep1 = ep1->next();
		 ep2 = ep2->next();
		 while (ep1 != vlist.head())
		  { qdel.sub (ep2->vh->q(), ep1->vh->q());
		    if (qdel.dot(qdel) > tol)
		     { equal = 0;
		       break;
		     }
		    ep1 = ep1->next();
		    ep2 = ep2->next();
		  }
		 return (equal);
	       }
	      ep2 = ep2->next();
	    }
	   while (ep2 != p2.vlist.head());
	   return (0);
	 }
	return (1);
}
