#ifndef _RXCVXPOLYGON2_H_WRAP_
#define _RXCVXPOLYGON2_H_WRAP_

#include <stdio.h>   
#include <RxClist.h>
#include <RxVecd2.h>
#include <RxLine2.h>
#include <stdlib.h>

#define RX_POLYTOPE2_EPS (1e-13)

class RxEdge2;

class RxVrtx2
{
	friend class RxCvxPolygon2;
	friend class RxEdge2;

private:
	static RxVrtx2 *cache;  // private cache for fast allocation
	RxVrtx2 *link;		// link used for private cache

	RxVecd2 q_;
	RxEdge2 *e;

	RxVrtx2& setV(double qx, double qy)
	 { q_.x() = qx;
	   q_.y() = qy;
	   return *this;
	 }
	RxVrtx2& setV(const RxVecd2 &q)
	 { q_ = q;
	   return *this;
	 }
public:
	// access

	const RxVecd2 &q() const
	 { return q_;
	 }

	RxVrtx2 ()
	 { e = NULL;
	 }
	RxEdge2 *nextEdge();
	RxEdge2 *prevEdge();

	// Private allocation and deallocation routines.

	static RxVrtx2 *alloc();
	static RxVrtx2 *allocList(int n);
	static void free (RxVrtx2 *vp);
	static void freeList (RxVrtx2 *vp);
};

class RxEdge2 : public RxCnodeBase
{
	/* Each polytope is described by a double circular linked
	   list of these elements which describe the vertices and
	   edges.
	 */
	friend class RxCvxPolygon2;
private:
	static RxEdge2 *cache;	// private cache for fast allocation

	RxVecd2 u_;	// vector in the direction of the line
	RxVrtx2 *vh;	// head vertex
	RxVrtx2 *vt;	// tail vertex; NULL unless the edge is a "free" edge

public:
	RxEdge2 *link;		// link used for private cache

	// access 

	const RxVecd2 &u() const
	 { return u_;
	 }
	RxEdge2& setU(double ux, double uy)
	 { u_.x() = ux;
	   u_.y() = uy;
	   return *this;
	 }
	RxEdge2& setU(const RxVecd2 &u)
	 { u_ = u;
	   return *this;
	 }
	RxEdge2& setV(const RxVecd2 &qh, const RxVecd2 &qt);
	RxEdge2& setV(double xh, double yh, double xt, double yt);
	RxEdge2& setV(const RxVecd2 &qh);
	RxEdge2& setV(double xh, double yh);

	RxEdge2()
	 { vt = NULL;
	   vh = NULL;
	 }
	RxEdge2(const RxVecd2 &v1, const RxVecd2 &v2)
	 { setV (v1, v2);
	 }
	~RxEdge2();

	double intersect (RxVecd2 *p, const RxLine2 &l2, 
			  double tol=RX_POLYTOPE2_EPS) const;

	// data structure reference

	RxEdge2 *nextEdge()
	 { return next();
	 }
	RxEdge2 *prevEdge()
	 { return prev();
	 }
	RxVrtx2 *nextVrtx() const
	 { return vh;
	 }
	RxVrtx2 *prevVrtx() const
	 { return vt ? vt : prev()->vh;
	 }

	// Private allocation and deallocation routines.

	static RxEdge2 *alloc();
	static RxEdge2 *allocList(int n);
	static void free (RxEdge2 *vp);
	static void freeList (RxEdge2 *vp);

	RxEdge2& operator=(const RxEdge2)
	 { printf ("bail!\n");
	   exit (1);
	   return (*this);
	 }

	RX_CNODE_DECLS (RxEdge2);	
};

typedef RxClist<RxEdge2> RxEdgeList;

typedef struct {
	/* When intersecting two polytopes, information and results
	   of geometric computations are kept in this structure. */

	RxVrtx2 *vsupply;	// list of vertices for building new polytope.
	RxEdge2 *esupply;	// list of edges for building new polytope.
	RxEdge2 *ep1;		// current edge on poly1 
	RxEdge2 *ep2;		// current edge on poly2
	RxVrtx2 *pp1, *vp1;	// last and current vertices on poly1 
	RxVrtx2 *pp2, *vp2;	// last and current vertices on poly2 
	RxVecd2 q0;		// first intersection point
	int inflag;		// which polytope is currently inside 
	int pcnt;		
	int qcnt;	
	double ah1;
	double ah2;
	double at1;
	double at2;
	double cross;
	int colinear;
	RxVecd2 dh2;
	RxVecd2 dt2;
	double tol;
} IsectCtx;

class RxCvxPolygon2
{
	friend class RxFace3;
public:
	// eps is the tolerance used for area calculations, and
	// based either on the bounding box size or on a supplied
	// chardist. Once chardist is set by the user, it remains
	// in effect for all operations involving that polygon,
	// until reset.

	double eps;
	double chardist;

	RxEdgeList vlist;
	RxEdge2 *curEdge;

	double xmin, xmax;
	double ymin, ymax;

	int numv;

	int addIsectPoint (const RxVecd2 &p, IsectCtx &ctx);
	RxCvxPolygon2 *finishIsect (IsectCtx &ctx);
	RxCvxPolygon2 *copyIsect (RxCvxPolygon2 &poly, IsectCtx &ctx);

	void calcBoundingBox();

	RxCvxPolygon2 *intersectLineSeg(RxCvxPolygon2 &poly1, RxCvxPolygon2 &poly2);
	void linkNodes (RxEdge2 *ep0, RxEdge2 *ep1);
	int isectEdge (int *np, RxVecd2 *v1, RxVecd2 *v2, const RxEdge2 &e1);

public:
	static int debug;
	static void setDebug(int on);

	void init();
	int consistencyCheck(int printInfo);

	int setVertices (double *vtxs, int n);
	int set (const RxCvxPolygon2 &poly);
	int setRect (double x, double y, double w, double h);
	int setBegin (int n=0);
	RxVrtx2 *addVrtx (double x, double y);
	int setEnd();

	int getVertices (double *vtxs) const;

	RxCvxPolygon2 ()
	 { init();
         }
	RxCvxPolygon2 (double *vtxs, int n)
	 { init();
	   setVertices (vtxs, n);
	 }
	RxCvxPolygon2 (const RxCvxPolygon2 &poly)
	 { init();
	   set (poly);
	 }
	RxCvxPolygon2 (double x, double y, double w, double h)
	 { init();
	   setRect (x, y, w, h);
	 }

	~RxCvxPolygon2 ()
	 { clear();
	 }
	void clear ();

	RxEdge2 *headEdge()
	 { return (vlist.head());
	 }
	RxEdge2 *tailEdge()
	 { return (vlist.tail());
	 }
	const RxEdge2 *headEdge() const
	 { return (vlist.head());
	 }
	const RxEdge2 *tailEdge() const
	 { return (vlist.tail());
	 }

	double getCharDist();
	double getEps();
	void setCharDist(double cdist);

	void shiftVtxs (int n);

	void xform (double px, double py, double ang);
	void xform (const RxCvxPolygon2 &poly1, 
		    double px, double py, double ang);

	RxCvxPolygon2 *intersect (RxCvxPolygon2 &poly1, 
		RxCvxPolygon2 &poly2, int bboxCheck=0);
	RxCvxPolygon2 *intersect (
		RxCvxPolygon2 &poly1, int bboxCheck=0);

	RxCvxPolygon2 *intersect (RxCvxPolygon2 &poly1,
		const RxLine2 &hp, int bboxCheck=0);
	RxCvxPolygon2 *intersect (const RxLine2 &hp, int bboxCheck=0);

	int intersect (RxEdge2 *er, const RxEdge2 &e1);
	int intersect (RxEdge2 *er)
	 { return intersect (er, *er);
	 }

	int intersectLine (double *isec, const RxLine2 &hp);

	int maxPoint (RxVecd2 *v, const RxVecd2 &dir) const;
	int numVtxs() const
	 { return numv;
	 }

	void centroid (RxVecd2 *v) const;
	double area () const;
	int pointIsInside (const RxVecd2 &v);
	
	int scan (FILE *fp=stdin);
	int xscan (FILE *fp);
	int scan (const char *buf, int *cntp=NULL);
	istream &scan(istream &is);
	int scan (class RxCinput *ci);
	virtual int oscan (class RxParseObj *obj);

	int print (FILE *fp, const char *fmt) const;
	int print (const char *fmt) const
	 { return print (stdout, fmt);
	 }
	int print (char *buf, const char *fmt, int n=0) const;
	ostream &print(ostream &os) const;
	virtual int cprint (class RxCoutput *co, class RxPrintSpec *fs) const;

	int equal (RxCvxPolygon2 &p2, double prec);
};

inline RxEdge2 *RxVrtx2::nextEdge()
{
	return (e);
}

inline RxEdge2 *RxVrtx2::prevEdge()
{
	return (e->prev());
}

inline ostream &operator<<(ostream &os, const RxCvxPolygon2 &poly)
{
	return poly.print(os);
}

inline istream &operator>>(istream &is, RxCvxPolygon2 &poly)
{
	return poly.scan(is);
}

#endif
