#include <RxLine2.h>
#include <RxPrintSpec.h>
#include <cdefs.h>
#include <math.h>

const double RxLine2::EPS = 1e-13;

RxLine2 &RxLine2::set (const RxVecd2 &q1, const RxVecd2 &u1)
{
#if NORMALIZE_U
	double magSqr;

	magSqr = u1.dotQ(u1);
	if (ABS(magSqr-1) > EPS)	// normalize if necessary
	 { u_.scaleQ (1/sqrt(magSqr), u1);
	 }
	else
#endif
	 { u_ = u1;
	 }
	q_ = q1;
	ulen = -1;	
	return (*this);
}

RxLine2 &RxLine2::setPlucker (double x1, const RxVecd2 &u1)
{
#if NORMALIZE_U
	double magSqr;

	magSqr = u1.dotQ(u1);
	if (ABS(magSqr-1) > EPS)	// normalize if necessary
	 { u_.scaleQ (1/sqrt(magSqr), u1);
	 }
	else
#endif
	 { u_ = u1;
	 }
	q_.x() =  u1.y()*x1;
	q_.y() = -u1.x()*x1;
	ulen = -1;	
	return (*this);
}

void RxLine2::getPlucker (double *x1, RxVecd2 *u1)
{
	if (u1)
	 { *u1 = u_;
	 }
	if (x1)
	 { *x1 = q_.cross(u_);
	 }
}

RxLine2& RxLine2::setHP (const RxVecd2 &a, double b)
{
	u_.x() =  a.y();
	u_.y() = -a.x();

	double magSqr = u_.dotQ(u_);
#if NORMALIZE_U
	if (ABS(magSqr-1) > EPS)	// normalize if necessary
	 { u_.scaleQ (1/sqrt(magSqr), u_);
	 }
#endif
	q_.scaleQ (b/magSqr, a);
	ulen = -1;	
	return (*this);
}	

void RxLine2::getHP (RxVecd2 *a, double *b)
{
	if (a)
	 { a->y() = u_.x();
	   a->x() = -u_.y();
	 }
	if (b)
	 { *b = q_.cross(u_);
	 }
}

#if 0
double RxLine2::lineto (RxLine2 *lr, const RxVecd2 &p) const
{
	double l;

	l = project (&lr->q_, p);
	lr->u_.x() = -u_.y();
	lr->u_.y() =  u_.x();
	return (l);
}
#endif

double RxLine2::sdist (const RxVecd2 &p)
{
#if NORMALIZE_U
	return (-u_.x()*(p.y()-q_.y()) + u_.y()*(p.x()-q_.x()));
#else
	if (ulen == -1)
	 { ulen = u_.norm2();
	 }
	return (-u_.x()*(p.y()-q_.y()) + u_.y()*(p.x()-q_.x()))/ulen;
#endif
}

double RxLine2::sdist (RxVecd2 *pr, const RxVecd2 &p1)
{
	double delx, dely;
	double d;

	delx = p1.x()-q_.x();
	dely = p1.y()-q_.y();
#if NORMALIZE_U
	d = -u_.x()*dely + u_.y()*delx;
#else
	d = (-u_.x()*dely + u_.y()*delx)/u().norm2();
#endif
	if (pr)
	 { pr->x() = q_.x() + delx - d*u_.y();
	   pr->y() = q_.y() + dely + d*u_.x();
	 }
	return d;
}

double RxLine2::project (RxVecd2 *pr, const RxVecd2 &p1) const
{
	double delx, dely, l;

	delx = p1.x()-q_.x();
	dely = p1.y()-q_.y();
#if NORMALIZE_U
	l = delx*u_.x()+dely*u_.y();
#else
	l = (delx*u_.x()+dely*u_.y())/u().dot(u());
#endif
	if (pr)
	 { pr->x() = q_.x() + l*u_.x();
	   pr->y() = q_.y() + l*u_.y();
	 }
	return l;
}

int RxLine2::side (const RxVecd2 &p, double tol) const
{
	double x = u_.x()*(p.y()-q_.y()) - u_.y()*(p.x()-q_.x());
	if (x > tol)
	 { return 1;
	 }
	else if (x < -tol)
	 { return -1;
	 }
	else
	 { return 0;
	 }
}

double RxLine2::intersect (RxVecd2 *p, const RxLine2 &l2, double tol) const
{
	/*
	   Suppose we have two line segments, given by

		v = q1 + lam1 u1   and    v = q2 + lam2 u2

	   where q1, u1, q2, and u2 are 2 vectors. In general, finding 
	   the intersection of these two points amounts to solving

		[ u1x  -u2x ] [ lam1 ]
		[           ] [      ] = q2 - q1 = del;
		[ u1y  -u2y ] [ lam2 ] 

	   for which we get

		       -u2y delx + u2x dely    u2 X del
		lam1 = --------------------  = -------
			 u2x u1y - u1x u2y     u2 X u1

		       -u1y delx + u1x dely    u1 X del
		lam2 = --------------------  = -------
			 u2x u1y - u1x u2y     u2 X u1
	*/

	RxVecd2 del;
	double x, l;

	x = l2.u_.cross(u_);
	if (ABS(x) <= tol)
	 { // then the lines are parallel
	   if (p)
	    { del.sub(l2.q_,q_);
	      l = del.x()*u_.x()+del.y()*u_.y();
	      p->scaleQ (l, u_);
	      p->addQ(*p, q_);
	    }
	   return (HUGE_VAL);
	 }
	else
	 { del.sub(l2.q_,q_);
	   l = l2.u_.cross(del)/x;
	   if (p)
	    { p->scaleQ (l, u_);
	      p->addQ(*p, q_);
	    }
	   return (l);
	 }
}

int RxLine2::scan(FILE *fp)
{
	int stat;

	if ((stat = q_.scan(fp)) != 1 ||
	    (stat = u_.scan(fp)) != 1)
	 { return (stat);
	 }
#if NORMALIZE_U
	u_.unit();
#endif
	return (1);
}

int RxLine2::xscan(FILE *fp)
{
	int stat;

	if ((stat = q_.xscan(fp)) != 1 ||
	    (stat = u_.xscan(fp)) != 1)
	 { return (stat);
	 }
#if NORMALIZE_U
	u_.unit();
#endif
	return (1);
}

int RxLine2::scan (const char *buf, int *cntp)
{
	int stat;
	int cnt = 0;
	int idx;

	if ((stat = q_.scan(&buf[cnt], &idx)) != 1)
	 { return (stat);
	 }
	cnt += idx;
	if ((stat = u_.scan(&buf[cnt], &idx)) != 1)
	 { return (stat);
	 }
#if NORMALIZE_U
	u_.unit();
#endif
	if (cntp)
	 { *cntp = cnt;
	 }
	return 1;	
}

int RxLine2::print (FILE *fp, const char *fmt) const
{
	char defaultFmt[32] = "%g";
	RxPrintSpec spec;
	const RxFmtNode *fnode;
	const char *pfmt = defaultFmt;
	const char *ufmt = defaultFmt;
	int ntotal = 0;
	int n;

	spec.set(fmt);
	if (spec.prefix())
	 { if ((n = fprintf (fp, spec.prefix())) < 0)
	    { return n;
	    }
	   ntotal += n;
	 }
	if ((fnode = spec.convNode(0)))
	 { pfmt = fnode->fmt();
	   if (fnode->next())
	    { ufmt = fnode->next()->fmt();
	    }
	   else
	    { strcpy (defaultFmt, pfmt);
	    }
	 }
	if ((n = q_.print (fp, pfmt)) < 0)
	 { return n;
	 }
	ntotal += n;
	if (ufmt == defaultFmt)
	 { putc (' ', fp);
	   ntotal += 1;
	 }
	if ((n = u_.print (fp, ufmt)) < 0)
	 { return n;
	 }
	ntotal += n;
	if (spec.suffix())
	 { if ((n = fprintf (fp, spec.suffix())) < 0)
	    { return n;
	    }
	   ntotal += n;
	 }
	return (ntotal);	
}

int RxLine2::print (char *buf, const char *fmt) const
{
	char defaultFmt[32] = "%g";
	RxPrintSpec spec;
	const RxFmtNode *fnode;
	const char *pfmt = defaultFmt;
	const char *ufmt = defaultFmt;
	int ntotal = 0;
	int n;

	spec.set(fmt);
	if (spec.prefix())
	 { if ((n = sprintf (&buf[ntotal], spec.prefix())) < 0)
	    { return n;
	    }
	   ntotal += n;
	 }
	if ((fnode = spec.convNode(0)))
	 { pfmt = fnode->fmt();
	   if (fnode->next())
	    { ufmt = fnode->next()->fmt();
	    }
	   else
	    { strcpy (defaultFmt, pfmt);
	    }
	 }
	if ((n = q_.print (&buf[ntotal], pfmt)) < 0)
	 { return n;
	 }
	ntotal += n;
	if (ufmt == defaultFmt)
	 { buf[ntotal] = ' ';
	   ntotal += 1;
	 }
	if ((n = u_.print (&buf[ntotal], ufmt)) < 0)
	 { return n;
	 }
	ntotal += n;
	if (spec.suffix())
	 { if ((n = sprintf (&buf[ntotal], spec.suffix())) < 0)
	    { return n;
	    }
	   ntotal += n;
	 }
	return (ntotal);	
}

ostream &RxLine2::print(ostream &os) const
{
	os << q_ << "  " << u_;
	return os;
}

istream &RxLine2::read(istream &is)
{
	is >> q_ >> u_;
#if NORMALIZE_U
	u_.unit();
#endif
	return is;
}

int RxLine2::equal (const RxLine2 &l1, double tol) const
{
	double x0, x1;

	x0 = q_.cross(u_);
	x1 = l1.q_.cross (l1.u_);
	return (ABS(x0-x1) <= tol && u_.equal(l1.u_, tol));
}
