#include <RxSpatialInertia.h>
#include <RxMathError.h>
#include <RxMatdLu.h>
#include <stdarg.h>
#include <math.h>
#include <cdefs.h>
#include <RxParseObj.h>
#include <RxCio.h>
#include <RxSysError.h>

#define SMALL	(1e-12)

const RxSpatialInertia RxSpatialInertia::UNIT (RxMatd3x3(1,0,0,0,1,0,0,0,1), 1);
const RxSpatialInertia RxSpatialInertia::ZERO (RxMatd3x3(0,0,0,0,0,0,0,0,0), 0);

void RxSpatialInertia::doInit()
{
	A_.setZero();
	B_.setZero();
	C_.setZero();
	D_.setZero();
}

void RxSpatialInertia::doZeros()
{
	A_.xx() = 0;
	A_.yy() = 0;
	A_.zz() = 0;
	D_.xx() = 0;
	D_.yy() = 0;
	D_.zz() = 0;
	B_.xy() = 0;
	B_.xz() = 0;
	B_.yx() = 0;
	B_.yz() = 0;
	B_.zx() = 0;
	B_.zy() = 0;
}

void RxSpatialInertia::setMass (double m)
{
	if (m == 0)
	 { setZero ();
	 }
	updateMass (m);
}

void RxSpatialInertia::setCom (const RxVecd3 &com)
{
	double mcx, mcy, mcz;
	com_.set (com);

	mcx = B_.xx()*com_.x();
	mcy = B_.xx()*com_.y();
	mcz = B_.xx()*com_.z();

	A_.xy() =  mcz;
	A_.xz() = -mcy;
	A_.yx() = -mcz;
	A_.yz() =  mcx;
	A_.zx() =  mcy;
	A_.zy() = -mcx;

	D_.xy() = -mcz;
	D_.xz() =  mcy;
	D_.yx() =  mcz;
	D_.yz() = -mcx;
	D_.zx() = -mcy;
	D_.zy() =  mcx;
}

RxSpatialInertia& RxSpatialInertia::set (const RxMatd3x3& I, double m)
{
	if (m == 0)
	 { setZero();
	   return *this;
	 }
	C_ = I;
	com_.setZero();
	updateMass (m);
	return (*this);
}

RxSpatialInertia& RxSpatialInertia::set (const RxMatd3x3& I, double m, const RxVecd3 &com)
{
	if (m == 0)
	 { setZero();
	   return *this;
	 }
	C_ = I;
	com_.set(com);
	updateMass (m);
	return (*this);
}

RxSpatialInertia &RxSpatialInertia::add (const RxSpatialInertia &m1, 
					 const RxSpatialInertia &m2)
{
	double m;

	if ((m = m1.getMass() + m2.getMass()) == 0)
	 { setZero();
	   return *this;
	 }
	C_.add (m1.C_, m2.C_);
	com_.x() = (m1.A_.yz() + m2.A_.yz())/m;
	com_.y() = (m1.A_.zx() + m2.A_.zx())/m;
	com_.z() = (m1.A_.xy() + m2.A_.xy())/m;
	updateMass (m);
	return *this;
}

RxSpatialInertia operator+(const RxSpatialInertia &m1, const RxSpatialInertia &m2)
{
	RxSpatialInertia mr;
	return mr.add(m1,m2);
}

RxSpatialInertia &RxSpatialInertia::sub (const RxSpatialInertia &m1, 
					 const RxSpatialInertia &m2)
{
	double m;

	if ((m = m1.getMass() - m2.getMass()) == 0)
	 { setZero();
	   return *this;
	 }
	C_.sub (m1.C_, m2.C_);
	com_.x() = (m1.A_.yz() - m2.A_.yz())/m;
	com_.y() = (m1.A_.zx() - m2.A_.zx())/m;
	com_.z() = (m1.A_.xy() - m2.A_.xy())/m;
	updateMass (m);
	return *this;
}

RxSpatialInertia operator-(const RxSpatialInertia &m1, const RxSpatialInertia &m2)
{
	RxSpatialInertia mr;
	return mr.sub(m1,m2);
}

RxSpatialInertia& RxSpatialInertia::scale (double a, const RxSpatialInertia &m1)
{
	double m;

	if ((m = a*m1.getMass()) == 0)
	 { setZero();
	   return *this;
	 }
	C_.scale (a, m1.C_);
	com_.set (m1.com_);
	updateMass (m);
	return *this;
}

RxSpatialInertia& RxSpatialInertia::scale (double a)
{
	double m;

	if ((m = a*getMass()) == 0)
	 { setZero();
	   return *this;
	 }
	C_.scale (a);
	updateMass (m);
	return *this;
}

RxSpatialInertia operator*(const RxSpatialInertia &m1, double a)
{
	RxSpatialInertia mr;
	return mr.scale (a, m1);
}

RxSpatialInertia operator*(double a, const RxSpatialInertia &m1)
{
	RxSpatialInertia mr;
	return mr.scale (a, m1);
}

RxSpatialInertia operator/(const RxSpatialInertia &m1, double a)
{
	RxSpatialInertia mr;
	return mr.scale (1/a, m1);
}

#define XX	d.x
#define YY	d.y
#define ZZ	d.z
#define XY	o.z
#define XZ	o.y
#define YZ	o.x
#define YX	XY
#define ZX	XZ
#define ZY	YZ

RxSpatialInertia& RxSpatialInertia::xform (const RxSpatialInertia &m1, const RxRot& R)
{
	//
	// transform M into the frame pointed to by R
	//
	double m;

	m = m1.getMass();
	C_.xform (m1.C_, R);
	com_.xform (m1.com_, R);
	updateMass (m);
	return *this;
}

RxSpatialInertia& RxSpatialInertia::Ixform (const RxSpatialInertia &m1, const RxRot& R)
{
	double m;

	if ((m = m1.getMass()) == 0)
	 { setZero();
	   return *this;
	 }
	C_.Ixform (m1.C_, R);
	com_.Ixform (m1.com_, R);
	updateMass (m);
	return *this;
}

RxSpatialInertia& RxSpatialInertia::xform (const RxSpatialInertia &m1, const RxTrsf& T)
{
	// 
	// transform M into the frame pointed to by T
	//
	RxVecd3 como;	// old mpc value, rotated into frame pointed to by T
	double cxx, cxy, cxz;	// correction factors for I
	double cyy, cyz, czz;
	double m;

	if ((m = m1.getMass()) == 0)
	 { setZero();
	   return *this;
	 }
	
	como.xform (m1.com_, T.R);
	com_.xform (m1.com_, T);
	C_.xform (m1.C_, T.R);

	cxx = (-como.z()*como.z() - como.y()*como.y() + 
		com_.z()*com_.z() + com_.y()*com_.y()) * m;
	cyy = (-como.z()*como.z() - como.x()*como.x() + 
		com_.z()*com_.z() + com_.x()*com_.x()) * m;
	czz = (-como.y()*como.y() - como.x()*como.x() + 
		com_.y()*com_.y() + com_.x()*com_.x()) * m;
	cxy =  (como.x()*como.y() - com_.x()*com_.y()) * m;
	cxz =  (como.x()*como.z() - com_.x()*com_.z()) * m;
	cyz =  (como.y()*como.z() - com_.y()*com_.z()) * m;

	C_.xx() += cxx;
	C_.xy() += cxy;
	C_.xz() += cxz;

	C_.yx() += cxy;
	C_.yy() += cyy;
	C_.yz() += cyz;

	C_.zx() += cxz;
	C_.zy() += cyz;
	C_.zz() += czz;

	updateMass (m);
	return *this;
}

RxSpatialInertia& RxSpatialInertia::Ixform (const RxSpatialInertia &m1, const RxTrsf& T)
{
	// 					     -1
	// transform M into the frame pointed to by T
	//
	RxVecd3 como;	// old mpc value, rotated into frame pointed to by T
	double cxx, cxy, cxz;	// correction factors for I
	double cyy, cyz, czz;
	double m;

	if ((m = m1.getMass()) == 0)
	 { setZero();
	   return *this;
	 }

	como.Ixform (m1.com_, T.R);
	com_.Ixform (m1.com_, T);
	C_.Ixform (m1.C_, T.R);

	cxx = (-como.z()*como.z() - como.y()*como.y() + 
		com_.z()*com_.z() + com_.y()*com_.y()) * m;
	cyy = (-como.z()*como.z() - como.x()*como.x() + 
		com_.z()*com_.z() + com_.x()*com_.x()) * m;
	czz = (-como.y()*como.y() - como.x()*como.x() + 
		com_.y()*com_.y() + com_.x()*com_.x()) * m;
	cxy =  (como.x()*como.y() - com_.x()*com_.y()) * m;
	cxz =  (como.x()*como.z() - com_.x()*com_.z()) * m;
	cyz =  (como.y()*como.z() - com_.y()*com_.z()) * m;

	C_.xx() += cxx;
	C_.xy() += cxy;
	C_.xz() += cxz;

	C_.yx() += cxy;
	C_.yy() += cyy;
	C_.yz() += cyz;

	C_.zx() += cxz;
	C_.zy() += cyz;
	C_.zz() += czz;

	updateMass (m);
	return *this;
}


RxSpatialVec &RxSpatialInertia::Vmul (RxSpatialVec *vr, const RxSpatialVec &v1) const
{
	RxVecd3 a, b, c;
	double m = getMass();

	a.subQ (v1.b(), c.cross (com_, v1.a()));
	a.scaleQ (m, a);
	C_.mulV (&b, v1.a());
	c.cross (com_, v1.b());
	b.addQ (b, c.scaleQ (m, c));
	
	vr->a() = a;
	vr->b() = b;
	return (*vr);
}

RxSpatialVec &RxSpatialInertia::mulV (RxSpatialVec *vr, const RxSpatialVec &v1) const
{
	RxVecd3 a, b, c;
	double m = getMass();

	a.subQ (v1.b(), c.cross (com_, v1.a()));
	a.scaleQ (m, a);
	C_.mulV (&b, v1.a());
	c.cross (com_, v1.b());
	b.addQ (b, c.scaleQ(m, c));
	
	vr->a() = a;
	vr->b() = b;
	return (*vr);
}

RxSpatialInertia &RxSpatialInertia::neg(const RxSpatialInertia &m1)
{
	scale (-1, m1);
	return (*this);
}

RxSpatialInertia &RxSpatialInertia::setUnit()
{
	C_.setIdent();
	com_.setZero();
	updateMass (1);
	return *this;
}

int RxSpatialInertia::oscan (RxParseObj *obj)
{
	RxParseObj *lobj;
	int len;
	double m;

	len = obj->getListLength();
	if (len == 3)
	 { RxParseObj *mobj;
	   RxParseObj *cobj;
	   RxParseObj *iobj;
	   RxVecd3 vc;
	   RxVecd3 vo;


	   mobj = obj->firstChild();
	   cobj = obj->nextChild();
	   iobj = obj->nextChild();

	   if (!mobj->isNumber() ||
	       !cobj->isNumberList() || cobj->getListLength() != 3 ||
	       !iobj->isNumberList() || iobj->getListLength() != 6)
	    { RxSysError (RxSysError::ParsingError,
"RxSpatialInertia::oscan: illegal object format\n").post();
	      return 0;
	    }
	   m = mobj->getNumberValue();
	   if (m == 0)
	    { setZero();
	    }
	   else
	    { cobj->getNumberListValues (vc.buf(), 3);
	      com_.set (vc);
	      lobj = iobj->firstChild();
	      iobj->getNumberListValues (vc.buf(), 3, &lobj);
	      iobj->getNumberListValues (vo.buf(), 3, &lobj);
	      C_.setSymm (vc, vo);
	      updateMass (m);
	    }
	 }
	else if (len == 36)
	 { double buf[36];
	   if (!obj->isNumberList())
	    { RxSysError (RxSysError::ParsingError,
"RxSpatialInertia::oscan: illegal object format\n").post();
	      return 0;
	    }
	   obj->getNumberListValues (buf, 36);
	   m = buf[0*6 + 3];
	   if (m == 0)
	    { setZero();
	    }
	   else
	    { com_.z() =  buf[0*6 + 1]/m;
	      com_.y() = -buf[0*6 + 2]/m;
	      com_.x() =  buf[1*6 + 2]/m;

	      C_(0,0)  =  buf[3*6 + 0];
	      C_(1,1)  =  buf[4*6 + 1];
	      C_(2,2)  =  buf[5*6 + 2];
	      C_(0,1)  =  buf[3*6 + 1];
	      C_(0,2)  =  buf[3*6 + 2];
	      C_(1,2)  =  buf[4*6 + 2];
	      C_(1,0)  =  C_(0,1);
	      C_(2,0)  =  C_(0,2);
	      C_(2,1)  =  C_(1,2);
	      updateMass (m);
	    }
	 }
	return 1;
}

int RxSpatialInertia::cprint (RxCoutput *co, RxPrintSpec *ps) const
{
	if (ps->findSpecial ("%M"))
	 { // then just print as a matrix
	   return RxMatdBase::cprint (co, ps);
	 }
	else
	 { const RxFmtNode *cf0;
	   const char *fmt = "%8.3f";	// default format
	   int width = 8;
	   RxVecd3 vc;
	   RxVecd3 vo;
	   int lspc;

	   lspc = ps->getLeadSpacing();
	   // figure out format strings
	   cf0 = ps->convNode(0);
	   if (cf0)
	    { fmt = cf0->fmt();
	      width = cf0->width();
	    }
	   co->output_str (ps->prefix());
	   co->output_str ("[ ");
	   co->output_double (fmt, getMass());

	   co->output_str (" [ ");
	   co->output_doubleArray (fmt, NULL, NULL, com_.buf(), 3);
	   co->output_str (" ]");
	   co->output_newline (lspc+2);
	   C_.getSymm (&vc, &vo);
	   co->output_str ("[ ");
	   co->output_doubleArray (fmt, NULL, NULL, vc.buf(), 3);
	   co->output_char (' ');
	   co->output_doubleArray (fmt, NULL, NULL, vo.buf(), 3);
	   co->output_str (" ] ]");

	   if (ps->numConvs() == 0)
	    { co->output_char ('\n');
	    }
	   else
	    { co->output_str (ps->suffix());
	    }
	   return co->nchars();
	 }
}

