/* New stuff for event handling. Here is the basic idea:

   Grf4 defines several different types of button and valuator
   devices. 

   The user can poll the state of a device directly, using 

		grf4PollEvent (code, &event)

   The user can also arrange for a callback function to be attached
   to changes in the device state, using 

		grf4AttachEvent (type, fxn, arg)

   or ask for device state changes to be queued, using

		grf4QueueEvent (type)

   or release the event, using

	 	grf4ReleaseEvent (type);

   The event queue can be read using

		grf4ReadQueue (&event)

   which returns 0 or 1 depending on if anything is there.

   The queue can be flushed using 

		grf4FlushQueue()

   and can be tested using 

		grf4TestQueue()

   The functions grf4Wait(fd) and grf4Select() provide ways to process
   incoming GRF events while waiting for activity on other file descriptors.
   grf4CycleWait() returns with a specified period, while meanwhile
   allowing GRF events to be handled with a user-specified callback.
   Events can be processes explicitly using grf4Cycle().


   The GRF functions in turn use the following interface functions
   to the underlying graphics:

	getEventFd()

		returns a file descriptor while can be used to check
 		for incoming device events.

	enableEvent(win,type)
	disableEvent(win,type)

		Enables or disables the event type described by 'type' on 
		the window described by 'win'. At the moment, 'win' is one of
		MAINWIN (the main graphics window), or CTRLWIN (the 
		control panel window). The previous setting is returned.

	pollEvent(win,code,event*)

		Directly polls for the status of event 'code' with
		respect to the window 'win' (if applicable). Information
		is returned in the event structure.

	readEvent(event*)

		If an event is in the queue, returns 1 and reads the
		event into the event structure. Otherwise, 0 is returned.
*/

#include <errno.h>

#include <cdefs.h>

#include <grf4.h>
#include "grf4Tval.h"

#include "grf4Internal.h"

#include <sys/types.h>
#include <sys/time.h>

static int winw;
static int winh;
static int panw;
static int panh;

#define fxntab grf4FunctionTable
extern struct grf4fxnTable fxntab;

struct intStringPair {
	int code;
	char *str;
};

static EventDispatch *evDispatch[GRF_MAX_EVENT_TYPES+1];

/* The event queue is a fixed ring fifo, with 'head' and 'tail' indicies:

	|X|X|X|_|_|_|_|_|_|_|_|_|_|_|_|
	 |     |
  tail --+     +-- head

  We insert at the head and remove at the tail. 
 */

static Grf4Event queueBuf[GRF_EVENTQSIZE];
static int queueHead = 0;
static int queueTail = 0;

static int doViewControl;
static int doControlPanel;

static void processEvents();
static void reserveEvent();
static void unreserveEvent();
static int advanceQIndex();
static void queueEvent();
static void queueNewEvent();
static int dispatchEvent();
static int dispatchNewEvent();

static long keepEnabled;	/* MAINWIN events that must remain enabled */

struct intStringPair devNameDefs[] = {
	GRF_KB_A,		"A",
	GRF_KB_B,		"B",
	GRF_KB_C,		"C",
	GRF_KB_D,		"D",
	GRF_KB_E,		"E",
	GRF_KB_F,		"F",
	GRF_KB_G,		"G",
	GRF_KB_H,		"H",
	GRF_KB_I,		"I",
	GRF_KB_J,		"J",
	GRF_KB_K,		"K",
	GRF_KB_L,		"L",
	GRF_KB_M,		"M",
	GRF_KB_N,		"N",
	GRF_KB_O,		"O",
	GRF_KB_P,		"P",
	GRF_KB_Q,		"Q",
	GRF_KB_R,		"R",
	GRF_KB_S,		"S",
	GRF_KB_T,		"T",
	GRF_KB_U,		"U",
	GRF_KB_V,		"V",
	GRF_KB_W,		"W",
	GRF_KB_X,		"X",
	GRF_KB_Y,		"Y",
	GRF_KB_Z,		"Z",
	GRF_KB_0,		"0",
	GRF_KB_1,		"1",
	GRF_KB_2,		"2",
	GRF_KB_3,		"3",
	GRF_KB_4,		"4",
	GRF_KB_5,		"5",
	GRF_KB_6,		"6",
	GRF_KB_7,		"7",
	GRF_KB_8,		"8",
	GRF_KB_9,		"9",
	GRF_KB_Break,		"Break",
	GRF_KB_Ctrl,		"Ctrl",
	GRF_KB_LeftCtrl,	"LeftCtrl",
	GRF_KB_CapsLock,	"CapsLock",
	GRF_KB_RightShift,	"RightShift",
	GRF_KB_LeftShift,	"LeftShift",
	GRF_KB_Esc,		"Esc",
	GRF_KB_Tab,		"Tab",
	GRF_KB_Ret,		"Ret",
	GRF_KB_Space,		"Space",
	GRF_KB_Linefeed,	"Linefeed",
	GRF_KB_Backspace,	"Backspace",
	GRF_KB_Del,		"Del",
	GRF_KB_Semicolon,	"Semicolon",
	GRF_KB_Period,		"Period",
	GRF_KB_Comma,		"Comma",
	GRF_KB_Quote,		"Quote",
	GRF_KB_Backquote,	"Backquote",
	GRF_KB_Minus,		"Minus",
	GRF_KB_Slash,		"Slash",
	GRF_KB_Backslash,	"Backslash",
	GRF_KB_Equal,		"Equal",
	GRF_KB_LeftBracket,	"LeftBracket",
	GRF_KB_RightBracket,	"RightBracket",
	GRF_KB_LeftArrow,	"LeftArrow",
	GRF_KB_DownArrow,	"DownArrow",
	GRF_KB_RightArrow,	"RightArrow",
	GRF_KB_UpArrow,		"UpArrow",
	GRF_KB_Pad0,		"Pad0",
	GRF_KB_Pad1,		"Pad1",
	GRF_KB_Pad2,		"Pad2",
	GRF_KB_Pad3,		"Pad3",
	GRF_KB_Pad4,		"Pad4",
	GRF_KB_Pad5,		"Pad5",
	GRF_KB_Pad6,		"Pad6",
	GRF_KB_Pad7,		"Pad7",
	GRF_KB_Pad8,		"Pad8",
	GRF_KB_Pad9,		"Pad9",
	GRF_KB_PadF1,		"PadF1",
	GRF_KB_PadF2,		"PadF2",
	GRF_KB_PadF3,		"PadF3",
	GRF_KB_PadF4,		"PadF4",
	GRF_KB_PadPeriod,	"PadPeriod",
	GRF_KB_PadMinus,	"PadMinus",
	GRF_KB_PadComma,	"PadComma",
	GRF_KB_PadEnter,	"PadEnter",
	GRF_KB_LeftAlt,		"LeftAlt",
	GRF_KB_RightAlt,	"RightAlt",
	GRF_KB_RightCtrl,	"RightCtrl",
	GRF_KB_F1,		"F1",
	GRF_KB_F2,		"F2",
	GRF_KB_F3,		"F3",
	GRF_KB_F4,		"F4",
	GRF_KB_F5,		"F5",
	GRF_KB_F6,		"F6",
	GRF_KB_F7,		"F7",
	GRF_KB_F8,		"F8",
	GRF_KB_F9,		"F9",
	GRF_KB_F10,		"F10",
	GRF_KB_F11,		"F11",
	GRF_KB_F12,		"F12",
	GRF_KB_PrintScreen,	"PrintScreen",
	GRF_KB_ScrollLock,	"ScrollLock",
	GRF_KB_Pause,		"Pause",
	GRF_KB_Insert,		"Insert",
	GRF_KB_Home,		"Home",
	GRF_KB_PageUp,		"PageUp",
	GRF_KB_End,		"End",
	GRF_KB_PageDown,	"PageDown",
	GRF_KB_NumLock,		"NumLock",
	GRF_KB_PadDivide,	"PadDivide",
	GRF_KB_PadMultiply,	"PadMultiply",
	GRF_KB_PadPlus,		"PadPlus",

	GRF_LEFT_MOUSE,		"LeftMouse",
	GRF_MIDDLE_MOUSE,	"MiddleMouse",
	GRF_RIGHT_MOUSE,	"RightMouse",
	GRF_CURSOR,		"Cursor",

	GRF_REDRAW,		"Redraw",
	GRF_RESIZE,		"Resize",
};

static char *devNameMap[256];

static int eventCodeToType (int code)
{
	/* Returns the event type which corresponds to the event
	   described by 'code' */

	if (GRF_IS_KEY(code))
	 { return (GRF_KEY_EVENT);
	 }
	else if (GRF_IS_BUTTON(code))
	 { return (GRF_BUTTON_EVENT);
	 }
	else
	 { switch (code)
	    { case GRF_CURSOR:
	       { return (GRF_MOTION_EVENT);
	       }
	      case GRF_REDRAW:
	       { return (GRF_REDRAW_EVENT);
	       }
	      case GRF_RESIZE:
	       { return (GRF_RESIZE_EVENT);
	       }
	      case GRF_FOCUS_CHANGE:
	       { return (GRF_FOCUS_EVENT);
	       }
	      default:
	       { return (-1);
	       }
	    }
	 }
}

void grf4InitEvents(int flags, int w, int h, int pw, int ph)
{
	int i;

	for (i=0; i<ARRAY_SIZE(devNameDefs); i++)
	 { devNameMap[devNameDefs[i].code] = devNameDefs[i].str;
	 }
	if (flags & GRF_VIEW_CONTROL)
	 { doViewControl = 1; 
	 }
	if (flags & GRF_CONTROL_PANEL)
	 { doControlPanel = 1;
	 }
	if (doViewControl || doControlPanel)
	 { (*fxntab.enableEvent) (CTRLWIN, GRF_MOTION_EVENT);
	   (*fxntab.enableEvent) (CTRLWIN, GRF_FOCUS_EVENT);
	   (*fxntab.enableEvent) (CTRLWIN, GRF_BUTTON_EVENT);
	   (*fxntab.enableEvent) (MAINWIN, GRF_MOTION_EVENT);
	   (*fxntab.enableEvent) (MAINWIN, GRF_FOCUS_EVENT);
	   (*fxntab.enableEvent) (MAINWIN, GRF_BUTTON_EVENT);
	   keepEnabled = (GRF_MOTION_MASK|GRF_BUTTON_MASK|GRF_FOCUS_MASK);
	 }
/* TODO: fix this using an event query */	
	winw = w;
	winh = h;
	panw = pw;
	panh = ph;
}


char *grf4DevName (int code)
{
	return (devNameMap[code]);
}

void grf4Wait(int sd)
{
	fd_set fdset, rset;
	int qfd, n;

	qfd = (*fxntab.getEventFd)();
	do
	 { FD_ZERO (&fdset);
	   if (sd >= 0) FD_SET (sd, &fdset);
	   if (qfd >= 0) FD_SET (qfd, &fdset);
	   do
	    { rset = fdset;
	      n = select (FD_SETSIZE, &rset, (fd_set*)0, (fd_set*)0, 
					     (struct timeval*)0);
	    }
	   while (n < 0 && errno == EINTR);
	   if (n < 0)
	    { perror ("grf4Wait() -- error in select");
	      exit (-1);
	    }
	   if (qfd >= 0 && FD_ISSET(qfd, &fdset))
	    { processEvents();	/* queue event occurred; handle */
	    }
	   if (queueHead != queueTail)
	    { break;
	    }
	 }
	while (sd >= 0 ? !FD_ISSET(sd, &fdset) : 1);
}

int grf4Select (int *gEvnt, fd_set *rfds, fd_set *wfds, fd_set *efds, 
		struct timeval *timeout)
{
	struct timeval startTime, curTime, tv;
	struct timeval *tvp;
	fd_set rset;
	int qfd, n;

	if (!rfds)
	 { rfds = &rset;
	   FD_ZERO (rfds);
	 }
	if (gEvnt)
	 { *gEvnt = 0;
	 }

	qfd = (*fxntab.getEventFd)();
	if (qfd >= 0) FD_SET (qfd, rfds);
	if (timeout)
	 { tvp = &tv;
	   *tvp = *timeout;
	   gettimeofday (&startTime, (struct timezone*)0);
	 }
	else
	 { tvp = NULL;
	 }
	do
	 { n = select (FD_SETSIZE, rfds, wfds, efds, tvp);
	   if (tvp && n < 0 && errno == EINTR)
	    { gettimeofday (&curTime, (struct timezone*)0);
	      *tvp = grf4SubTval(*timeout,grf4SubTval(curTime,startTime));
	      if (grf4CmpTval(*tvp, grf4ZeroTval()) < 0)
	       {  *tvp = grf4ZeroTval();
	       }
	    }
	 }
	while (n < 0 && errno == EINTR);
	if (n > 0)
	 { if (qfd >= 0 && FD_ISSET(qfd, rfds))
	    { processEvents();	/* queue event occurred; handle */
	      FD_CLR (qfd, rfds);
	      if (gEvnt) *gEvnt = 1;
	      n--;
	    }
	 }
	return(n);
}

void grf4Cycle()
{
	processEvents();
}

int grf4QueueEvent (int type)
{
	EventDispatch *disp;

	if (type <= 0 || type > GRF_MAX_EVENT_TYPES)
	 { 
	   return (-1);
	 }
	if ((disp=evDispatch[type]) == NULL)
	 { disp = CALLOC(1,EventDispatch);
	   disp->type = type;
	   (*fxntab.enableEvent)(MAINWIN,type);
	   evDispatch[type] = disp;
	 }
	disp->flags &= ~EVENT_ATTACH;
	disp->flags |= EVENT_QUEUE;
	return (1);
}

int grf4AttachEvent (int type, void (*fxn)(Grf4Event *, int), int arg)
{
	EventDispatch *disp;

	if (type <= 0 || type > GRF_MAX_EVENT_TYPES)
	 { return (-1);
	 }
	if ((disp=evDispatch[type]) == NULL)
	 { disp = CALLOC(1,EventDispatch);
	   disp->type = type;
	   (*fxntab.enableEvent)(MAINWIN,type);
	   evDispatch[type] = disp;
	 }
	disp->flags &= ~EVENT_QUEUE;
	disp->flags |= EVENT_ATTACH;
	disp->fxn = fxn;
	disp->arg = arg;
	return (1);
}

int grf4ReleaseEvent (int type)
{
	EventDispatch *disp;

	if (type <= 0 || type > GRF_MAX_EVENT_TYPES)
	 { return (-1);
	 }
	if ((disp=evDispatch[type]) != NULL)
	 { FREE (disp);
	   evDispatch[type] = NULL;
	   if (!((1<<type) & keepEnabled))
	    { (*fxntab.disableEvent)(MAINWIN,type);
	    }
	 }
	return (1);
}


static int doReadEvent (Grf4Event *ev)
{
	int rv;

	/* Read the event queue, weeding out NULLEVENTS */

	do
	 { rv = (*fxntab.readEvent)(ev);
	 }
	while (rv && ev->code == GRF_NULLEVENT);
	return (rv);
}

static void processEvents()
{
	/* TODO: finish putting in event queueing and dispatching */

	Grf4Event ev, evx;

	while (doReadEvent(&ev))
	 { switch (ev.code)
	    { 
	      default:
	       { if (GRF_IS_KEY(ev.code))
		  { dispatchEvent (GRF_KEY_EVENT, &ev);
		  }
		 else if (GRF_IS_BUTTON(ev.code))
		  { 
#if 0
		    printf ("%s %s\n", devNameMap[ev.code], 
					ev.val1 ? "pressed" : "released");
#endif
		    dispatchEvent (GRF_BUTTON_EVENT, &ev);
		  }
		 else
		  { printf ("code = %d\n", ev.code);
		  }
		 break;
	       }
	      case GRF_NULLEVENT:
	       { break;
	       }
	      case GRF_REDRAW:
	       { grf4OnRedrawEvent (ev.win);
		 dispatchEvent (GRF_REDRAW_EVENT, &ev);
		 break;
	       }
	      case GRF_RESIZE:
	       { if (ev.win == MAINWIN)
		  { winw = ev.val1;
		    winh = ev.val2;
		    grf4OnResizeEvent (ev.win, ev.val1, ev.val2);
		  }
		 dispatchEvent (GRF_RESIZE_EVENT, &ev);
		 break;
	       }
	      case GRF_LEFT_MOUSE:
	       { (*fxntab.pollEvent)(ev.win, GRF_CURSOR, &evx);
		 grf4OnButtonEvent (ev.win, 1, (ev.val1!=0), evx.val1, evx.val2);
		 dispatchEvent (GRF_BUTTON_EVENT, &ev);
		 break;
	       }
	      case GRF_MIDDLE_MOUSE:
	       { (*fxntab.pollEvent)(ev.win, GRF_CURSOR, &evx);
		 grf4OnButtonEvent (ev.win, 2, (ev.val1!=0), evx.val1, evx.val2);
		 dispatchEvent (GRF_BUTTON_EVENT, &ev);
		 break;
	       }
	      case GRF_RIGHT_MOUSE:
	       { (*fxntab.pollEvent)(ev.win, GRF_CURSOR, &evx);
		 grf4OnButtonEvent (ev.win, 3, (ev.val1!=0), evx.val1, evx.val2);
		 grf4JackOnButton (&ev, (void*)0, 0);
		 dispatchEvent (GRF_BUTTON_EVENT, &ev);
		 break;
	       }
	      case GRF_CURSOR:
	       { grf4OnMotionEvent (ev.win, ev.val1, ev.val2);
		 grf4CalcVirtualCursor (&ev.vx1, &ev.vx2, ev.val1, ev.val2);
		 grf4JackOnCursor (&ev, (void*)0, 0);
		 dispatchEvent (GRF_MOTION_EVENT, &ev);
/*		 printf ("Motion: %d %d\n", ev.val1, ev.val2); */
		 break;
	       }
	      case GRF_FOCUS_CHANGE:
	       { grf4OnFocusChangeEvent (ev.val1); /* val1 contains the new window */
/* TODO: fix this up */
		 dispatchEvent (GRF_FOCUS_EVENT, &ev);
		 break;
	       }
	    }
	 }
}

static int dispatchEvent (int type, Grf4Event *ev)
{

	EventDispatch *disp;


	if (ev->win != MAINWIN)
	 { return (0);
	 }
	disp = evDispatch[type];
	if (disp == 0)
	 { 
	   return (0);
	 }
	else if (disp->flags & EVENT_QUEUE)
	 { queueEvent (ev);
	   return (1);
	 }
	else if (disp->flags & EVENT_ATTACH)
	 { (*disp->fxn)(ev, disp->arg);
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static int dispatchNewEvent (int type, int code, int val1, int val2)
{
	Grf4Event ev;

	ev.code = code;
	ev.val1 = val1;
	ev.val2 = val2;
	return (dispatchEvent (type, &ev));
}

int grf4PollEvent (int code, Grf4Event *ev)
{
	Grf4Event evx;

	if (code == GRF_FOCUS_CHANGE)
	 { ev->code = code;
	   ev->val2 = -1;

	   (*fxntab.pollEvent)(MAINWIN, GRF_FOCUS_CHANGE, &evx);
	   ev->val1 = (evx.val1 == MAINWIN);
	 }
	else
	 { (*fxntab.pollEvent)(MAINWIN, code, ev);
	   if (code == GRF_CURSOR)
	    { grf4CalcVirtualCursor (&ev->vx1, &ev->vx2, ev->val1, ev->val2);
	    }
	 }
	return (1);
}

static int advanceQIndex (int i)
{
	return ((i+1) % GRF_EVENTQSIZE);
}

static void queueEvent (Grf4Event *ev)
{
	queueBuf[queueHead] = *ev;
	queueHead = advanceQIndex(queueHead);
	if (queueHead == queueTail)
	 { queueTail = advanceQIndex(queueTail);
	 }
}

static void queueNewEvent (int code, int val1, int val2)
{
	Grf4Event ev;

	ev.code = code;
	ev.val1 = val1;
	ev.val2 = val2;
	queueEvent (&ev);
}

int grf4ReadQueue (Grf4Event *ev)
{
	if (queueHead == queueTail)
	 { return (0);
	 }
	else
	 { *ev = queueBuf[queueTail];
	   queueTail = advanceQIndex(queueTail);
	   return (1);
	 }
}

int grf4TestQueue ()
{
	return (queueHead != queueTail);
}

int grf4FlushQueue ()
{
	queueHead = queueTail;
}

void grf4CycleWait(int interval, struct timeval *nextCycleTime, 
		   int (*handleEvent)())
{
	static struct timeval zeroTime;
	static struct timeval minTime;
	struct timeval currentTime;
	struct timeval intervalTv;
	struct timeval naptime;

	Grf4Event ev;
	int gset;

	intervalTv.tv_sec = interval/1000;
	intervalTv.tv_usec = (interval%1000)*1000;

	if (nextCycleTime->tv_sec == 0 && nextCycleTime->tv_usec == 0)
	 { gettimeofday (&currentTime, (struct timezone *)0);
	   *nextCycleTime = grf4AddTval (currentTime, intervalTv);
	   zeroTime = grf4ZeroTval();
	   minTime = grf4ZeroTval();
	 }
	else
	 { *nextCycleTime = grf4AddTval (*nextCycleTime, intervalTv);
	 }

	if (interval < 17)
	 { /* XXX XXX XXX rate > 60 */
	   minTime.tv_usec = 10000;
	 }
	else
	 { minTime.tv_usec = 0;
	 }

	while(1)
	 { gettimeofday (&currentTime, (struct timezone *)0);
	   naptime = grf4SubTval (*nextCycleTime, currentTime);
	   if (grf4CmpTval (naptime, minTime) < 0)
	    { *nextCycleTime = grf4AddTval (currentTime, minTime);
	      naptime = minTime;
	    }
	   grf4Select (&gset, (fd_set*)0, (fd_set*)0, (fd_set*)0, &naptime);
	   if (gset)
	    { while (grf4ReadQueue(&ev))
	       { (*handleEvent) (&ev);
	       }
	    }
	   else
	    { return;
	    }
	 }
}

static int computeDeltaTime (struct timeval *tinc, struct timeval *tabs)
{
	/* Compute tinc to be the difference between the current
	   time and some future reference time tabs. If this is
	   is in fact less than or equal to zero, set tinc to
	   0 and return 0; otherwise, return 1 */

	static struct timeval tzero = { 0, 0 };
	struct timeval tcur;

	gettimeofday (&tcur, (struct timezone *)0);
	*tinc = grf4SubTval (*tabs, tcur);
	if (grf4CmpTval(*tinc, tzero) <= 0)
	 { *tinc = tzero;
	   return (0);
	 }
	else
	 { return (1);
	 }
}

int grf4CycleSelect (int *gCode, fd_set *rfds, fd_set * wfds, fd_set *efds, 
		     struct timeval *expireTime)
{
	/* This routine acts like a select call, except that it
	   also checks for events requested in *gCode.

	   These include:

	   graphic events (GRF_GRAPHICS_EVENT)
	   a timer event  (GRF_TIMER_EVENT)

	   A timer event, if requested, occurs when the current
	   system time equals or exceeds expireTime.
	 */
	   
	struct timeval *tvp;
	struct timeval tinc;
	fd_set rset;
	int qfd, n;
	int code;
	int gevent = 0;
	int tevent = 0;

	if (!rfds)
	 { rfds = &rset;
	   FD_ZERO (rfds);
	 }
	qfd = (*fxntab.getEventFd)();
	if (qfd >= 0) FD_SET (qfd, rfds);

	code = *gCode;
	if (code & GRF_TIMER_EVENT)
	 { tvp = &tinc;
	   computeDeltaTime (tvp, expireTime);
	 }
	else
	 { tvp = NULL;
	 }
	do
	 { n = select (FD_SETSIZE, rfds, wfds, efds, tvp);
	   if (tvp)
	    { tevent = !computeDeltaTime (tvp, expireTime);
	    }
	   if (n < 0 && errno == EINTR)
	    { n = 0;
	    }
	   else if (n > 0 && qfd >= 0 && FD_ISSET(qfd, rfds))
	    { processEvents();  /* queue event occurred; handle */
	      FD_CLR (qfd, rfds);
	      n--;
	      if (code & GRF_GRAPHICS_EVENT)
	       { gevent = 1;
	       }
	    }
	   
	 }
	while (n == 0 && !tevent && !gevent);

	*gCode = 0;
	if (tevent)
	 { *gCode |= GRF_TIMER_EVENT;
	 }
	if (gevent)
	 { *gCode |= GRF_GRAPHICS_EVENT;
	 }
	return(n);
}
