#ifndef _GRF4INTERNAL_H_
#define _GRF4INTERNAL_H_

/* Event flags */

#include <grf4.h>

#ifndef PROTO
#ifndef NO_PROTOTYPES
#define PROTO(x) x
#else
#define PROTO(x) ()
#endif
#endif

#ifdef __cplusplus
extern "C"
{
#endif

struct localFxnTable {
	void (*draw);
	void (*clear);
	void (*flush);
  };

typedef struct buttonDesc {
	struct buttonDesc *next;
	int num;
	char *name;
	int flags;
	int sizeK;
	int x, y, w, h;
	int coffX;
	int coffY;
	int minw;
	int minh;
	void (*fxn)(int, int, long);
	long arg;
	int row;
	int rowSizeK;
	int rowMinW;
	int rowN;
  } BUTTON_DESC;

#if HLS

#define HIDE_EVENT	1
#define EXPOSE_EVENT	2

#define P_DONT_INTERSECT	0
#define P_INTERSECT		1
#define P_POLY1_INSIDE		2
#define P_POLY2_INSIDE		3

typedef struct pevent {
	struct pevent *next;	/* used for event list */
	float r;		/* point along edge/vector of event */
	int type;		/* type of event: expose or hide */
} PEVENT;

typedef struct vector {
	/* Describes a single vector */

	struct vector *next;	/* forms double linked list of */
	struct vector *prev;	/*   vectors that must be drawn */
	float x1, y1, ez1;	/* endpoints coordinates for the vector */
	float x2, y2, ez2;	/*   x,y in screen and ez in eye coords */

	int visible;		/* indicates entire vector is invisible */
	PEVENT *events;		/* list of hide/expose events */
	
	/* fields below are used by the intersection code */

	float rOut;		/* point where vector exits polygon */
	float rInto;		/* point where vector enters polygon */
} VECTOR;

typedef struct pedge {
	/* Describes a vertex associated with a polygon, as well as
	   the edge formed by travelling in the counter-clockwise
	   ('next') direction to the next vertex.

	   An edge can be obscured by other polygons. Information
	   on how the edge is obscured is kept as follows:

	   'visible' denotes whether the beginning of the edge (at
	   the vertex) is visible. 'nextEvents' is a list of hide
	   and expose events for the edge. If 'visible' is true and
	   'nextEvents' is NULL, the edge is completely visible. */

	struct pedge *next;	/* pointers for linked list of */
	struct pedge *prev;	/*   vertices/edges */
	PEVENT *nextEvents;	/* list of hide/expose events */
	PEVENT *prevEvents;	/* closest event on previous edge, if any
				   (not used at time of this writing) */

	double x, y;		/* x and y screen coordinates for vertex */
	float ez;		/* z eye coordinates for vertex */
	int visible;		/* true if the vertex is visible. */

	/* the fields below are used by the polygon intersect code. */

	int intersect;		/* there is an intersection on this edge */
#define INTO	1		/* edge is ``entering'' the other polygon */
#define OUT	2		/* edge is ``exiting'' the other polygon */
	float rInto;		/* distance along edge for enter point */
	float rOut;		/* distance along edge for exit point */
} PEDGE;

typedef struct polygon {
	struct polygon *next;	/* prev and next create a double linked */
	struct polygon *prev;	/* list of the polygons to be drawn */
	int visible;		/* is polygon is completely invisible? */ 
	int n;			/* number of edges in the polygon */

	/* polygons edges are stored as a double circular linked list
	   of vertices, in counter-clockwise order */

	PEDGE *edges;		/* pointer to edge list */
	float nrml[3];		/* normal vector for the polygon */

	float xmax, xmin;	/* these define a bounding box for the */
	float ymax, ymin;	/* polygon. x and y are in screen */
	float ezmax, ezmin;	/* coordinates; ez is in eye coordinates */
	int objCnt;		/* indicates parent object, if any */
} SCR_POLYGON;

typedef struct _IEDGE {

	/* vertices/edges for a polygon created by the intersection
	   of two polygons */

	struct _IEDGE *next;
	struct _IEDGE *prev;

	int intersection;	/* vertex is an intersection point */
	SCR_POLYGON *poly;	/* pointer to current inside polygon */
	PEDGE *edgeIn;		/* pointer to current inside edge */

	/* these next three fields are set only if there is an 
	   intersection. */

	float rIn;		/* inside edge distance to intersection */
	PEDGE *edgeOut;		/* pointer to current outside edge */
	float rOut;		/* outside edge distance to intersection */
} IEDGE;

typedef struct {
	/* Polygon created by the intersection of two polygons */
	int n;			/* number of vertices */
	IEDGE *edges;		/* pointer to list of edges */
} INTER_POLYGON;

typedef struct {
	int left, right;
} AlistElem;

#define NULLWIN	0
#define MAINWIN	1
#define CTRLWIN	2

#define NUMWINS	3

typedef struct {
	int type;		/* backpointer to the event */
	int flags;
#define EVENT_QUEUE	0x01
#define EVENT_ATTACH	0x02
#define EVENT_RESERVED	0x04	/* event is reserved for internal use */
	void (*fxn)(Grf4Event *, int);
	int arg;
} EventDispatch;

typedef struct {
	int placed;
	int w, h;
	int x, y;
	int minW, minH;
} WinPlaceDesc;

void alistBuild PROTO((AlistElem**, AlistElem**, AlistElem*, int));
int alistSearch PROTO((int*, int, AlistElem*, int));

VECTOR* grf4NewVector();
PEDGE* grf4NewPedge();
PEVENT* grf4NewPevent();
SCR_POLYGON* grf4NewPolygon();
int grf4VectorIntersect PROTO((SCR_POLYGON *poly, VECTOR *vect, 
			       float *xp, float *yp));
void grf4VectorClip PROTO((VECTOR *v));
void grf4VectorDraw PROTO((VECTOR *v));
int grf4IntersectPolygons PROTO((SCR_POLYGON *poly1, SCR_POLYGON *poly2,
	int draw, PEDGE **fp, PEDGE **fq, 
	float *cntrXp, float *cntrYp, INTER_POLYGON *isect));

#endif

struct grf4fxnTable {
	long (*capacity)PROTO(());

	void (*ortho)PROTO((double, double, double, double, double, double));
	void (*window)PROTO((double, double, double, double, double, double));
	void (*loadTrans)PROTO((TRSF*));
	void (*color)PROTO((unsigned long));
	void (*clear)PROTO(());
	void (*flush)PROTO(());
	void (*move)PROTO((double, double, double));
	void (*draw)PROTO((double, double, double));
	void (*point)PROTO((double, double, double));
	void (*move2)PROTO((double, double));
	void (*draw2)PROTO((double, double));
	void (*point2)PROTO((double, double));
	void (*moveRaw)PROTO((double, double));
	void (*drawRaw)PROTO((double, double));
	void (*pointRaw)PROTO((double, double));
	void (*drawPrism)PROTO((Grf4Object*));
	void (*drawSphere)PROTO((Grf4Object*));
	void (*drawPolygon)PROTO((Grf4Object*));
	void (*drawLineSet)PROTO((Grf4Object*));
	void (*drawPolyhed3)PROTO((void*));
	PRISM* (*genDisc)PROTO((int, double, double));
	PRISM* (*genCone)PROTO((int, double, double, double));
	PRISM* (*genPrism)PROTO((int, float*, float*, char*, double));
	SPHERE* (*genSphere)PROTO((int, double));
	POLYGON* (*genPolygon)PROTO((int, float*));
	POLYGON* (*genRegPolygon)PROTO((int, double));
	LINESET* (*genLineSet)PROTO((int, float*));
	double (*setDepth)PROTO((Grf4Object*, double));
	int (*overPrism)();
	int (*overSphere)();

	void (*putText)PROTO((char*,int,int));
	void (*getTextExtent)PROTO((int*,int*,int*,int*,char*));
	void (*drawButton)PROTO((BUTTON_DESC*));
	void (*getButtonExtent)PROTO((BUTTON_DESC*));

	void (*pxlDraw)PROTO((int, int, int, int));
	void (*pxlPoint)PROTO((int, int));
	void (*pxlClear)PROTO((int, int, int, int));
	void (*pxlFlush)PROTO((int, int, int, int));
	void (*pxlPutText)PROTO((char*,int,int));

	void* (*defLight)PROTO((unsigned long, unsigned long, 
				double, double, double, double));
	void* (*defMaterial)PROTO((unsigned long, unsigned long, unsigned long,
				   unsigned long, double, double));
	void* (*defMaterialD)PROTO((MATERIAL_DEF*));
	void (*setMaterial)PROTO((void*));
	void (*setLight)PROTO((int, void*));
	void (*resetLights)PROTO(());

	int (*getEventFd)PROTO(());
	int (*enableEvent)PROTO((int, int));
	int (*disableEvent)PROTO((int, int));
	int (*pollEvent)PROTO((int, int, Grf4Event*));
	int (*readEvent)PROTO((Grf4Event*));

	void (*getWindowSize)PROTO((int*, int*, int*));
  };

void PostScriptInit (FILE *fp, char *name, int flags, double w, double h);
void PostScriptXform (float *xmp, float *xbp, float *ymp, float *ybp);
void PostScriptBoxSet (double llx, double lly, double urx, double ury);
void PostScriptClose();
void PostScriptFinish();
void PostScriptMove (double x, double y);
void PostScriptDraw (double x, double y);
void PostScriptVector (VECTOR *v);
void PostScriptPolygon (SCR_POLYGON *pgn);

int grf4glSetupFxns (struct grf4fxnTable *);
int grf4glwfSetupFxns (struct grf4fxnTable *);
int grf4GLSetupFxns (struct grf4fxnTable *);
int grf4SvSetupFxns (struct grf4fxnTable *);
int grf4X11SetupFxns (struct grf4fxnTable *);

int grf4glInit (char *, int, int, WinPlaceDesc *, int);
int grf4glwfInit (char *, int, int, WinPlaceDesc *, int);
int grf4GLInit (char *, int, int, WinPlaceDesc *, int);
int grf4SvInit (char *, int, int, WinPlaceDesc *, int);
int grf4X11Init (char *, int, int, WinPlaceDesc *, int);

int grf4DoFxnInit (int flags);

void grf4OnButtonEvent(int win, int num, int pressed, int x, int y);
void grf4OnMotionEvent (int win, int x, int y);
void grf4OnFocusChangeEvent (int win);
void grf4OnResizeEvent(int win, int w, int h);
void grf4OnRedrawEvent(int win);
void grf4ButtonInit(int *w, int *h);
void grf4CalcVirtualCursor (double *vxp, double *vyp, int x, int y);
int grf4GetDrawObjCnt();
int grf4IncDrawObjCnt();

void grf4JackOnCursor (Grf4Event *ev, void *ptr, int arg);
void grf4JackOnButton (Grf4Event *ev, void *ptr, int arg);

TRSF *grf4GetJackTrsfPtr();
TRSF *grf4GetViewTrsfPtr();
TRSF *grf4GetTransPtr();

#ifdef __cplusplus
}
#endif

#endif
