#include <RxP2RKin.h>
#include <math.h>
#include <mathx.h>
#include <cdefs.h>

#define J1	0
#define J2	1

#define A1	dhpSet[J1].a
#define A2	dhpSet[J2].a

#define S1	dhpSet[J1].sT
#define S2	dhpSet[J2].sT
#define S12	s12

#define C1	dhpSet[J1].cT
#define C2	dhpSet[J2].cT
#define C12	c12

#define EPSILONF	(1e-13)
#define SING_TOL	(0.0001*DTOR)

#define SINCOS(s,c,ang)	s = sin(ang); c = cos(ang);
void RxP2RKin::p2Rinit()
{
	s12 = sin(dhpSet[J1].theta + dhpSet[J2].theta);
	c12 = cos(dhpSet[J1].theta + dhpSet[J2].theta);
}

void RxP2RKin::setDefaultP2R ()
{
	DHParams dhp[2];
	int i;

	for (i=0; i<2; i++)
	 { dhp[i].type = 'R';
	   dhp[i].theta = 0;
	   dhp[i].alpha = 0;
	   dhp[i].d = 0;
	   dhp[i].a = 0;
	 }
	dhp[0].a     =  10.0;
	dhp[1].a     =  10.0;
	set (dhp, 2);
}

int RxP2RKin::setP2RJnts (const RxVecd2 &jnts)
{
	double *jvec = jnts.buf();

	dhpSet[J1].theta = jvec[J1];
	SINCOS (S1, C1, jvec[J1]);
	dhpSet[J2].theta = jvec[J2];
	SINCOS (S2, C2, jvec[J2]);

	s12 = sin(jvec[J1]+jvec[J2]);
	c12 = cos(jvec[J1]+jvec[J2]);

	flags = 0;
	return (1);
}

int RxP2RKin::fwdK (RxTrsf *Tr, int *conf)
{
	int rcode = 0;

	if (Tr)
	 { 
	   Tr->R.n.x() = 1.0;
	   Tr->R.n.y() = 0.0;
	   Tr->R.n.z() = 0.0;

	   Tr->R.o.x() = 0.0;
	   Tr->R.o.y() = 1.0;
	   Tr->R.o.z() = 0.0;

	   Tr->R.a.x() = 0.0;
	   Tr->R.a.y() = 0.0;
	   Tr->R.a.z() = 1.0;

	   Tr->p.x() = A1 * C1 + A2 * C12;
	   Tr->p.y() = A1 * S1 + A2 * S12;
	   Tr->p.z() = 0.0;
	 }

	if (conf != NULL)
	 { /* Determine the configurations: */

	   *conf = 0;

	   if (S2 >= 0.0)
	    { *conf |= RX_PLANAR2RKIN_CONF_DOWN;
	    }
	   if (ABS(S2) <= SING_TOL)
	    { rcode |= RX_PLANAR2RKIN_SING_ELBOW;
	    }
	 }
	return (rcode);
}

void RxP2RKin::buildJacobian()
{
	RxTrsf Um, Am;

	// The Jacobian for this robot is simply
	//
	//     [ J11 ]
	// J = [     ]
	//     [  0  ]
	//
	// where J11 is 2X2 of the form
	//
	// [ A1 S2	  0  ]
	// [  		     ]
	// [ A2 + A1 C2   A2 ]
	//
	// Note that with respect to base coordinates, J is instead
	//
	// [ -A1 S1 - A2 S12	-A2 S12 ]
	// [  				]
	// [  A1 C1 + A2 C12	 A2 C12 ]

	j00 = A1*S2;
	j10 = A2 + A1*C2;
	j01 = 0;
	j11 = A2;

	flags |= RX_GENKIN_JACOBIAN_VALID;
}

RxMatd& RxP2RKin::Jset (RxMatd *Jr)
{
	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { buildJacobian();
	 }
	(*Jr)(0,0) = j00;
	(*Jr)(1,0) = j10;
	(*Jr)(2,0) = 0.0;
	(*Jr)(3,0) = 0.0;
	(*Jr)(4,0) = 0.0;
	(*Jr)(5,0) = 0.0;

	(*Jr)(0,1) = j01;
	(*Jr)(1,1) = j11;
	(*Jr)(2,1) = 0.0;
	(*Jr)(3,1) = 0.0;
	(*Jr)(4,1) = 0.0;
	(*Jr)(5,1) = 0.0;

	return (*Jr);
}

RxTwist& RxP2RKin::Jmul (RxTwist *tw, const RxVecd &djnts)
{
	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { buildJacobian();
	 }

	tw->v().x() = j00 * djnts(J1) + j01 * djnts(J2);
	tw->v().y() = j10 * djnts(J1) + j11 * djnts(J2);
	tw->v().z() = 0.0;
	tw->w().x() = 0.0;
	tw->w().y() = 0.0;
	tw->w().z() = 0.0;
	
	return (*tw);
}

RxVecd& RxP2RKin::JTmul (RxVecd *tor, const RxWrench &wr)
{
	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { RxGenKin::buildJacobian();
	 }

	(*tor)(J1) = j00 * wr.f().x() + j10 * wr.f().y();
	(*tor)(J2) = j01 * wr.f().x() + j11 * wr.f().y();

	return (*tor);
}

int RxP2RKin::invK (RxVecd *jnts, const RxTrsf &T, int conf,
		     const RxVecd *jref, int update)
{
	double a, b, k, u;
	double ang1, ang2, c2, s2;
	int rcode = 0;

	a = 2*T.p.y()*A1;
	b = 2*T.p.x()*A1;
	k = A1*A1 - A2*A2 + T.p.x()*T.p.x() + T.p.y()*T.p.y();

	u = (2*A1*A1-k)/(2*A1*A2);
	if (u > 1.0)
	 { u = 1.0;
	   rcode |= RX_PLANAR2RKIN_CLIP_OUTER;
	 }
	else if (u < -1.0)
	 { u = -1.0;
	   rcode |= RX_PLANAR2RKIN_CLIP_INNER;
	 }
	if (1-u*u <= SING_TOL*SING_TOL)
	 { rcode |= RX_PLANAR2RKIN_SING_ELBOW;
	 }

	if (conf == RX_PLANAR2RKIN_CONF_DOWN)
	 { ang2 = M_PI - acos(u);
	 }
	else
	 { ang2 = acos(u) - M_PI;
	 }

	c2 = cos(ang2);
	s2 = sin(ang2);
	if (ABS(T.p.x()) < EPSILONF*(A1+A2) && ABS(T.p.y()) < EPSILONF*(A1+A2))
	 { ang1 = (jref ? (*jref)(J1) : 0.0);
	   rcode |= (RX_GENKIN_INDETERMINATE | RX_PLANAR2RKIN_SING_ELBOW);
	 }
	else
	 { ang1 = atan2(T.p.y(),T.p.x()) - atan2(A2*s2,(A1+A2*c2));
	 }
	(*jnts)(J1) = ang1;
	(*jnts)(J2) = ang2;

	if (update)	/* Update sines and cosines if needed */
	 { S1 = sin(ang1);
	   C1 = cos(ang1);
	   dhpSet[J1].theta = (*jnts)(J1);
	   S2 = s2;
	   C2 = c2;
	   dhpSet[J2].theta = (*jnts)(J2);
	   S12 = s12;
	   C12 = c12;
	 }
	return(rcode);
}

int RxP2RKin::Jsolve (RxVecd *dj, const RxTwist &tw)
{
	int code = 0;

	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { RxGenKin::buildJacobian();
	 }

	if (ABS(S2) > SING_TOL)
	 { (*dj)(J1) = tw.v().x()/(A1*S2);
	   (*dj)(J2) = (tw.v().y() - (A1*C2+A2)*(*dj)(J1))/A2;
	 }
	else
	 { // ignore tw.v().x()
	   if (C2 > 0)
	    { // arm is outstretched

	      (*dj)(J1) = tw.v().y()/(A1+A2);
	      (*dj)(J2) = 0;
	    }
	   else
	    { // arm is folded up

	      (*dj)(J1) = tw.v().y()/(2*A2-A1);
	      (*dj)(J2) = tw.v().y()/(2*A2-A1);
	    }
	   code |= RX_PLANAR2RKIN_SING_ELBOW;
	 }
	return (code);
}

int RxP2RKin::JTsolve (RxWrench *wr, const RxVecd &tor)
{
	int code = 0;

	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { RxGenKin::buildJacobian();
	 }

	wr->f().z() = 0.0;
	wr->m().setZero();

	if (ABS(S2) > SING_TOL)
	 { wr->f().y() = tor(J2)/A2;
	   wr->f().x() = (tor(J1) - (A2+C2*A1)*wr->f().y())/(A1*S2);
	 }
	else
	 { wr->f().y() = tor(J2)/A2;
	   wr->f().x() = 0;
	   code |= RX_PLANAR2RKIN_SING_ELBOW;
	 }
	return (code);
}
