#ifndef _RCPATHKNOTS_H_WRAP_
#define _RCPATHKNOTS_H_WRAP_

#include <RxDlist.h>
#include <RxVecd3.h>
#include <RxVecd.h>
#include <RxTrsf.h>
#include <RxTwist.h>
#include <RxCvxPolygon2.h>

#include <RcPathTrajectory.h>

// Indices into the vectors of augmented coordinates
#define J1 0			// joint 1 ...
#define J2 1
#define J3 2
#define J4 3
#define J5 4
#define J6 5
#define J7 6			// joint 7
// XXX the following defs need the variable numj
#define JS	(numj+0)	// value of path parameter s 
#define JP	(numj+1)	// position path length
#define JR	(numj+2)	// rotational path length

// Codes describing different tests used when dividing the knot points
#define TEST_POS_TOL		0x01
#define TEST_JNT_DEL		0x02
#define TEST_DERIV_LEFT		0x20
#define TEST_DERIV_RIGHT	0x40
#define TEST_MONO		0x80
#define TEST_MONO_LEFT		0x04
#define TEST_MONO_RIGHT		0x08

#define TEST_DERIV      (TEST_DERIV_LEFT|TEST_DERIV_RIGHT)
#define TEST_ALL	(TEST_POS_TOL|TEST_JNT_DEL|TEST_DERIV)

// linked list of plot items

class RcPathKnot : public RxDnodeBase
{
	friend class RcPathTrajectory;

	static RcPathKnot *cache;	// cache of allocated knots
	int checkcode;			// used for debugging
	double *store;			// buffer space for vectors

	// These routines are to quickly allocate knots, by keeping
	// those already allocated around in a special cache.
	static RcPathKnot *allocK();
	static RcPathKnot *allocKList(int n);
	static void freeK (RcPathKnot *kp);
	static void freeKList (RcPathKnot *kp, RcPathKnot *lp);

	char *flagStr (char *buf);
	void print (const char *str, const char *jtype);
	void makeCorner();
	
public:
	RcPathKnot *link;		// link for knot caching
	static void reserve (int nj, int n);

	int numj;			// number of robot coordinates
	int numx;			// number of augmented coordinates

	int idx;			// knot index number

	int flags;			// misc flags:
#define KNOT_CORNER	0x01		// corresponds to a corner
#define KNOT_INTERPOLATED	0x02		// was inserted by interpolation
#define KNOT_CLIPPED	0x04		// X corresponds to clipped location

	RxTrsf X;			// spatial location along path

	// The stored coordinates include the joint coordinate values
	// (up to numj-1), the positional and rotational path lengths,
	// and the path parameter.

	RxVecd coordx;			// coordinate values

	// Each knot is used to store derivative and difference 
	// information for the succeeding interval. In the following
	// comments, assume that knot the current knot has an idx value
	// of k. Then knot->next() will have an idex value of k+1.
	// 
	// Both left and right derivative information is stored in
	// case of discontinuity. However, because computations
	// are usually done within a given interval, we store the
	// left derivative information associated with knot k+1.
	//
	//    |<-right derivatives	left derivatives->|
	//    |						  |
	//    +-------------------------------------------+
	//    k                                          k+1

	int jx;		    // index of driving coordinate x over [k,k+1]
	RxVecd delcx;	    // coordx[k+1] - coordx[k]
	RxVecd dcxdsR;     // right derivative of coordx WRT s, at k
	RxVecd dcxdsL;     // left derivative of coordx WRT s, at k+1
	RxVecd dcxdxR;	    // right derivative of coordx WRT x, at k
	RxVecd dcxdxL;	    // left derivative of coordx WRT x, at k+1
	RxVecd dcxdxA;	    // average derivative of coordx WRT x over [k,k+1]
	RxVecd d2cxdx2;    // average 2nd deriv of coordx WRT x over [k,k+1]

	RxCvxPolygon2 E;	    // polytope giving limits on energy at k and k+1
	double er;	    // path energy value of x at k
	double el;	    // path energy value of x at k+1

	RX_DNODE_DECLS (RcPathKnot);

	// Initialization and constructors

	void init(int nj);
	int set (RcPathInvKinFxn kfxn, double s, double dels, 
		 const char *jtype, int nj, RxVecd *old);

	RcPathKnot ()
	 { numj = 0;
	 }
	RcPathKnot (int nj)
	 { numj = 0;
	   init (nj);
	 }
	RcPathKnot (RcPathInvKinFxn kfxn, double s, double dels,
		    const char *jtype, int nj, RxVecd *old)
	 { numj = 0;
	   set (kfxn, s, dels, jtype, nj, old);
	 }
	~RcPathKnot ();

	// sets values that depend on adjacent knots
	void setPair (RcPathKnot *next, const RxVecd &range,
		      const char *jtype, int useLen);
	void set2ndDeriv();

	// Tests used to decide if we need to subdivide the path further
	int jntDelOK (RxVecd &maxDel, int defaultJx);
	int posTolOK (RcPathInvKinFxn, RcPathFwdKinFxn, 
	    double posTol, double rotTol, const char *jtype,RcPathKnot *next);

	int monoOK();
	int monoLeftOK();
	int monoRightOK();

	int derivLeftOK (const RxVecd &maxAcc, double jntxAccLim, 
		double maxAccErrRatio);
	int derivRightOK (const RxVecd &maxAcc, double jntxAccLim, 
		double maxAccErrRatio);

	// estimate an upper bound on acceleration for joint jx
	double estAccBound (const RxVecd &maxVel, 
		const RxVecd &maxAcc,int jx);

	// integrate a trajectory across a single knot value
	double integrate (double *lam, RcPathKnot *next, double t);

	double xToCoord (double delx, int j);
};

typedef RxDlist<RcPathKnot>  RcPathKnotList;

#if 0
double lamToJval (RcPathKnot *knot, double lam, int jx);
double lamToS (RcPathKnot *knot, double lam, int numj);
#endif
void rangeJnts (RxVecd *jr, const RxVecd &j1, const RxVecd *ref, 
		const char *jtype, int n);

#endif
